/*    .
 
  ..  ..*/

#ifndef _NORMACNF_H_
#define _NORMACNF_H_

typedef int BOOL;
#define TRUE 1
#define FALSE 0

#define MAX_LEN_NAME 32 /*   */
#define MAX_INDENT_LEN 64 /*  */
#define MAX_STR_LEN 200 /*    */

#ifdef _WIN32
#pragma warning(disable: 4127)  /* constant expression */
#define DIR_SEPARATOR '\\'
#define MYMAXPATH _MAX_PATH

#else

#define DIR_SEPARATOR '/'
#define strnicmp strncasecmp
#define stricmp strcasecmp
#define MYMAXPATH PATH_MAX

#endif

/*  */
typedef enum {GNS,PVM,FMPI,C,CMPI,F77} OUT_FILE_TYPE;
#define FMPI_OR_F77 (out_file_type==FMPI||out_file_type==F77)
#define C_OUT (out_file_type==C||out_file_type==CMPI)
#define MPI (out_file_type==FMPI||out_file_type==CMPI)

/**********************************************************/
#define DIAPAZON struct Diapazon
#define PARAM struct Param
DIAPAZON /* */
{
	char *index;
	long first;
	long last;
	DIAPAZON *next;
	char *fExpression; /*  ,   */
	char *lExpression;
	BOOL bfUnknown; /* */
	BOOL blUnknown;
};

typedef enum
{
	TYP_UNKNOWN,
	TYP_INT,
	TYP_REAL,
	TYP_DOUBLE,
	TYP_PARTNAME,
	TYP_FUNCNAME,
	TYP_INT1,
	TYP_INT2,
	TYP_INT8,
	TYP_INT16,
	TYP_BOOL,
	TYP_CHAR,
	TYP_VOID,
	TYP_OTHER,
	TYP_PROGRAMNAME
} VARTYPES;

typedef enum
{
	VRT_UNKNOWN,
	VRT_SCALAR,
	VRT_DIMENSION,
	VRT_FUNCTION,
	VRT_POINTER,
} VARIETY;

PARAM /* */
{
	char name[MAX_LEN_NAME+1]; /**/
	VARTYPES type; /**/
	VARIETY variety; /* , ,  */
	DIAPAZON *diapazon; /* ,  */
	BOOL bCommon; /*   COMMON   */
	long strlen; /* char  */
	PARAM *row; /*  */
	PARAM *exp; /*  .  */
	PARAM *next;
};

#define CALLPART struct callpart
CALLPART /*  */
{
	char name[MAX_LEN_NAME+1]; /* */
	char filename[MYMAXPATH]; /* */
	long line; /* */
	long intrinsicline; /*    */
	long meaningline; /*    */
	long mpi_initline; /*    MPI_Init*/
	long endline; /*  */
	int inparam; /*  rezult*/
	char find; /*0- , 1- , 2- include-*/
	VARTYPES type; /*,   */
	PARAM *parameters; /* */
	CALLPART *next; /* */
};

#define COMMON struct common1
COMMON /*  COMMON*/
{
	char name[MAX_LEN_NAME+1]; /* */
	PARAM *NextPosition; /*  ,      */
	PARAM *parameters; /* */
	BOOL bMainPart; /* COMMON    */
	COMMON *next; /* */
};

#define STRINGS struct strings3
STRINGS
{
	char *s;
	long num;
	STRINGS *next;
};

/*  ( d=)*/
#define DEFINES struct defines
DEFINES
{
	char *name;
	DEFINES *next;
};

/*     */
#define SUBROUTINE struct subroutine
SUBROUTINE
{
	char *name;
	CALLPART *formal;/* */
	CALLPART *actual;
	SUBROUTINE *next;
};

/*  */
#define FILES struct file
FILES
{
	char *name;
	BOOL exists_without_config;/*    ..*/
	STRINGS *sFirstStr;/*    */
	BOOL was_write;
	BOOL was_all_write;
	BOOL is_user_file;
	long crc;/*    */
	FILES *next;
};

#define SENDS struct sends11
SENDS
{
	int i,j;
	SENDS *next;
};

#define ERRORS struct errors
ERRORS
{
	int cod;
	char message[80];
	int return_cod; /* -1 -     */
	ERRORS *next;
};

#define CALLS struct calls
CALLS
{
	char message[80];
	char com_line[MAX_STR_LEN];
	char file1[120];
	char file2[120];
	ERRORS *e;
	CALLS *next;
};

/*      */
#define RASPR struct raspr
RASPR
{
	char name[MAX_INDENT_LEN];
	int i,j;
	int in;
	int jn;
	SENDS *s;
	RASPR *next;
};

#define STR struct str1123
STR
{
	char name[MAX_INDENT_LEN];
	BOOL see_now;
	BOOL without_config_information;
	BOOL insert_debug_info;
	FILES *where;
	STR *next;
};

/*    -       */
#define HIERARHY struct hierarhy
HIERARHY
{
	char mainname[MAX_INDENT_LEN];
	STR *scal;
	FILES *where;
	RASPR *raspr;
	HIERARHY *next;
};

#define CNFFILE struct cnffile
CNFFILE
{
	FILE* file;
	char filename[MYMAXPATH];
	long line;
};

/**********************************************************/

/*   */
extern BOOL bPrintDetail;

/*      INCLUDE*/
extern BOOL bNoInclude;

/*,   ,     */
extern BOOL bUserInterface;

/*  */
extern BOOL bDebug;

/*    */
extern BOOL bDebugFortran;

/*  */
extern OUT_FILE_TYPE out_file_type;

/*   */
extern BOOL bNormaGeneratedProgram;

/* */
extern char **message;

/*   */
extern char strCurrentPart[MAX_INDENT_LEN];
/*     */
extern char strOutPath[MYMAXPATH];
extern int nLenOutPath;
/*    */
extern char main_out_name[64];
/*   */
extern BOOL bUppercaseExt;
/* ,        */
extern BOOL bUserOutName;

/*  */
extern FILES *src_file;
/* header-*/
extern FILES *head_file;
/*  */
extern RASPR *begraspr;
/* */
extern HIERARHY begin_hi;

/**********************************************************/

void OutOfMemory(void);
void ErrWrite(void);
void CnfInfDestroyed(STRINGS *s);
void CnfExitSay(CNFFILE *f, STRINGS *s, int MssgCode, ...);
void CnfError(STRINGS *s, int MssgCode, ...);
void CnfWarning(STRINGS *s, int MssgCode, ...);
void CnfNfErr(STRINGS *s, int MssgCode, ...);

void CnfParamError(CALLPART *fcall, CALLPART *acall, char* partname, int MssgCode, int n_set, int n_param);
void CnfParamWarning(CALLPART *fcall, CALLPART *acall, char* partname, int MssgCode, int n_set, int n_param);
void CnfParamNfErr(CALLPART *fcall, CALLPART *acall, char* partname, int MssgCode, int n_set, int n_param);

/*  (    )*/
BOOL CompareNames(const char *name1, const char *name2);
/*     .
 0-
			  1-rus
			  2-eng*/
int IsCnfString(const char *s, int n);
char *GetDistrName(char *s, int j);
/*       s*/
BOOL GetFileLine(STRINGS *s, char** name, long *line);
/*    s*/
long GetStringLine(STRINGS *s);

FILE *OpenTmpFile(char *start_path, char *mode);

/*    */
CNFFILE *cnf_open(const char *filename, const char *mode);
char *cnf_gets(char *s, int n, CNFFILE *f);
int cnf_close(CNFFILE *stream);

/*     header-  NULL
        */
STRINGS* ReadHeaderFile(char*filename);

char* GetNextInfStr(STRINGS **nov);

#endif
