/*    .
   
  ..*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32

#define unlink _unlink

#else

#include <sys/types.h>
#include <unistd.h>
#include <limits.h>

#endif


#include "normacnf.h"
#include "memcheck.h"
#include "paramcnf.h"
#include "outptcnf.h"

/***********************************************************/

/*      */
static void ChopBuffer(char* buf, FILE *fout)
{
	char c;
	int n, m;
	BOOL string = FALSE;
	int len = strlen(buf);

	if(len<72)
		return;


	for(n=len-1; n>30; n--)
	{
		if(buf[n]=='\'')
			string = !string;
		else if(n<70 && !string && (buf[n]==' ' || buf[n]==','))
			break;
	}
	c = buf[n+1];
	buf[n+1]=0;

	if(fprintf(fout, buf)==EOF) ErrWrite();
	if(fprintf(fout, "\n")==EOF) ErrWrite();

	strcpy(buf, "     >");
	buf[6] = c;
	for(m=7; ; m++,n++)
	{
		buf[m]=buf[n+2];
		if(buf[n+2] == 0)
			break;
	}
}

/***********************************************************/

/*   */
static void PrintCommonParam(PARAM *prm, FILE *f1, int nNewNamesCount)
{
	char buf[MAX_STR_LEN];

	DIAPAZON *dprm;
	BOOL f = FALSE;

	if(prm->type == TYP_CHAR)
		sprintf(buf, "      %s*%ld ", GetPrintedType(prm->type), prm->strlen);
	else
		sprintf(buf, "      %s ", GetPrintedType(prm->type));

	if(prm->bCommon)
		sprintf(&buf[strlen(buf)], "%s", prm->name);
	else
		sprintf(&buf[strlen(buf)], "RCNF%d", nNewNamesCount);

	ChopBuffer(buf, f1);

	if(prm->variety == VRT_SCALAR || prm->variety == VRT_FUNCTION)
	{
		fprintf(f1, buf);
		fprintf(f1, "\n");
		return;
	}

	sprintf(&buf[strlen(buf)], "(");
	for(dprm=prm->diapazon; dprm; dprm=dprm->next)
	{
		if(f) sprintf(&buf[strlen(buf)], ", ");
		else f=TRUE;

		if(dprm->first != 1)
			sprintf(&buf[strlen(buf)], "%ld:", dprm->first);
		sprintf(&buf[strlen(buf)], "%ld", dprm->last);

		ChopBuffer(buf, f1);
	}

	fprintf(f1, buf);
	fprintf(f1, ")\n");
}

/*  */
static void WriteCommonBlocks(FILE *f1)
{
	char buf[MAX_STR_LEN];
	PARAM *pNextVar;
	COMMON* pCommon = NULL;

	int nNewNamesCount=0;

	while(TRUE)
	{
		int nNewCommonNamesCount=0;

		pCommon = GetNextCommon(pCommon);
		if(pCommon == NULL)
			break;
		/*     */
		if(pCommon->bMainPart)
			continue;
		/*      ,     ? */
		for(pNextVar = pCommon->parameters; pNextVar != NULL; pNextVar = pNextVar->next)
		{
			if(pNextVar->bCommon)
				break;
		}
		if(pNextVar == NULL)
			continue;

		/* */
		for(pNextVar = pCommon->parameters; pNextVar != NULL; pNextVar = pNextVar->next)
		{
			PrintCommonParam(pNextVar, f1, nNewNamesCount + nNewCommonNamesCount);
			if(!pNextVar->bCommon)
				nNewCommonNamesCount++;
		}

		sprintf(buf, "      COMMON /%s/", pCommon->name);

		ChopBuffer(buf, f1);

		nNewCommonNamesCount=0;

		/* */
		for(pNextVar = pCommon->parameters; pNextVar != NULL; pNextVar = pNextVar->next)
		{
			if(pNextVar->bCommon)
			{
				sprintf(&buf[strlen(buf)], " %s", pNextVar->name);
			}
			else
			{
				sprintf(&buf[strlen(buf)], " RCNF%d", nNewNamesCount);
				nNewNamesCount++;
			}

			if(pNextVar->next)
				sprintf(&buf[strlen(buf)], ",");

			ChopBuffer(buf, f1);
		}
		fprintf(f1, buf);
		fprintf(f1, "\n");

		nNewNamesCount += nNewCommonNamesCount;
	}
}

static void CallRegVar(PARAM* var, FILE *f1, BOOL bCommon);

/*  */
static void WriteRegCommonBlocks(FILE *f1)
{
	PARAM *pNextVar;
	COMMON* pCommon = NULL;

	while(TRUE)
	{
		pCommon = GetNextCommon(pCommon);
		if(pCommon == NULL)
			break;
		if(pCommon->parameters == NULL)
			continue;

		for(pNextVar = pCommon->parameters; pNextVar != NULL; pNextVar = pNextVar->next)
			CallRegVar(pNextVar, f1, TRUE);
	}
}

/***********************************************************/

/* */
static void CheckSpecialComment(const char *s, FILE *f1)
{
	if(MPI && strcmp(s, "C $DIR CASE MPIFLG\n")==0)
	{
		int i=1;
		RASPR *endraspr;
		for(endraspr=begraspr; endraspr; endraspr=endraspr->next,i++)
		{
			if(C_OUT) fprintf(f1,"*/\n\t\tif(MPIFLG==%d) %s();\n/*\n",i,endraspr->name);
			else fprintf(f1,"      IF(MPIFLG.EQ.%d) CALL %s\n",i,endraspr->name);
		}
	}
	else if(MPI && strncmp(s, "C $DIR MPIFLG= ", 15)==0)
	{
		int i=1;
		RASPR *endraspr;
		for(endraspr=begraspr; endraspr; endraspr=endraspr->next,i++)
		{
			if(CompareNames(&s[15], endraspr->name))
				break;
		}
		if(endraspr==NULL) CnfExitSay(NULL, NULL, 31, s);
		if(C_OUT) fprintf(f1,"*/\n\tMPIFLG=%d;\n/*\n",i);
		else fprintf(f1,"      MPIFLG=%d\n",i);
	}
	else if(strncmp(s, "C $DIR OUTPUT", 13)==0)
	{
		printf(&s[13]);
	}
}

/*       ?*/
static BOOL IsScalarSaved(HIERARHY *t_hi, STR *t_scal)
{
	STR *next_scal;
	HIERARHY *tek_hi=t_hi->next;
	while(tek_hi)
	{
		next_scal=tek_hi->scal;
		while(next_scal)
		{
			if(CompareNames(t_scal->name, next_scal->name))
				return TRUE;
			next_scal=next_scal->next;
		}
		tek_hi=tek_hi->next;
	}
	return FALSE;
}

/***********************************************************/
/*Static variable for next fuctions*/
static STRINGS *strStatic;

STRINGS *GetStrStatic(void) { return strStatic; }

char* SetCurrentString(STRINGS *srcstr)
{
	strStatic = srcstr;
	return GetCurrentString();
}

char* GetCurrentString(void)
{
	return strStatic->s;
}

char* GetNextString(BOOL IsInCurrentFile)
{
	if(strStatic->next==NULL) return NULL;
	if(IsInCurrentFile)/*   */
	{
		FILES *tek_file;
		if(strStatic->next->num<0) /*header*/
		{
			for(tek_file=head_file;tek_file!=NULL;tek_file=tek_file->next)
			{
				if(tek_file->sFirstStr == NULL)
					continue;
				if(tek_file->sFirstStr->num == strStatic->next->num) return NULL;
				if(tek_file->sFirstStr->num < strStatic->next->num) break;
			}
		}
		else /*source*/
		{
			for(tek_file=src_file;tek_file!=NULL;tek_file=tek_file->next)
			{
				if(tek_file->sFirstStr == NULL)
					continue;
				if(tek_file->sFirstStr->num == strStatic->next->num) return NULL;
				if(tek_file->sFirstStr->num > strStatic->next->num) break;
			}
		}
	}
	strStatic=strStatic->next;
	return strStatic->s;
}

char* GetNextStringInf()
{
	char* str = GetNextString(TRUE);
	if(str == NULL) CnfInfDestroyed(GetStrStatic());
	return str;
}

char* OpenSourceFile(FILES *pFile)
{
	strStatic = pFile->sFirstStr;
	if(strStatic)
		return strStatic->s;
	return NULL;
}

void CloseSourceFile(void)
{
	strStatic = NULL;
}

/***********************************************************/
static char *StrOutStrings[]={
"      CALL %s",
" CALL %s",
"(\'%s\\000\', \'%s\\000\', %ld, %ld)\n",
"CDEC$ ATTRIBUTES C, REFERENCE :: %s,%s,%s\n",
"ndstexf",
"ndenexf",
"ndregvex",
"ndopstex",
"ndstpex",
"ndenprg",
"ndregvg",
"CDEC$ ATTRIBUTES C, REFERENCE :: %s,%s,%s,%s\n",
};

/*  -    */
static void CallRegVar(PARAM* var, FILE *f1, BOOL bCommon)
{
	int ranges=0;
	DIAPAZON *diapazon;
	char buf[MAX_STR_LEN];

	if(var->variety == VRT_FUNCTION)
		return;

	/*  ,     */
	if(var->variety==VRT_POINTER)
		return;

	if(!(bCommon && var->bCommon) && !(!bCommon && !var->bCommon))
		return;

	sprintf(buf, StrOutStrings[0], StrOutStrings[bCommon ? 10 : 6]);
	/*  */
	sprintf(&buf[strlen(buf)], "(\'%s\\000\',%s,",var->name,var->name);
	/*  - */
	for(diapazon = var->diapazon; diapazon!=NULL; diapazon=diapazon->next)
		ranges++;

	sprintf(&buf[strlen(buf)],"%d,%d", GetDebugerType(var), ranges);

	/*  ,   */
	ChopBuffer(buf, f1);

	for(diapazon = var->diapazon; diapazon!=NULL; diapazon=diapazon->next)
	{
		if(diapazon->fExpression)
			sprintf(&buf[strlen(buf)], ", %s,", diapazon->fExpression);
		else
			sprintf(&buf[strlen(buf)], ", %ld,", diapazon->first);

		if(diapazon->lExpression)
			sprintf(&buf[strlen(buf)], diapazon->lExpression);
		else
			sprintf(&buf[strlen(buf)], "%ld", diapazon->last);

		/*  ,   */
		ChopBuffer(buf, f1);
	}

	if(fprintf(f1, buf)==EOF) ErrWrite();
	if(fprintf(f1, ")\n")==EOF) ErrWrite();
}

/*    string    label*/
static void ExtractLabel(char* string, char* label)
{
	int n;
	for(n=0; n<5; n++)
	{
		if(string[n]=='\t' || string[n]==0 || string[n]=='\n')
		{
			while(n<5)
			{
				label[n] = ' ';
				n++;
			}
			break;
		}
		label[n] = string[n];
		string[n] = ' ';
	}
	label[n]=0;
}

/*   */
static BOOL WriteScalar(FILE *f1, BOOL insert_debug_info, char* partname, FILES *srcFile)
{
	char buf[MAX_STR_LEN];
	long srcline, lPrevLineFirst=0, lPrevLineLast=0;
	CALLPART* part = NULL;
	PARAM* var;
	char filename[MYMAXPATH];
	int prevoper=0;
	long lStartDebugText=0;
	BOOL bPartDecl = FALSE;
	char* string = GetCurrentString();

	if(insert_debug_info)
		part = PreparePartInfo(partname);

	if(part)
	{/*     */
		int pos = strlen(part->filename);
		for(; pos>0; pos--)
			if(part->filename[pos-1] == '\\' || part->filename[pos-1] == '/')
				break;
		strcpy(filename, &part->filename[pos]);
	}

	/*  */
	srcline = GetStringLine(GetStrStatic());

	/* ,   */
	if(part)
	{
		if(part->type == TYP_PROGRAMNAME)
			lStartDebugText = part->mpi_initline;
		else
			lStartDebugText = part->meaningline;
	}

	while(string)
	{
		/*  */
		if(part && srcline==part->intrinsicline)
		{
			if(fprintf(f1, StrOutStrings[11], StrOutStrings[4],
				StrOutStrings[5], StrOutStrings[6], StrOutStrings[7])==EOF) ErrWrite();
			if(fprintf(f1, StrOutStrings[3], StrOutStrings[8],
				StrOutStrings[9], StrOutStrings[10])==EOF) ErrWrite();

			if(part->type == TYP_PROGRAMNAME)
				WriteCommonBlocks(f1);
		}
		if(part && srcline >= lStartDebugText && srcline >lPrevLineLast)
		{
			long linescount;
			char label[10];
			int oper = IsStartOfOperator(string, &linescount);
			/*    */
			if(oper>0 && !bPartDecl && srcline>=lStartDebugText)
			{
				bPartDecl = TRUE;

				if(part->type == TYP_PROGRAMNAME)
					sprintf(buf, StrOutStrings[0], StrOutStrings[8]);
				else
					sprintf(buf, StrOutStrings[0], StrOutStrings[4]);
				sprintf(&buf[strlen(buf)],
					StrOutStrings[2], part->name, filename, part->line, srcFile->crc);
				ChopBuffer(buf, f1);
				if(fprintf(f1, buf)==EOF) ErrWrite();

				if(part->type == TYP_PROGRAMNAME)
					WriteRegCommonBlocks(f1);

				for(var=part->parameters; var!=NULL; var = var->next)
					CallRegVar(var, f1, FALSE);
			}

			/*   ,      */
			if(oper > 0)
				ExtractLabel(string, label);
			if(oper == 1 || oper == 4)
			{
				if(fprintf(f1, label)==EOF) ErrWrite();
				if(fprintf(f1, StrOutStrings[1], StrOutStrings[7])==EOF) ErrWrite();
				if(fprintf(f1, "(%d, %ld, %ld, %ld, %ld)\n", (oper == 4)?1:0,
					lPrevLineFirst, lPrevLineLast, srcline, srcline+linescount-1)==EOF) ErrWrite();
				lPrevLineFirst = srcline;
				lPrevLineLast = srcline+linescount-1;
			}
			else if(oper == 2 || (oper == 3 && prevoper != 2))
			{
				int fncname;
				if(fprintf(f1, label)==EOF) ErrWrite();
				if(part->type == TYP_PROGRAMNAME) fncname=9;
				else fncname=5;
				if(fprintf(f1, StrOutStrings[1], StrOutStrings[fncname])==EOF) ErrWrite();
				if(fprintf(f1, "(\'%s\\000\', %ld)\n", filename, srcline)==EOF) ErrWrite();
			}
			if(oper>0)
				prevoper=oper;
		}

		if(fputs(string, f1)==EOF) ErrWrite();
		CheckSpecialComment(string, f1);
		if(IsEndPart(string)) break;

		string = GetNextString(TRUE);
		srcline++;
	}

	if(string == NULL) return FALSE;
	return TRUE;
}

/*  ,       */
static BOOL WriteFileWithoutConfig(FILE *f1, FILES* srcFile)
{
	char partname[MAX_INDENT_LEN];
	char* string = GetCurrentString();
	partname[0]=0;

	srcFile->was_all_write=TRUE;

	while(string)
	{
		if(bDebugFortran && !bNormaGeneratedProgram && IsThisProgram(string, partname))
		{
			if(!WriteScalar(f1, TRUE, partname, srcFile)) return FALSE;
			partname[0]=0;
		}
		else if(bDebugFortran && IsThisFuncOrSub(string, partname))
		{
			if(!WriteScalar(f1, TRUE, partname, srcFile)) return FALSE;
			partname[0]=0;
		}
		else
		{
			if(fputs(string, f1)==EOF) ErrWrite();
		}
		string = GetNextString(TRUE);
	}

	return TRUE;
}

/***********************************************************/
/*     BLOCK DATA*/

static BOOL IsThisBlockData(char *s, char *name)
{
	int i=0, j=0;

	if(strchr("Cc!*\n#", s[0])) return FALSE;
	while(s[i]==' ' || s[i]=='\t') i++;

	if(!IsFortranKeyword(&s[i], 17))
		return FALSE;

	i+=10;
	while(s[i]==' ' || s[i]=='\t') i++;

	/* */
	j=0;
	while(strchr("( \n\t", s[i+j])==NULL && j+1<MAX_INDENT_LEN) j++;
	strncpy(name, &s[i], j);
	name[j]=0;
	return TRUE;
}

static void SaveBlockData(FILE *f1)
{
	char *buf_source;
	char block_name[MAX_INDENT_LEN];
	FILES *pFiles;

	if(C_OUT) return;

	for(pFiles = src_file; pFiles != NULL; pFiles = pFiles->next)
	{
		STRINGS *StrStart=NULL;

		BOOL IsBlockData = FALSE;
		/* normalib    */
		if(!pFiles->is_user_file || pFiles->was_all_write) continue;
		/* ,     
		if(!pFiles->was_write) continue;*/

		for(buf_source = OpenSourceFile(pFiles)
			;buf_source; buf_source = GetNextString(TRUE))
		{
			if(!IsBlockData)
			{
				IsBlockData = IsThisBlockData(buf_source, block_name);
				if(!IsBlockData) continue;
				StrStart = GetStrStatic();
				putc('\n', f1);
			}

			if(fputs(buf_source, f1)==EOF) ErrWrite();
			CheckSpecialComment(buf_source, f1);
			IsBlockData = !IsEndPart(buf_source);
		}

		/*   */
		if(IsBlockData)
			CnfExitSay(NULL, StrStart, 50, block_name);

		CloseSourceFile();
	}
}

/***********************************************************/

/*  ,   */
static BOOL ReplaceOldFile(char* oldfile, char*newfile)
{
	FILE *fOld, *fNew;
	int i, j;

	fOld=myfopen(oldfile,"rt");
	fNew=myfopen(newfile,"rt");

	/*  */
	do
	{
		i=getc(fOld);
		j=getc(fNew);
	}
	while(i!=EOF && j!=EOF && i==j);

	myfclose(fOld);
	myfclose(fNew);

	/*     ( ) -  */
	if(i!=j || bUserInterface)
	{
		unlink(oldfile);
		rename(newfile, oldfile);
	}
	else
	{
		unlink(newfile);
	}

	return (i!=j);
}

/***********************************************************/

/*       */
static void SaveOutScalar(FILE *f1, HIERARHY *rasprpart, STR *scal)
{
	if(scal == NULL) return;
/*    */
	SaveOutScalar(f1, rasprpart, scal->next);

	{
		char *buf_source;
		strcpy(strCurrentPart, scal->name);
		/*     */
		if(scal->where->was_all_write) return;
		/*    MPI       */
		if((FMPI_OR_F77||C_OUT) && IsScalarSaved(rasprpart, scal)) return;

		if(!scal->where->was_write)
		{
			printf("%s\n", scal->where->name);
			scal->where->was_write=TRUE;
		}
		putc('\n', f1);

		buf_source = OpenSourceFile(scal->where);
		/*  ,       */
		if(scal->where->exists_without_config)
		{
			if(!WriteFileWithoutConfig(f1, scal->where))
				CnfExitSay(NULL, GetStrStatic(), 50, scal->name);
		}
		else
		{
			/*  */
			for(;buf_source; buf_source = GetNextString(FALSE))
			{
				int i = IsCnfString(buf_source, 1);
				if(i==0) continue;
				buf_source = GetNextStringInf();
				if(!CompareNames(&buf_source[12+2*(i==1)], scal->name)) continue;

				while(IsCnfString(buf_source, 2)==0)
					buf_source = GetNextStringInf();

				GetNextStringInf();
				if(C_OUT)
					GetNextStringInf();
				break;
			}

			/*   */
			if(!WriteScalar(f1, (BOOL)(scal->insert_debug_info && out_file_type==FMPI), scal->name, scal->where))
				CnfExitSay(NULL, GetStrStatic(), 50, scal->name);
		}
		CloseSourceFile();
	}
}

/***********************************************************/

/*    */
static void SaveNormaOutFile(HIERARHY *rasprpart)
{
	static char out_filename[MYMAXPATH], tmp_name[MYMAXPATH];
	static FILE *f_old, *f_out;

	/*    */
	if(rasprpart->next != NULL)
		SaveNormaOutFile(rasprpart->next);

	{
		STRINGS *nov=NULL, *mains=NULL;
		BOOL bIsMainPart=(rasprpart==&begin_hi);
		BOOL bCommonWriten = FALSE, bCommonRegWriten = FALSE;
		BOOL bSlave;

		if((!FMPI_OR_F77&&!C_OUT)||rasprpart->next == NULL)
		{
			FILES* next_file;

			strcpy(out_filename, strOutPath);
			if(FMPI_OR_F77||C_OUT)
				strcpy(&out_filename[nLenOutPath], main_out_name);
			else
				strcpy(&out_filename[nLenOutPath], rasprpart->mainname);

			switch(out_file_type)
			{
			case GNS:
				if(bUppercaseExt)
					strcat(out_filename,".gns");
				else
					strcat(out_filename,".GNS");
				break;
			case C:
			case CMPI:
				if(bUppercaseExt)
					strcat(out_filename,".C");
				else
					strcat(out_filename,".c");
				break;
			default:
				if(bUppercaseExt)
					strcat(out_filename,".F");
				else
					strcat(out_filename,".f");
				break;
			}
			if((f_old=myfopen(out_filename,"rt"))==NULL)
			{
				if((f_out=fopen(out_filename,"wt")) == NULL) ErrWrite();
			}
			else
			{
				strOutPath[nLenOutPath]=0;
				f_out=OpenTmpFile(strOutPath, "wt");
				strcpy(tmp_name, strOutPath);
			}

			/*        */
			if(C_OUT&&fputs("#include <stdlib.h>\n#include <stdio.h>\n",f_out)==EOF) ErrWrite();
			if(C_OUT&&fputs("#include <stdarg.h>\n#include <math.h>\n",f_out)==EOF) ErrWrite();
			if(out_file_type==CMPI)
			{
				int ITASKJ=1;
				if(begraspr)
					ITASKJ = begraspr->i*begraspr->j;

				if(bDebug && fputs("#include \"MonitorConnect.h\"\n",f_out)==EOF) ErrWrite();
				if(fputs("#include <mpi.h>\n\n",f_out)==EOF) ErrWrite();
				if(fprintf(f_out,"int ITASKJ[%d];\n",ITASKJ)==EOF) ErrWrite();
				if(fputs("int IPRWWW, JPRWWW;\n",f_out)==EOF) ErrWrite();
				if(fputs("int IVVVVJ, IWWWWJ;\n",f_out)==EOF) ErrWrite();
			}

			/*   */
			for(next_file=src_file; next_file; next_file=next_file->next)
			{
				next_file->was_write = FALSE;
				next_file->was_all_write = FALSE;
			}
			printf(message[16], out_filename);
		}
		if(!rasprpart->where->was_write)
		{
			printf("%s\n", rasprpart->where->name);
			rasprpart->where->was_write = TRUE;
		}

		SaveOutScalar(f_out, rasprpart, rasprpart->scal);

		strcpy(strCurrentPart, rasprpart->mainname);

		nov=src_file->sFirstStr;
		while(TRUE)
		{
			if(bIsMainPart)
			{
				int j=IsCnfString(nov->s, 1);
				/*     */
				if(j!=0 && nov->next!=NULL && nov->next->next!=NULL &&
					CompareNames(&nov->next->next->s[12+2*(j==1)], rasprpart->mainname))
				{
					while(IsCnfString(nov->s, 2)==0)
						GetNextInfStr(&nov);

					GetNextInfStr(&nov);
					/*    ,  -*/
					if(out_file_type==CMPI)
					{
						mains = nov;
						while(!IsEndPart(nov->s))
						{
							if(nov->next==NULL) CnfExitSay(NULL, nov, 50, rasprpart->mainname);
							nov=nov->next;
						}
					}
					if(C_OUT) GetNextInfStr(&nov);
					break;
				}
			}
			else
			{
				/*  DISTRNAME*/
				int j=IsCnfString(nov->s, 3);
				if(j!=0 && CompareNames(GetDistrName(nov->s, j), rasprpart->mainname))
				{
				/*   */
					while(TRUE)
					{
						if(nov->next==NULL) CnfExitSay(NULL, nov, 24, rasprpart->mainname);
						nov=nov->next;
						if(FMPI_OR_F77||C_OUT)
						{
							if(IsThisFuncOrSub(nov->s, rasprpart->mainname)) break;
						}
						else
						{
							int i;
							if((nov->s[0]|32)=='c') continue;
							for(i=0; nov->s[i]; i++)
							{
								if(nov->s[i]==' ') continue;
								if(strnicmp(&(nov->s[i]),"TASK ",5)==0)
								{
									i+=4;
									continue;
								}
								break;
							}
							if(strnicmp(&(nov->s[i]),"PROGRAM",7)!=0)
								continue;
						}
					}
					break;
				}
			}
			GetNextInfStr(&nov);
		}

		putc('\n', f_out);
		/*    ,  bSlave == FALSE*/
		bSlave = (!bIsMainPart || !MPI);
		/*    DISRIBUTED   */
		while(TRUE)
		{
			if(fputs(nov->s,f_out)==EOF) ErrWrite();
			if(IsEndPart(nov->s))
			{
				if(bSlave) break;
				bSlave=TRUE;
				putc('\n', f_out);
				if(out_file_type==CMPI) nov = mains;
			}

			/*    COMMON*/
			if(bIsMainPart && bDebugFortran && !bCommonRegWriten)
			{
				if(strnicmp("C $DIR GLOBAL VARIABLES", nov->s, 23)==0)
				{
					WriteRegCommonBlocks(f_out);
					bCommonRegWriten = TRUE;
				}
			}

			CheckSpecialComment(nov->s, f_out);

			if(nov->next==NULL) CnfExitSay(NULL, nov, 50, rasprpart->mainname);
			nov=nov->next;

			/*  COMMON*/
			if(bIsMainPart && bDebugFortran && !bCommonWriten)
			{
				while(!IsNewFortranOperator(nov->s))
				{
					if(fputs(nov->s,f_out)==EOF) ErrWrite();
					CheckSpecialComment(nov->s, f_out);
					if(nov->next==NULL) CnfExitSay(NULL, nov, 50, rasprpart->mainname);
					nov=nov->next;
				}
				WriteCommonBlocks(f_out);
				bCommonWriten = TRUE;
			}
		}

		/*  */
		if((!FMPI_OR_F77&&!C_OUT)||bIsMainPart)
		{
			SaveBlockData(f_out);

			fclose(f_out);

			if(f_old!=NULL)
			{
				myfclose(f_old);
				if(ReplaceOldFile(out_filename, tmp_name))
					printf(message[17]);
				else
					printf(message[18]);
			}
			else
			{
				printf(message[17]);
			}

		}
	}
}

/***********************************************************/
/*    */

/*   -   COMMON ( )*/
static void PrepareCommonBlocks(void)
{
	FILES* next_file;
	char partname[MAX_INDENT_LEN]="";

	if(!bDebugFortran)
		return;

	for(next_file=src_file; next_file; next_file=next_file->next)
	{
		char* string = OpenSourceFile(next_file);

		while(string)
		{
			if(IsThisFuncOrSub(string, partname))
			{
				PreparePartInfo(partname);
				partname[0]=0;

				while(string)
				{
					if(IsEndPart(string)) break;
					string = GetNextString(TRUE);
				}
			}
			string = GetNextString(TRUE);
		}

	}
}

/*     -*/
static void SaveFortranOutFile(void)
{
	FILES* next_file;
	FILE *fold, *f_out;
	int i;
	char out_filename[MYMAXPATH];
	char tmp_name[MYMAXPATH];

	/*    -   COMMON
	       */
	PrepareCommonBlocks();

	/*  */
	strcpy(out_filename, strOutPath);
	for(i=nLenOutPath; out_filename[i]; i++)
	{
		if(out_filename[i]=='.')
		{
			out_filename[i]='\0';
			break;
		}
	}

	/* _dbg,          */
	if(!bUserOutName)
		strcat(out_filename,"_dbg");

	if(bUppercaseExt)
		strcat(out_filename,".F");
	else
		strcat(out_filename,".f");

	fold=myfopen(out_filename,"rt");
	if(fold==NULL)
	{
		if((f_out=fopen(out_filename,"wt")) == NULL) ErrWrite();
	}
	else
	{
		strOutPath[nLenOutPath]=0;
		f_out=OpenTmpFile(strOutPath, "wt");
		strcpy(tmp_name, strOutPath);
	}

	printf(message[16], out_filename);

	/*  */
	for(next_file=src_file; next_file; next_file=next_file->next)
	{
		printf("%s\n", next_file->name);
		if(fold!=NULL && strcmp(&out_filename[nLenOutPath], next_file->name)==0)
			CnfExitSay(NULL, NULL, 58);

		OpenSourceFile(next_file);
		if(!WriteFileWithoutConfig(f_out, next_file))
			CnfExitSay(NULL, GetStrStatic(), 50, "");

		putc('\n', f_out);
	}

	fclose(f_out);

	if(fold!=NULL)
	{
		myfclose(fold);
		if(ReplaceOldFile(out_filename, tmp_name))
			printf(message[17]);
		else
			printf(message[18]);
	}
	else
	{
		printf(message[17]);
	}
}

/***********************************************************/

void WriteOutFile(void)
{
	if(bNormaGeneratedProgram)
		SaveNormaOutFile(&begin_hi);
	else
		SaveFortranOutFile();

	printf(message[19]);
}

/***********************************************************/
