/* Header of TCP/IP communicating library */
#ifndef _CDMessages_H
#define _CDMessages_H
/* Messages */
#define CD_FAIL 2000
#define CD_OK 2001

#define CD_START_PARAMETERS 2002
/* [-wp] [-src NN] [-srct NN]  "task" [level]
:

task  -     ;

level -   .

    ,           .  level=0,     .

  wp , -src NN  -srct NN     .

NN
       (Comm)       (Communication),         
 
-wp
      (Execution Characteristics)       SRC Characteristics; 
 
-src NN
   SRC Characteristics,      SRC (omm); 
 
-srct NN
   SRC Characteristics,    SRC     
 */

#define CD_ERROR 2003
/* free text */

#define CD_ANALYSIS_READY 2004
/* %d %ld %s */
/* lines amount, CRC, protocol_name */

#define CD_GET_PROTOCOL_PARAMETERS 2030
/* "%s" */
/* protocol_name */

#define CD_TAKE_PROTOCOL_PARAMETERS 2031
/* %d %ld %s */
/* lines amount (-1 if file isn't available), CRC, protocol_name */


#define SOURCE_FILE_JUMP			0x1
#define TRACE_EVENT_JUMP			0x2
#define PROTOCOL_BY_TIME_JUMP		0x4
#define PROTOCOL_BY_LOCATION_JUMP	0x8
/*
    Jumpable information: unsigned int, where bits show if jump available or not:
    Bits:	0 (SOURCE_FILE_JUMP)			- into Source File text
			1 (TRACE_EVENT_JUMP)			- into Trace Events view
			2 (PROTOCOL_BY_TIME_JUMP)		- into Protocol text sorted by time
			3 (PROTOCOL_BY_LOCATION_JUMP)	- into Protocol text sorted by location

    Example: 5 (0101B, SOURCE_FILE_JUMP | PROTOCOL_BY_TIME_JUMP) means that jumps
	         in Source File text and Protocol text sorted by time are avalable.

    Jump type - unsigned int with only 1 bit set as above.
	Example - 2 (TRACE_EVENT_JUMP) means jump into Trace Events view.
 */

#define CD_GET_PROTOCOL_TEXT 2005
/* "%s" %d %d */
/* protocol_name, start_line, lines amount */

#define CD_TAKE_PROTOCOL_TEXT 2006
/* "%s" %d %d %d ... %d %s */
/* protocol_name, start_line, lines amount(n), jumpable_info_1 ... jumpable_info_n,
   free text (n lines, line's separator - '\n') */
/* On error: */
/* "%s" %d -1 */

#define CD_GET_PROTOCOL_JUMP 2028
/* "%s" %d %d */
/* protocol_name, line, jump type */

#define CD_TAKE_PROTOCOL_JUMP 2029
/* jump type = SOURCE_FILE_JUMP */
  /* "%s" %d %d "%s" %d */
  /* protocol_name, line, jump type, source_file_name, source_line_number */
/* jump type = TRACE_EVENT_JUMP */
  /* "%s" %d %d "%s" %d ... %d .... "%s" %d ... %d */
  /* protocol_name, line, jump type, trace_name1, event1_1 ... event1_n ... trace_namek, eventk_1 ... eventk_m */


#define CD_STOP 2007

#define CD_GET_TASKS_AVAILABLE 2008

#define CD_TAKE_TASKS_AVAILABLE 2009
/* %d "%s" ..."%s" */
/* amount task1 ... taskN */

#define CD_GET_TRACE_AVAILABLE 2010

#define CD_TAKE_TRACE_AVAILABLE 2011
/* %d "%s" ..."%s" */
/* amount trace_name1 ... trace_nameN */

#define CD_GET_TRACE_PARAMETERS 2012
/* "%s" */
/* trace_name */

#define CD_TAKE_TRACE_PARAMETERS 2013
/* %d %ld %s */
/* events amount, CRC, trace_name */

#define CD_GET_TRACE_EVENT 2014
/* "%s" %d %d */
/* trace_name, start_event, events amount */

#define CD_TAKE_TRACE_EVENT 2015
/* "%s" %d %d %d ... %d %s */
/* trace_name, start_event, events amount (n), jumpable_info_1 ... jumpable_info_n,
   free text (n lines, line's separator - '\n') */
/* On error: */
/* "%s" %d -1 */

#define CD_GET_TRACE_EVENT_JUMPABLE 2032
/* "%s" %d %d */
/* trace_name, start_event, events amount */

#define CD_TAKE_TRACE_EVENT_JUMPABLE 2033
/* "%s" %d %d %d ... %d */
/* trace_name, start_event, events amount (n), jumpable_info_1 ... jumpable_info_n */
/* On error: */
/* "%s" %d -1 */

#define CD_GET_TRACE_EVENT_TEXT 2024
/* "%s" %d */
/* trace_name, event number */

#define CD_TAKE_TRACE_EVENT_TEXT 2025
/* "%s" %d %d %s */
/* trace_name, event number, lines amount (n),
   free text (n lines, line's separator - '\n') */
/* On error: */
/* "%s" %d -1 */


#define CD_GET_TRACE_EVENT_JUMP 2026
/* "%s" %d %d */
/* trace_name, event number, jump type */

#define CD_TAKE_TRACE_EVENT_JUMP 2027
/* "%s" %d %d "%s" %d */
/* jump type = SOURCE_FILE_JUMP */
  /* trace_name, event number, jump type, source_file_name, source_line_number */
/* jump type = PROTOCOL_BY_TIME_JUMP | PROTOCOL_BY_LOCATION_JUMP */
  /* trace_name, event number, jump type, protocol_name, protocol_line_number */

#define CD_GET_SOURCE_FILE_PARAMETERS 2016
/* "%s" */
/* source_file_name */

#define CD_TAKE_SOURCE_FILE_PARAMETERS 2017
/* %d %ld %s */
/* lines amount (-1 if file isn't available), CRC, source_file_name */

#define CD_GET_SOURCE_FILE_TEXT 2018
/* "%s" %d %d */
/* source_file_name, start_line, lines amount */

#define CD_TAKE_SOURCE_FILE_TEXT 2019
/* "%s" %d %d %d ... %d %s */
/* source_file_name, start_line, lines amount (n), jumpable_info_1 ... jumpable_info_n,
   free text; line's separator - '\n' */

#define CD_GET_SOURCE_FILE_JUMP 2020
/* "%s" %d %d */
/* source_file_name, line, jump type */

#define CD_TAKE_SOURCE_FILE_JUMP 2021
/* jump type = TRACE_EVENT_JUMP */
  /* "%s" %d %d "%s" %d ... %d .... "%s" %d ... %d */
  /* source_file_name, line, jump type, trace_name1, event1_1 ... event1_n ... trace_namek, eventk_1 ... eventk_m */
/* jump type = PROTOCOL_BY_TIME_JUMP | PROTOCOL_BY_LOCATION_JUMP */
  /* "%s" %d %d "%s" %d ... %d */
  /* source_file_name, line, jump type, protocol_name, line1 ... line1 */

#define CD_GET_SOURCE_FILE_JUMPABLE 2022
/* "%s" %d %d */
/* source_file_name, start_line, lines amount */

#define CD_TAKE_SOURCE_FILE_JUMPABLE 2023
/* "%s" %d %d %d ... %d */
/* source_file_name, start_line, lines amount (n), jumpable_info_1 ... jumpable_info_n */

#define CD_INFO_SOURCE_FILE_JUMPABLE_CHANGED 2034
/* "%s" %d 1 %d */
/* source_file_name, start_line, lines amount (1), jumpable_info */

#define CD_START_BUILD_JUMP_TABLES 2035

#define CD_JUMP_TABLES_BUILD_COMPLETED 2036
/* %d */
/* percents completed (0-100) */


/**************************************************************************/
/*                 Analyzer of Correctness                                */
#define CD_START_CORRECTNESS_ANALYZER 2037
/* [-<options>] "task"
:

task  -     ;
*/

/* The answers that Correctness Analysis completed: */
#define CD_CORRECTNESS_ANALYZER_FAILED 2042
/* free text */
#define CD_CORRECTNESS_ANALYZER_READY 2043


#define CD_GET_CORRECTNESS_ERROR_SUMMARY_TABLE 2038

#define CD_TAKE_CORRECTNESS_ERROR_SUMMARY_TABLE 2039
/* "%s" %d %d %d .... "%s" %d %d %d */
/* Error_type_1, Nerr_1, Nproc_1, Nsrc_1 ... Error_type_N, Nerr_N, Nproc_N, Nsrc_N */

#define CD_GET_CORRECTNESS_ERROR_TABLE 2040

#define CD_TAKE_CORRECTNESS_ERROR_TABLE 2041
/* ("%s" %d "%s" (%d %f %d %d "%s" %s) .... (%d %f %d %d "%s" %s)) ...
    ... ("%s" %d "%s" (%d %f %d %d "%s" %s) .... (%d %f %d %d "%s" %s)) */
/* (Error_type1 line_amount1==K Text_1
      (Nproc_1_1, Time_1_1, Nsrc_1_1, Event_1_1, Source file_1_1: Source line_1_1, MPI_func_1_1) ...
      (Nproc_1_K, Time_1_K, Nsrc_1_K, Event_1_K, Source file_1_K: Source line_1_K, MPI_func_1_K)
   )
   ....
   (Error_typeN line_amountN==M  Text_N
      (Nproc_N_1, Time_N_1, Nsrc_N_1, Event_N_1, Source file_N_1: Source line_N_1, MPI_func_N_1) ...
      (Nproc_N_M, Time_N_M, Nsrc_N_M, Event_N_M, Source file_N_M: Source line_N_M, MPI_func_N_M)
   )
*/

#define CD_GET_STACK_BY_SRC 2044
/* %d - SRC number */

#define CD_TAKE_STACK_BY_SRC 2045
/* %d %d ("%s" %d %d %s) ... ("%s" %d %d %s) */
/* SRC number lines_amount (source_file_name, line, text_length, text) ... (source_file_name, line, text_length, text) */

#endif
