#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>

#include "TCPIPlib.h"
#include "Thread_Semaphore.h"
#include "Monitor.h"
#include "CombinedDebugger.h"
#include "CDMessages.h"
#include "NDMessages.h"

#ifdef DEBUG
static const char *MessageStr(TCPIPMessageType msg)
{
	static char buffer[40];

	switch(msg)
	{
	case CD_FAIL: return "CD_FAIL";
	case CD_OK: return "CD_OK";
	case CD_START_PARAMETERS: return "CD_START_PARAMETERS";
	case CD_ERROR: return "CD_ERROR";
	case CD_ANALYSIS_READY: return "CD_ANALYSIS_READY";
	case CD_GET_PROTOCOL_PARAMETERS: return "CD_GET_PROTOCOL_PARAMETERS";
	case CD_TAKE_PROTOCOL_PARAMETERS: return "CD_TAKE_PROTOCOL_PARAMETERS";
	case CD_GET_PROTOCOL_TEXT: return "CD_GET_PROTOCOL_TEXT";
	case CD_TAKE_PROTOCOL_TEXT: return "CD_TAKE_PROTOCOL_TEXT";
	case CD_GET_PROTOCOL_JUMP: return "CD_GET_PROTOCOL_JUMP";
	case CD_TAKE_PROTOCOL_JUMP: return "CD_TAKE_PROTOCOL_JUMP";
	case CD_STOP: return "CD_STOP";
	case CD_GET_TASKS_AVAILABLE: return "CD_GET_TASKS_AVAILABLE";
	case CD_TAKE_TASKS_AVAILABLE: return "CD_TAKE_TASKS_AVAILABLE";
	case CD_GET_TRACE_AVAILABLE: return "CD_GET_TRACE_AVAILABLE";
	case CD_TAKE_TRACE_AVAILABLE: return "CD_TAKE_TRACE_AVAILABLE";
	case CD_GET_TRACE_PARAMETERS: return "CD_GET_TRACE_PARAMETERS";
	case CD_TAKE_TRACE_PARAMETERS: return "CD_TAKE_TRACE_PARAMETERS";
	case CD_GET_TRACE_EVENT: return "CD_GET_TRACE_EVENT";
	case CD_TAKE_TRACE_EVENT: return "CD_TAKE_TRACE_EVENT";
	case CD_GET_TRACE_EVENT_JUMPABLE: return "CD_GET_TRACE_EVENT_JUMPABLE";
	case CD_TAKE_TRACE_EVENT_JUMPABLE: return "CD_TAKE_TRACE_EVENT_JUMPABLE";
	case CD_GET_TRACE_EVENT_TEXT: return "CD_GET_TRACE_EVENT_TEXT";
	case CD_TAKE_TRACE_EVENT_TEXT: return "CD_TAKE_TRACE_EVENT_TEXT";
	case CD_GET_TRACE_EVENT_JUMP: return "CD_GET_TRACE_EVENT_JUMP";
	case CD_TAKE_TRACE_EVENT_JUMP: return "CD_TAKE_TRACE_EVENT_JUMP";
	case CD_GET_SOURCE_FILE_PARAMETERS: return "CD_GET_SOURCE_FILE_PARAMETERS";
	case CD_TAKE_SOURCE_FILE_PARAMETERS: return "CD_TAKE_SOURCE_FILE_PARAMETERS";
	case CD_GET_SOURCE_FILE_TEXT: return "CD_GET_SOURCE_FILE_TEXT";
	case CD_TAKE_SOURCE_FILE_TEXT: return "CD_TAKE_SOURCE_FILE_TEXT";
	case CD_GET_SOURCE_FILE_JUMP: return "CD_GET_SOURCE_FILE_JUMP";
	case CD_TAKE_SOURCE_FILE_JUMP: return "CD_TAKE_SOURCE_FILE_JUMP";
	case CD_GET_SOURCE_FILE_JUMPABLE: return "CD_GET_SOURCE_FILE_JUMPABLE";
	case CD_TAKE_SOURCE_FILE_JUMPABLE: return "CD_TAKE_SOURCE_FILE_JUMPABLE";
	case CD_INFO_SOURCE_FILE_JUMPABLE_CHANGED: return "CD_INFO_SOURCE_FILE_JUMPABLE_CHANGED";
	case CD_START_BUILD_JUMP_TABLES: return "CD_START_BUILD_JUMP_TABLES";
	case CD_JUMP_TABLES_BUILD_COMPLETED: return "CD_JUMP_TABLES_BUILD_COMPLETED";
	case CD_START_CORRECTNESS_ANALYZER: return "CD_START_CORRECTNESS_ANALYZER";
	case CD_GET_CORRECTNESS_ERROR_SUMMARY_TABLE: return "CD_GET_CORRECTNESS_ERROR_SUMMARY_TABLE";
	case CD_TAKE_CORRECTNESS_ERROR_SUMMARY_TABLE: return "CD_TAKE_CORRECTNESS_ERROR_SUMMARY_TABLE";
	case CD_GET_CORRECTNESS_ERROR_TABLE: return "CD_GET_CORRECTNESS_ERROR_TABLE";
	case CD_TAKE_CORRECTNESS_ERROR_TABLE: return "CD_TAKE_CORRECTNESS_ERROR_TABLE";
	case CD_CORRECTNESS_ANALYZER_FAILED: return "CD_CORRECTNESS_ANALYZER_FAILED";
	case CD_CORRECTNESS_ANALYZER_READY: return "CD_CORRECTNESS_ANALYZER_READY";
	case CD_GET_STACK_BY_SRC: return "CD_GET_STACK_BY_SRC";
	case CD_TAKE_STACK_BY_SRC: return "CD_TAKE_STACK_BY_SRC";
	}

	sprintf(buffer, "Unknown message (%d)", msg);
	return buffer;
}
#endif

void OnPassMessage(void)
{
	int receiver = 1 - active_thread_recv;
#ifdef DEBUG
	printf("Message from %s to %s: %s %s\n",
		(active_thread_recv == 1 ? "UI" : "Tracer"),
		(active_thread_recv == 1 ? "Tracer" : "UI"),
		MessageStr(mess_id_recv[active_thread_recv]), task_buffer_recv[active_thread_recv] ? task_buffer_recv[active_thread_recv] : "<null>");
#endif
	assert(number_of_norma_tasks == 1);
	if(thread_should_stops[receiver])
	{// receiver aborted
		PrintErrorMessage("Can't send message to exited component");
		if(task_buffer_recv[active_thread_recv])
			free(task_buffer_recv[active_thread_recv]);
		return;
	}

	/* Take semaphore that we may use the send buffer */	
	sem_wait_not_intr(&send_thread_sem[receiver]);
	/* be sure that there is no data in buffer */
	assert(task_buffer_send[receiver] == NULL);
	mess_id_send[receiver] = mess_id_recv[active_thread_recv];
	task_buffer_send[receiver] = task_buffer_recv[active_thread_recv];
	sem_post(&send_thread_sem_has_work[receiver]);
}



/********************************************************************************/
/*						Combined Debugger message handlers						*/
/********************************************************************************/
void CD_main_switch(int *number_stops)
{
	/* Main switch - what kind of message received */
	switch(mess_id_recv[active_thread_recv])
	{
	/* UI send break */
	case ND_BREAK_EXECUTION:
	case CD_STOP:
#ifdef DEBUG
		fputs("Message: CD_STOP\n", stderr);
#endif
		if(active_thread_recv == number_of_norma_tasks)
		{/* this is UI command to break execution */
			OnBreakExecution(CD_STOP);
			*number_stops = 1;
		}
		/* this is notification that task breaking execution */
		assert(*number_stops == 1);
		if(OnNotifyBreakExecution())
		{
			StopPWaitTimer();
			*number_stops = 2;
		}
		break;

	default:
		OnPassMessage();
	}
}

