#ifndef _FORMULA_H
#define _FORMULA_H
#include "CommonMonitorDefs.h"


/* Whait is supported in expressions?
Variables : a, B, Hren;
Operators: +-/*, <>=;
Values: Integer and float: 34, 2.56
Index construct: [20], [1..2, k=5..6];
Function: MIN, MAX;
Operation orders: ();
*/

typedef struct Index
{
	char name[MAX_VARIABLE_NAME+1];
	int from, to;
	BOOL found;
	struct Index *next;
} Index;

typedef struct Value
{
	NormaVarType type;
	union
	{
		int int_value;
		float real_value;
		double double_value;
		char *char_value;
	} value;
	char *indexes;
} Value;

typedef struct ValuesArray
{
	int amount;
	Value *values;
} ValuesArray;

typedef struct ValuesProcArray
{
	int proc_amount;
	ValuesArray *array;
} ValuesProcArray;

typedef struct Variable
{
	char name[MAX_VARIABLE_NAME+1];
	Index *index;
} Variable;

typedef enum
{
	PLUS = 0,
	MINUS,
	DIV,
	MULT,
	EQUAL,
	NEQUAL,
	MORE,
	MOREEQ,
	LESS,
	LESSEQ,
	ALL_EQUAL,
	ALL_NEQUAL,
	ALL_MORE,
	ALL_MOREEQ,
	ALL_LESS,
	ALL_LESSEQ,
	ANY_EQUAL,
	ANY_NEQUAL,
	ANY_MORE,
	ANY_MOREEQ,
	ANY_LESS,
	ANY_LESSEQ,
	NOT,
	AND,
	OR
} OperatorType;

typedef struct Operator
{
	OperatorType type;
	struct Formula *left, *right;
} Operator;

typedef enum
{
	MAXFunction,
	MINFunction,
	WHEREFunction
} FunctionType;

typedef struct Function
{
	FunctionType type;
	struct Formula *operand;
} Function;
/* Whait is supported in expressions?
Variables : a, B[i=8], Hren[1..10];
Operators: +-/*, <>=;
Values: Integer and float: 34, 2.56
Index construct: [20], [1..2, k=5..6];
Function: MIN, MAX;
Operation orders: ();
*/
typedef enum
{
	NoItem,
	VariableItem,
	OperatorItem,
	ValueItem,
	FunctionItem

} FormulaItemType;

typedef struct Formula
{
	FormulaItemType type;
	union
	{
		Variable *variable;
		Operator *noperator;
		Value *value;
		Function *function;
	} item;
} Formula;

typedef struct FormulaList
{
	Formula *formula;
	int ID;
	struct FormulaList *next;
} FormulaList;

typedef struct FormulaArray
{
	Formula *formula[MAX_FORMULA_ITEMS];
	int amount;
} FormulaArray;

typedef enum
{
	ErrorToken,
	EmptyToken,
	MAXToken,
	MINToken,
	WHEREToken,
	PlusToken,
	MinusToken,
	MultToken,
	DivToken,
	EqualToken,
	NotEqToken,
	MoreEqToken,
	MoreToken,
	LessEqToken,
	LessToken,
	All_EqualToken,
	All_NotEqToken,
	All_MoreEqToken,
	All_MoreToken,
	All_LessEqToken,
	All_LessToken,
	Any_EqualToken,
	Any_NotEqToken,
	Any_MoreEqToken,
	Any_MoreToken,
	Any_LessEqToken,
	Any_LessToken,
	NotToken,
	AndToken,
	OrToken,
	IndexStartToken,
	IndexEndToken,
	ParenthesisStartToken,
	ParenthesisEndToken,
	IntegerToken,
	RealToken,
	DoubleToken,
	VariableToken,
	AssignToken,
	RangeToken,
	CommaToken,
	SemicolonToken
} FormulaTokenType;

/* Public functions */
Formula* BuildFormula(const char *formula, int len, int *error_pos,
					  char **error_message);
void FreeFormula(Formula *formula);
void FreeFormulaListItem(FormulaList *formula_list_item);
void FreeFormulaList(FormulaList *formula_list);
FormulaList *MakeFormulaListItem(Formula *formula, int ID);

ValuesProcArray *CalculateFormula(const Formula *formula, char **error_message);
ValuesProcArray *CalculateVariable(const Variable *variable, char **error_message, BOOL retrieve_index_info);

char *PrintValuesProcArray(const ValuesProcArray *value_array, int ID /* just for print */);

BOOL IsValueTrue(const Value *value);
BOOL IsAnyValueTrue(const ValuesProcArray *value);
int ValueEqual(Value *v1, Value *v2);

Value* MakeValues(int amount);
Value* MakeValue(void);
ValuesProcArray *MakeValuesProcArray(int nproc, int amount_in_each);

void FreeValues(Value *value, int amount);
void FreeValue(Value *value);
void FreeValuesProcArray(ValuesProcArray *values);

#endif
