#ifdef WIN32
#include <process.h>
#include <direct.h>
#else
#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>

#include "TCPIPlib.h"
#include "NDMessages.h"
#include "CDMessages.h"
#include "CommonMonitorDefs.h"
#include "MonitorDefs.h"
#include "Formula.h"
#include "Thread_Semaphore.h"
#include "Monitor.h"
#include "CombinedDebugger.h"

/********************************************************/
TCPIPConnectionHandle *tcp_ip_handle;
static TCPIPConnectionHandle ui_handle;

/* The amount of running Norma tasks */
int number_of_norma_tasks = 0;
static BOOL is_fortran_program = FALSE;

/* Thread identifiers: */
/*	threads that sends data to Norma tasks and UI*/
pthread_t *task_thread_send, 
/*	threads that receives data from Norma tasks and UI*/
	*task_thread_recv;
/* Semaphores: */
/* Semaphore that recv tread may receive */
sem_t *recv_thread_sem, 
/* Semaphore that send tread ready to send */
	*send_thread_sem,
/* Semaphore that send tread has anything to send */
	*send_thread_sem_has_work,
/* Semaphore that main tread is doing anything and can't process anything more */
	main_thread_sem_busy, 
/* Semaphore that main tread has work from any recv thread */
	main_thread_sem_has_work,
	/* Semaphores for calculating variable values */
	calc_wait_sem, calc_ready_sem;
/* Variable pointed what thread has received anything */
int active_thread_recv;

/* ID of messages */
int *mess_id_send, *mess_id_recv;
/* Data of messages */
char **task_buffer_send, **task_buffer_recv;
/* Is this task wait the answer from UI; */
/* 2 means that necessary call EndPart on continue */
/* 3 means that it is on program start (no cur_part_info) */
/* 4 means that it is end of distribution part and necessary
 call EndPart and OnContinueSignal(1) on continue */
/* 5 means that it is end of Main part and necessary
 call EndPart and OnContinueSignal(0) on continue */
static int *task_wait_answer; 
/* will be on when this recv and send threads should stops */
BOOL *thread_should_stops;
/* 0 - part isn't in wait state;
   1 - part is in wait state;
   2 - part got signal continue, so answer for 'wait' should be send immediately
   Not used: 3 - part may be in potentially wait (Send/Receive) state
*/
static int *part_wait_cont_signal;
static int main_part_thread_ID;
static char *value_received_str;
static BOOL value_received_error;
static pthread_attr_t attr; /* attributes for thread - for set them PTHREAD_CREATE_DETACHED */
static BOOL ALL_tasks_successfully_connected, Server_has_canceled, server_for_UI;
static int wait_connect_from_tasks = 3000;
static int number_stops; /* 0 - nobody stopped, 1 - tasks, 2 - tasks and UI */
static CheckedFiles *beg_CheckedFiles = NULL;
static TaskGroups *beg_TaskGroups = NULL, *current_looked_TaskGroup;
static BOOL tasksGroupsCreated, tasksGroupsActual, task_groups_shall_be_reported;


static THREADRETURN start_task_thread_send(void *arg);
static THREADRETURN start_task_thread_recv(void *arg);
static THREADRETURN thread_IsStopHere(void *arg);
static THREADRETURN thread_CalculateFormula(void *arg);

/********************************************************/
/* check breakpionts stuff */
static sem_t sem_IsStopHere_has_work, sem_IsStopHere_busy;
static FormulaList *beg_calc_formula_list = NULL;
static sem_t sem_calc_formula_has_work, sem_calc_formula_busy, sem_calc_formula_list;
/********************************************************/
static PartInfo **cur_part_info;
static VarInfo **global_var_info;

/*  TIMER for wait any event */
#ifdef WIN32
static UINT_PTR pwait_timer;
static DWORD main_thread_ID;
static UINT event_message;
#else
static int pwait_timer;
#endif

static BOOL is_step_into = TRUE;

static Breakpoint *beg_breakpoint = NULL;
static int breakpoints_count = 0;

/* flag: the kind of Monitor: tracer or Norma */
static BOOL is_combined_debugger = FALSE;

void PrintErrorMessage(const char *msg)
{
//	char str[10];
	fputs(msg, stderr);
	fputc('\n',stderr);
//	gets(str);
}
static void PrintNotifyMessage(const char *msg)
{
	fputs(msg, stderr);
	fputc('\n',stderr);
}
static void out_memory(void) {PrintErrorMessage("Out memory!");}

/********************************************************/

static BOOL ReadCfgFile(const char *MonitorCfgFileName, IPInfo *mi)
{
	FILE *f;
	char buffer[256];
	
	if((f = fopen(MonitorCfgFileName, "r")) == NULL)
	{
		mi->is_server = TRUE;
		mi->port_number = DEFAULT_PORT_NUMBER;
		return TRUE;
	}
	
	if(fscanf(f, "%s\n", mi->address) != 1)
		goto monitor_format_error;
	if(fscanf(f, "%d\n", &mi->port_number) != 1)
		goto monitor_format_error;
	mi->is_server = FALSE;

	fclose(f);
	return TRUE;

monitor_format_error:
	fclose(f);
	sprintf(buffer, MONITOR_CFG_FILE_FORMATERROR, MonitorCfgFileName);
	PrintErrorMessage(buffer);
	return FALSE;
}

/********************************************************/

static BOOL BuildMonitorInfoFile(const char *MonitorInfoFileName, IPInfo *mi)
{
	FILE *f;
	
	if((f = fopen(MonitorInfoFileName, "w")) == NULL)
	{
		PrintErrorMessage(ER_WR);
		return FALSE;
	}
	fprintf(f, "%s\n%d\n", mi->address, mi->port_number);
	fputs(SERVER_ID, f);
	fclose(f);
	return TRUE;
}

/********************************************************/
static void FreeVarInfo(VarInfo *vars)
{
	VarInfo *tek;
	while(vars)
	{
		tek = vars;
		vars = vars->next;
		if(tek->number_of_dimensions)
			free(tek->dimensions);
		free(tek);
	}
}
/********************************************************/

static void RotateOperartorInfo(PartInfo *part_info,
	int current_operator,
	int current_operator_line_begin,
	int current_operator_line_end)
{
	part_info->previous_operator = part_info->current_operator;
	part_info->current_operator = current_operator;
	part_info->previous_operator_line_begin =
		part_info->current_operator_line_begin;
	part_info->previous_operator_line_end =
		part_info->current_operator_line_end;
	part_info->current_operator_line_begin = current_operator_line_begin;
	part_info->current_operator_line_end = current_operator_line_end;
}

static void StartNewPart(const char *name, const char *filename,
						 int thread_number, BOOL is_External)
{
	PartInfo *new_part_info;

#ifdef DEBUG
	fputs("StartNewPart\n", stderr);
#endif

	if((new_part_info = malloc(sizeof(PartInfo))) == NULL)
	{
		out_memory();
		return;
	}

	new_part_info->is_step_over = new_part_info->is_tmp_breakpoint = FALSE;
	if(cur_part_info[thread_number] == NULL)
	{ /* First part strated in program */
		if(!is_fortran_program &&
			(thread_number != main_part_thread_ID)) /* this is the main part that called distribution */
			new_part_info->is_step_into = cur_part_info[main_part_thread_ID]->is_step_into;
		else
			new_part_info->is_step_into = is_step_into;
	}
	else
	{
		new_part_info->is_step_into = cur_part_info[thread_number]->is_step_into;
		cur_part_info[thread_number]->is_step_into = FALSE;
	}

	new_part_info->next = cur_part_info[thread_number];
	cur_part_info[thread_number] = new_part_info;

	assert(strlen(name) < MAX_PART_NAME);
	assert(strlen(filename) < MAX_FILE_NAME);
	
	strcpy(new_part_info->name, name);
	strcpy(new_part_info->filename, filename);
	new_part_info->previous_operator = 0;
	new_part_info->current_operator = 0;
	new_part_info->previous_operator_line_begin = -1; /* means that we at the begin of the part, and differ from when at the begin of the first operator (== 0) */
	new_part_info->previous_operator_line_end = 0;
	new_part_info->current_operator_line_begin = 0;
	new_part_info->current_operator_line_end = 0;
	new_part_info->i = 0; /* not distr* part */
	new_part_info->vars = NULL;
	new_part_info->number_from_root = 
		(new_part_info->next ? new_part_info->next->number_from_root + 1 : 1);
	new_part_info->is_External = is_External;
}

static BOOL EndPart(int thread_number)
{
	PartInfo *was_part_info = cur_part_info[thread_number];
	BOOL was_tmp_breakpoint;
	

#ifdef DEBUG
	fprintf(stderr, "EndPart %d\n", thread_number);
#endif

	cur_part_info[thread_number] = cur_part_info[thread_number]->next;

	if(cur_part_info[thread_number])
	{
		if((was_tmp_breakpoint = cur_part_info[thread_number]->is_tmp_breakpoint) == TRUE)
		{
			cur_part_info[thread_number]->is_tmp_breakpoint = FALSE;
			cur_part_info[thread_number]->is_step_into = TRUE;
		}
		else if(was_part_info->is_step_into || was_part_info->is_step_over)
			cur_part_info[thread_number]->is_step_over = TRUE;
	}
	else
		was_tmp_breakpoint = FALSE;

	FreeVarInfo(was_part_info->vars);
	free(was_part_info);
	return was_tmp_breakpoint;
}

static void FreeFileCRC(void)
{
	CheckedFiles *tek_CheckedFiles = beg_CheckedFiles;
	while(beg_CheckedFiles)
	{
		tek_CheckedFiles = beg_CheckedFiles;
		beg_CheckedFiles = beg_CheckedFiles->next;
		free(tek_CheckedFiles);
	}
}
/********** Add the file CRC. ***************** */
static void AddFileCRC(const char *filename, long CRC)
{
	CheckedFiles *tek_CheckedFiles;
	if((tek_CheckedFiles = malloc(sizeof(CheckedFiles))) == NULL)
	{
		out_memory();
		return;
	}
	strcpy(tek_CheckedFiles->filename, filename);
	tek_CheckedFiles->CRC = CRC;
	tek_CheckedFiles->next = beg_CheckedFiles;
	beg_CheckedFiles = tek_CheckedFiles;
}
/********** Check the file CRC. First look in cache, and, if it is, return TRUE;
            Otherwise install in cache, ask UI and return FALSE */
static void CheckFileCRC(const char *filename, long CRC)
{
	CheckedFiles *tek_CheckedFiles = beg_CheckedFiles;
	char *buffer;

	assert(strlen(filename) + 1 < MAX_FILE_NAME);
	while(tek_CheckedFiles)
	{
		if((tek_CheckedFiles->CRC == CRC) &&
			(strcmp(tek_CheckedFiles->filename, filename) == 0))
			return;  // this file already was checked
		tek_CheckedFiles = tek_CheckedFiles->next;
	}
	// insert in list
	AddFileCRC(filename, CRC);
	if((buffer = malloc(MAX_FILE_NAME + MAX_NUMBER_LEN + 4)) == NULL)
	{
		out_memory();
		return;
	}
	sprintf(buffer, "\"%s\" %ld", filename, CRC);
	/* Take semaphore that we may use the send buffer */	
	sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
	/* be sure that there is no data in buffer */
	assert(task_buffer_send[number_of_norma_tasks] == NULL);
	mess_id_send[number_of_norma_tasks] = ND_CHECK_CRC;
	task_buffer_send[number_of_norma_tasks] = buffer;
	sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
}

/********************************************************/
static void ErrorSpawn(const char *message)
{
	PrintNotifyMessage(message);
	/* send the message to UI */
	if(!TCPIPSendMessage(ui_handle, 
		ND_DEBUGED_PROGRAM_CANT_RUN, message))
		PrintErrorMessage("Error send message");
	TCPIPDisconnect(ui_handle);
}
/********************************************************/
/* Thread that should stop Monitor's server when debugging program can't start */
#ifdef WIN32
static THREADRETURN thread_Cancel(void *arg)
{
	DWORD status;
	pid_t *chpid = (pid_t *)arg;
	
	Sleep(wait_connect_from_tasks);
	if(ALL_tasks_successfully_connected)
		return;

	if((GetExitCodeProcess((HANDLE)(*chpid), &status) == FALSE) ||
		((status != STILL_ACTIVE) && (status != 0)))
	{
		ErrorSpawn("Executed program has terminated unexpectedly");
		TCPIPCancelServer();
		Server_has_canceled = TRUE;
	}
}
#endif
/* Thread that should stop Monitor if UI send the break command while monitor waits the connect from tasks */
static THREADRETURN thread_ListenBreak(void *arg)
{
	TCPIPMessageType rez;

	TCPIPSetConnectionNonBlockedMode(ui_handle);
	while(!ALL_tasks_successfully_connected)
	{
		if(TCPIPGetMessage(ui_handle, &rez, NULL))
		{
			if((rez == ND_BREAK_EXECUTION) || (rez == CD_STOP))
			{
				if(!server_for_UI)
					TCPIPDisconnect(ui_handle);
				TCPIPCancelServer();
				Server_has_canceled = TRUE;
				return;
			}
			else if(rez != SYS_NO_MESSAGES)
				PrintErrorMessage("Unexpected message from UI\n");
		}
		else
		{
			if(!server_for_UI)
				TCPIPDisconnect(ui_handle);
			TCPIPCancelServer();
			Server_has_canceled = TRUE;
			PrintErrorMessage("Connection with UI has been lost");
			return;
		}
		Sleep(100);
	}
	TCPIPSetConnectionBlockedMode(ui_handle);
	sem_post(&calc_ready_sem); /* indicate that ui_handle returned */
}
/********************************************************/
static BOOL AssignArg(char **arg, const char *string)
{
	if(strchr(string, ' '))
	{ /* space in argument */
		if((*arg = malloc(strlen(string) + 3)) == NULL)
		{
			out_memory();
			*arg = (char *)string;
			return FALSE;
		}
		sprintf(*arg, "\"%s\"", string); 
		return TRUE;
	}
	*arg = (char *)string;
	return FALSE;
}
/********************************************************/

 
int main(int argc, char **argv)
{
	int i, j, k;
	IPInfo mi;
	char **mpi_args, *exec_prg;
	pid_t chpid;
	static int main_loop(void);
	BOOL is_remote_server = FALSE;
	BOOL *free_mpi_arg;
	TCPIPConnectionHandle task_handle;
	char *data;
	TCPIPMessageType rez;
	int expected_tasks_amount = 0;
	int gate_ID = 0;
	BOOL is_renamed_as_CD = FALSE;
	
	puts("Monitor started");

#ifdef DEBUG
	puts("With args:");
	for(i = 0; i < argc; i++)
		puts(argv[i]);
#endif

	if(strcmp(argv[0], "monitorCD") == 0)
		is_renamed_as_CD = TRUE;

	if(argc == 1)
	{
		printf("Monitor %sfor Combined Debugger.\nVersion %s built at %s.\n",
			is_renamed_as_CD ? "" : "for debugging Norma programs and ", VERSIONID, __DATE__);
		puts("Usage:\nmonitor <arguments for mpi_run> [options]\n"
			"options are:\n"
			"  -server <host>[:<port>[:<gate ID>]]\n"
			"  -remote[:<port>]\n"
			"  -port:<port>\n"
			"  -np <amount>\n"
			"  -exec <programname>\n"
			"  -tracer <programname>\n"
			"  -wd <path>\n"
			"Example:\nmonitor -server first.applmat.ru:80 -np 5 mpi_go\n");
		return 1;
	}

	if(((mpi_args = malloc(argc * sizeof(char*))) == NULL) ||
		((free_mpi_arg = calloc(argc, sizeof(BOOL))) == NULL))
	{
		out_memory();
		return 1;
	}

	mi.is_server = TRUE;
	mi.port_number = 0;

	mpi_args[0] = exec_prg = "mpirun";
	for(i = 1, j = 1; i < argc; i++)
	{
		if((strcasecmp(argv[i], "-np") == 0) && (i + 1 < argc))
			expected_tasks_amount = atoi(argv[i + 1]);
		if(strncasecmp(argv[i], "-remote", 7) == 0)
		{
			is_remote_server = TRUE;
			mi.is_server = TRUE;
			if(argv[i][7] == ':')
				mi.port_number = atoi(&argv[i][8]);
			else
				mi.port_number = DEFAULT_PORT_NUMBER;
			continue;
		}
		if(strncasecmp(argv[i], "-port:", 6) == 0)
		{
			mi.is_server = TRUE;
			mi.port_number = atoi(&argv[i][6]);
			continue;
		}
		else if(strcasecmp(argv[i], "-server") == 0)
		{
			if(++i == argc)
			{
				PrintErrorMessage("Option '-server' request the hostname:port followed");
				return 1;
			}
			mi.is_server = FALSE;
			for(k = 0; argv[i][k]; k++)
				if(argv[i][k] == ':')
					break;
			if(k > MAX_INTERNET_ADDRESS_SIZE)
				k = MAX_INTERNET_ADDRESS_SIZE;
			memcpy(mi.address, argv[i], k);
			mi.address[k] = 0;
			if(argv[i][k] == ':')
			{
				int pos = k + 1;
				while(isdigit(argv[i][pos]))
				    pos++;
				if(argv[i][pos] == ':')
				{
					argv[i][pos] = '\0';
					gate_ID = atoi(&argv[i][pos + 1]);
				}
				mi.port_number = atoi(&argv[i][k + 1]);
			}
			else
				mi.port_number = DEFAULT_PORT_NUMBER;
			continue;
		}
		else if(strcasecmp(argv[i], "-exec") == 0)
		{
			if(++i == argc)
			{
				PrintErrorMessage("Option '-exec' request the program file name for executing followed");
				return 1;
			}
			free_mpi_arg[0] = AssignArg(&mpi_args[0], argv[i]);
			exec_prg = argv[i];
			continue;
		}

		else if(strcasecmp(argv[i], "-wd") == 0)
		{
			if(++i == argc)
			{
				PrintErrorMessage("Option '-wd' request the path followed");
				return 1;
			}
			if(chdir(argv[i]))
			{
				char buffer[100];
				sprintf(buffer, "Can't change current dir to %s\n", argv[i]);
				PrintErrorMessage(buffer);
				return 1;
			}
#ifdef DEBUG
			{
				char str[200];
				printf("Current dir has been changed to: %s\n",
					getcwd(str, 200));
			}
#endif
			continue;
		}

		else if(strcasecmp(argv[i], "-tracer") == 0)
		{
			if(++i == argc)
			{
				PrintErrorMessage("Option '-tracer' request the program file name for executing followed");
				return 1;
			}
			free_mpi_arg[0] = AssignArg(&mpi_args[0], argv[i]);
			exec_prg = argv[i];
			is_combined_debugger = TRUE;
			continue;
		}

		/* option for executed program */
		free_mpi_arg[j] = AssignArg(&mpi_args[j], argv[i]);
		j++;
	}

	if(is_renamed_as_CD)
	{
		free_mpi_arg[0] = AssignArg(&mpi_args[0], "CDmain");
		exec_prg = "CDmain";
		is_combined_debugger = TRUE;
	}

	mpi_args[j] = NULL;

	if(!is_remote_server && (mi.port_number == 0) && mi.is_server &&
		(ReadCfgFile(MONITOR_CFG_FILENAME, &mi) == FALSE))
		return 1;

	server_for_UI = mi.is_server;
	if(mi.is_server)
	{
		mi.type = 0; /* we wait the UI */
		if(TCPIPServerCreateSocket(&mi) == FALSE) /* for the 'address' field filling */
			return 1;
		if(BuildMonitorInfoFile(MONITOR_INFO_FILENAME, &mi) == FALSE)
			return 1;
	}
	else
	{
		if(gate_ID)
			mi.type = gate_ID;
		else
			mi.type = 2; /* we are the monitor-client */
	}
	
	puts("Wait for the UI connect...");

	if(TCPIPConnect(&mi, &ui_handle, NULL) == FALSE)
		return 1;

	if(server_for_UI) /* say that there is no gate in connection */
		TCPIPSendMessage(ui_handle, SYS_GATE_ID, NULL);

	puts("UI successfully connected");

	/* Build Monitor info file */
	if(mi.is_server == FALSE) /* change this - for norma tasks Monitor is always server */ 
	{
		mi.port_number = DEFAULT_PORT_NUMBER;
		mi.is_server = TRUE;
		if(TCPIPServerCreateSocket(&mi) == FALSE) /* for the 'address' field filling */
			return 1;
		if(BuildMonitorInfoFile(MONITOR_INFO_FILENAME, &mi) == FALSE)
			return 1;
	}
	mi.type = 1;

	if(is_remote_server)
	{
		puts("MonitorInfoFile created");
		chpid = -1;
	}
	else
	{
		/* start MPI: */
		if(!is_combined_debugger)
			puts("Start debugged programs...");
		else
			puts("Start tracer...");

#ifdef WIN32
		chpid = _spawnv(P_NOWAIT, exec_prg, mpi_args);
		if(chpid == -1)/* error start the new process */
		{
			char buf[MAX_FILE_NAME + 80];
			sprintf(buf, "Can't start %s", exec_prg);
			ErrorSpawn(buf);
			return 1;
		}
#else
		chpid = fork();
		if(chpid == -1) /* error fork */
		{
			ErrorSpawn("Error fork()");
			return 1;
		}
		if(chpid == 0) /* Child process */
		{
			char buf[MAX_FILE_NAME + 80];
			execvp(exec_prg, mpi_args);
			/* error */
			sprintf(buf, "Can't start %s", exec_prg);
			ErrorSpawn(buf);
			return 1;
		}
#endif
	}
	/* Parent process */
	for(i = 0; mpi_args[i]; i++)
		if(free_mpi_arg[i])
			free(mpi_args[i]);
	free(mpi_args);
	free(free_mpi_arg);

	/* Init the attributes for thread creating */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, 2 * PTHREAD_STACK_MIN); /* We don't need the large stack !!! */

	ALL_tasks_successfully_connected = FALSE;
	Server_has_canceled = FALSE;

/* Busy */
	if(sem_init(&main_thread_sem_busy, 0, 0) ||
/* Semaphore will be released when any work appear */
		sem_init(&main_thread_sem_has_work, 0, 0) ||
	/* Init the calculating semaphores */
		sem_init(&sem_IsStopHere_busy, 0, 0) ||
		sem_init(&sem_IsStopHere_has_work, 0, 0) ||
		sem_init(&sem_calc_formula_busy, 0, 0) ||
		sem_init(&sem_calc_formula_has_work, 0, 0) ||
		sem_init(&sem_calc_formula_list, 0, 1) ||
		sem_init(&calc_wait_sem, 0, 0) ||
		sem_init(&calc_ready_sem, 0, 0))
	{
		PrintErrorMessage("error_init sem");
		return 1;
	}

#ifdef WIN32
	if(chpid != -1)
	/* start the cancel thread */
	{
		pthread_t tn;

		if(pthread_create(&tn, &attr, thread_Cancel, &chpid))
		{
			PrintErrorMessage("error_init cancel thread");
			return 1;
		}
	}
#endif

	/* start the listen UI thread */
	{
		pthread_t tn;

		if(pthread_create(&tn, &attr, thread_ListenBreak, NULL))
		{
			PrintErrorMessage("error_init listen thread");
			return 1;
		}
	}

	/* connect with debugged programs */
	if(!is_combined_debugger)
		puts("Connecting with debugged programs...");
	else
		puts("Connecting with tracer programs...");

	number_of_norma_tasks = -1;

	for(i = 0; i != number_of_norma_tasks; i++)
	{
		if(TCPIPConnect(&mi, &task_handle, NULL) == FALSE)
		{
			if(!Server_has_canceled)
			{
				PrintErrorMessage("Connection failed");
				TCPIPDisconnect(ui_handle);
				if(i > 0)
					for(j = 0; j < number_of_norma_tasks; j++)
						if(tcp_ip_handle[j])
							TCPIPDisconnect(tcp_ip_handle[j]);
			}
			return 1;
		}
		
		if(is_combined_debugger)
		{ /* Tracer monitor */
			k = 1; /* currently we have only 1 tracer available */
			j = 0;
		}
		else
		{ /* Norma debugger */
			if(!TCPIPGetMessage(task_handle, &rez, &data) ||
				(rez != ND_TASK_ID) || (data == NULL) ||
				(sscanf(data, "%d %d", &k, &j) != 2))
			{
				PrintErrorMessage("Error get message from task");
				TCPIPDisconnect(ui_handle);
				if(i > 0)
					for(j = 0; j < number_of_norma_tasks; j++)
						if(tcp_ip_handle[j])
							TCPIPDisconnect(tcp_ip_handle[j]);
				return 1;
			}

			free(data);
		}

		if(i == 0)
		{
			assert(k > 0);
			number_of_norma_tasks = k;

			if(j == -1)
			{ /* Fortran program */
				is_fortran_program = TRUE;
			}

			if(!is_combined_debugger)
				printf("Amount of expected tasks = %d\n", number_of_norma_tasks);
			
			if((tcp_ip_handle = calloc(sizeof(TCPIPConnectionHandle), number_of_norma_tasks + 1)) == NULL)
			{
				out_memory();
				TCPIPDisconnect(ui_handle);
				return 1;
			}
			
			tcp_ip_handle[number_of_norma_tasks] = ui_handle;
		}
		else
		{
			assert(is_fortran_program || (number_of_norma_tasks == k));
		}
		if(is_fortran_program)
			j = i;
		assert(j < number_of_norma_tasks);
		assert(j >= 0);
		assert(tcp_ip_handle[j] == 0);
		tcp_ip_handle[j] = task_handle;
	}

	if(!is_remote_server)
		unlink(MONITOR_INFO_FILENAME);

	ALL_tasks_successfully_connected = TRUE;

	sem_wait_not_intr(&calc_ready_sem); /* indicate that ui_handle returned */
	if(is_combined_debugger)
		TCPIPSendMessage(ui_handle, CD_OK, NULL);

	puts("Start threads...");
#ifdef WIN32
	main_thread_ID = GetCurrentThreadId();
	event_message = RegisterWindowMessage("EventMessage");
#endif

/* Alloc the memory for threads staff */
	if(((task_thread_send = malloc(sizeof(pthread_t) * (number_of_norma_tasks + 1))) == NULL) ||
		((task_thread_recv = malloc(sizeof(pthread_t) * (number_of_norma_tasks + 1))) == NULL) ||
		((recv_thread_sem = malloc(sizeof(sem_t) * (number_of_norma_tasks + 1))) == NULL) ||
		((send_thread_sem = malloc(sizeof(sem_t) * (number_of_norma_tasks + 1))) == NULL) ||
		((send_thread_sem_has_work = malloc(sizeof(sem_t) * (number_of_norma_tasks + 1))) == NULL) ||
		((mess_id_send = malloc(sizeof(int) * (number_of_norma_tasks + 1))) == NULL) ||
		((mess_id_recv = malloc(sizeof(int) * (number_of_norma_tasks + 1))) == NULL) ||
		((task_buffer_send = calloc(sizeof(char *), (number_of_norma_tasks + 1))) == NULL) ||
		((task_buffer_recv = calloc(sizeof(char *), (number_of_norma_tasks + 1))) == NULL) ||
		((thread_should_stops = calloc(sizeof(BOOL), (number_of_norma_tasks + 1))) == NULL) ||
		((cur_part_info = calloc(sizeof(PartInfo), number_of_norma_tasks)) == NULL) ||
		((global_var_info = calloc(sizeof(VarInfo), number_of_norma_tasks)) == NULL) ||
		((part_wait_cont_signal = calloc(sizeof(int), number_of_norma_tasks)) == NULL) ||
		((task_wait_answer = calloc(sizeof(int), number_of_norma_tasks)) == NULL))
	{
		out_memory();
		return 1;
	}

	for(i = 0; i <= number_of_norma_tasks; i++)
	{
		
		if(sem_init(&send_thread_sem[i], 0, 1) ||
			sem_init(&send_thread_sem_has_work[i], 0, 0))
		{
			PrintErrorMessage("error_init sem send");
			return 1;
		}

		mess_id_send[i] = i;

		if(pthread_create(&task_thread_send[i], &attr, 
			start_task_thread_send, &mess_id_send[i]) ||
			pthread_create(&task_thread_recv[i], &attr, 
			start_task_thread_recv, &mess_id_send[i]))
		{
			PrintErrorMessage("error_init thread: start");
			return 1;
		}
	}

	/* Init the calculations threads */
	{
		pthread_t tn;
		if(pthread_create(&tn, &attr, thread_IsStopHere, NULL) ||
			pthread_create(&tn, &attr, thread_CalculateFormula, NULL))
		{
			PrintErrorMessage("error_init calc thread");
			return 1;
		}
	}

	puts("The threads started");

	sem_wait_not_intr(&calc_ready_sem);
	sem_post(&calc_ready_sem);

	main_loop();

	/* stop the check breakpoints thread */
	sem_wait_not_intr(&sem_IsStopHere_busy);
	assert(current_looked_TaskGroup == NULL);
	sem_post(&sem_IsStopHere_has_work);
	sem_wait_not_intr(&sem_IsStopHere_busy);
	/* stop the calculation user's formula thread */
	sem_wait_not_intr(&sem_calc_formula_busy);
	assert(beg_calc_formula_list == NULL);
	sem_post(&sem_calc_formula_has_work);
	sem_wait_not_intr(&sem_calc_formula_busy);

	pthread_attr_destroy(&attr);

	sem_destroy(&main_thread_sem_busy);
	sem_destroy(&main_thread_sem_has_work);
	sem_destroy(&sem_IsStopHere_busy);
	sem_destroy(&sem_IsStopHere_has_work);
	sem_destroy(&sem_calc_formula_busy);
	sem_destroy(&sem_calc_formula_has_work);
	sem_destroy(&sem_calc_formula_list);
	sem_destroy(&calc_wait_sem);
	sem_destroy(&calc_ready_sem);
	for(i = 0; i < number_of_norma_tasks + 1; i++)
	{
		sem_destroy(&send_thread_sem[i]);
		sem_destroy(&send_thread_sem_has_work[i]);
		sem_destroy(&recv_thread_sem[i]);
	}

	free(task_thread_send);
	free(task_thread_recv);
	free(recv_thread_sem);
	free(send_thread_sem);
	free(send_thread_sem_has_work);
	free(mess_id_send);
	free(mess_id_recv);
	free(task_buffer_send);
	free(task_buffer_recv);
	free(thread_should_stops);
	free(cur_part_info);
	free(global_var_info);
	free(part_wait_cont_signal);
	free(task_wait_answer);
	
	FreeFileCRC();

	for(i = 0; i < number_of_norma_tasks + 1; i++)
		TCPIPDisconnect(tcp_ip_handle[i]);

	free(tcp_ip_handle);

	if(!is_remote_server)
	{
		puts("Before wait child end");
	#ifdef WIN32
		_cwait(&i, chpid, 0);
	#else /* UNIX */
		for(j = 0; j < 5; j++)
		{
			if(waitpid(-1, &i, WNOHANG) > 0)/* i - status of process */
				break;
			sleep(1);
		}
		if(j == 5)
		{
			printf("Wait child end fail\n");
			return 1;
		}
	#endif
		printf("Return code is %d\n", i);
		return i;
	}
	puts("Monitor finished successfully");
	return 0;
}

/**********************************/
static void SendContinue(int id)
{
#ifdef DEBUG
	fprintf(stderr, "Send 'Continue' to %d\n", id);
#endif
	sem_wait_not_intr(&send_thread_sem[id]);
	assert(task_buffer_send[id] == NULL);
	mess_id_send[id] = ND_CONTINUE_EXECUTION;
	sem_post(&send_thread_sem_has_work[id]);
	part_wait_cont_signal[id] = 0;
}
/***************** called when wait signal comes*****************/
static void CheckTasksState(void);
static void OnWaitSignal(void)
{
#ifdef DEBUG
	fprintf(stderr, "OnWaitSignal of task (%d)\n", active_thread_recv);
#endif
	if(part_wait_cont_signal[active_thread_recv] == 2) /* continue already took place */
		SendContinue(active_thread_recv);
	else
	{
		part_wait_cont_signal[active_thread_recv] = 1;
		CheckTasksState();
	}
}
/***************** called when waiting process may continue*****************/
static void OnContinueSignal(BOOL is_for_main_task)
{
	int i;

#ifdef DEBUG
	fprintf(stderr, "OnContinueSignal %s main task\n", 
		(is_for_main_task ? "for" : "from"));
#endif

	if(is_for_main_task)
	{
		if(part_wait_cont_signal[main_part_thread_ID] == 1)
			SendContinue(main_part_thread_ID);
		else
			part_wait_cont_signal[main_part_thread_ID] = 2; /* Set that the part already got the 'Continue' signal */
	}
	else
		for(i = 0; i < number_of_norma_tasks; i++)
			if(i != main_part_thread_ID)
				if(part_wait_cont_signal[i] == 1)
					SendContinue(i);
				else
					part_wait_cont_signal[i] = 2; /* Set that the part already got the 'Continue' signal */
}
/************ These functions checks the states of tasks if they shoud
              be stoped or should continue execution *******************/
/********* This function DestroyTaskGroups *******************/
static void DestroyTaskGroups(void)
{
	TaskGroups *tek_TaskGroups;
#ifdef DEBUG
	fputs("DestroyTaskGroups\n", stderr);
#endif

	/* wait untill any tasks group is been checked for breakpoint - we can't destroy it at this moment */
	sem_wait_not_intr(&sem_IsStopHere_busy);
	sem_post(&sem_IsStopHere_busy);
	
	while(beg_TaskGroups)
	{
#ifdef DEBUG
		fprintf(stderr, "DestroyTaskGroup with %d tasks\n", beg_TaskGroups->amount);
#endif
		tek_TaskGroups = beg_TaskGroups;
		beg_TaskGroups = beg_TaskGroups->next;
		free(tek_TaskGroups->task_numbers);
		if(tek_TaskGroups->IDs)
			free(tek_TaskGroups->IDs);
		free(tek_TaskGroups);
	}
	tasksGroupsCreated = FALSE;
	tasksGroupsActual = FALSE;
	current_looked_TaskGroup = NULL;
	task_groups_shall_be_reported = FALSE;
}
/********* This function report the tasks groups to UI *******************/
static int Print_Group_info(char *buffer, int number_groups, TaskGroups *beg_TaskGroups)
{
	int pos, i;
	sprintf(buffer, " %d ", number_groups);
	pos = strlen(buffer);
	while(beg_TaskGroups)
	{
		/* file and line numbers */
		sprintf(&buffer[pos], "(\"%s\" %d %d",
			cur_part_info[beg_TaskGroups->task_numbers[0]]->filename,
			cur_part_info[beg_TaskGroups->task_numbers[0]]->current_operator_line_begin,
			cur_part_info[beg_TaskGroups->task_numbers[0]]->current_operator_line_end);
		pos += strlen(&buffer[pos]);
		/* process IDs */
		for(i = 0; i < beg_TaskGroups->amount; i++)
		{
			sprintf(&buffer[pos], " %d", beg_TaskGroups->task_numbers[i]);
			pos += strlen(&buffer[pos]);
		}
		buffer[pos++] = ')';
		beg_TaskGroups = beg_TaskGroups->next;
	}
	return pos;
}
static void ReportTasksGroups()
{
	int number_normal_groups = 0, number_blocked_groups = 0;
	char *buffer;
	TaskGroups *tek_TaskGroups, *beg_blocked = NULL;
	int i, pos;
	
	/* Calculate the amount of normal groups */
	tek_TaskGroups = beg_TaskGroups;
	while(tek_TaskGroups)
	{
		number_normal_groups++;
		tek_TaskGroups = tek_TaskGroups->next;
	}

	/* Create and calculate amount of blocked groups, calculate amount of not started processes */
	for(i = 0; i < number_of_norma_tasks; i++)
	{
		if((task_wait_answer[i] == 0) && cur_part_info[i])
		{
			TaskGroups *tek_TaskGroups = beg_blocked;
			while(tek_TaskGroups)
			{
				int index = tek_TaskGroups->task_numbers[0];
				if((cur_part_info[index]->current_operator_line_begin ==
					cur_part_info[i]->current_operator_line_begin) &&
					(strcmp(cur_part_info[index]->filename, cur_part_info[i]->filename) == 0))
				{ /* add to existing group */
					tek_TaskGroups->task_numbers[tek_TaskGroups->amount++] = i;
					break;
				}
				tek_TaskGroups = tek_TaskGroups->next;
			}
			if(tek_TaskGroups == NULL) /* new group */
			{
#ifdef DEBUG
				fprintf(stderr, "Create new TaskGroup\n");
#endif
				if(((tek_TaskGroups = malloc(sizeof(TaskGroups))) == NULL) ||
					((tek_TaskGroups->task_numbers = malloc(number_of_norma_tasks * sizeof(int))) == NULL))
				{
					out_memory();
					return;
				}
				tek_TaskGroups->amount = 1;
				tek_TaskGroups->task_numbers[0] = i;
				tek_TaskGroups->IDs = NULL;
				tek_TaskGroups->next = beg_blocked;
				beg_blocked = tek_TaskGroups;
				number_blocked_groups++;
			}
		}
	}
	
	if((buffer = malloc((MAX_FILE_NAME + 2 * 15) * (number_normal_groups + number_blocked_groups) +
		(breakpoints_count + number_of_norma_tasks) * 15)) == NULL)
	{
		out_memory();
		return;
	}
	
	/* print in buffer information about breakpoints */
	buffer[0] = '(';
	pos = 1;

	/* Check if any task has step into/over state
	   in this case don't report any reached breakpoints */
	tek_TaskGroups = beg_TaskGroups;
	while(tek_TaskGroups)
	{
		for(i = 0; i < tek_TaskGroups->amount; i++)
			if(cur_part_info[tek_TaskGroups->task_numbers[i]]->is_step_over ||
				cur_part_info[tek_TaskGroups->task_numbers[i]]->is_step_into)
				break;
		if(i < tek_TaskGroups->amount)
			break;
		tek_TaskGroups = tek_TaskGroups->next;
	}

	if(tek_TaskGroups == NULL)
	{
		tek_TaskGroups = beg_TaskGroups;
		while(tek_TaskGroups)
		{
			if(tek_TaskGroups->IDs)
				for(i = 0; tek_TaskGroups->IDs[i]; i++)
				{   /* remove the same breakpoints IDs from different groups. This may happens when breakpoint isn't related to place in file */
					TaskGroups *tg = beg_TaskGroups;
					int j;
					while(tg != tek_TaskGroups)
					{
						if(tg->IDs)
						{
							for(j = 0; tg->IDs[j]; j++)
								if(tg->IDs[j] == tek_TaskGroups->IDs[i])
									break;
							if(tg->IDs[j])
								break;
						}
						tg = tg->next;
					}
					if(tg == tek_TaskGroups)
					{  /* no same breakpoint ID found - report it */
						sprintf(&buffer[pos], " %d", tek_TaskGroups->IDs[i]);
						pos += strlen(&buffer[pos]);
					}
				}
			tek_TaskGroups = tek_TaskGroups->next;
		}
	}
	buffer[pos++] = ')';

	/* print in buffer information about normal groups */
	pos += Print_Group_info(&buffer[pos], number_normal_groups, beg_TaskGroups);
	/* print in buffer information about blocked groups */
	pos += Print_Group_info(&buffer[pos], number_blocked_groups, beg_blocked);
	/* free info about blocked groups */
	while(beg_blocked)
	{
		tek_TaskGroups = beg_blocked;
		beg_blocked = beg_blocked->next;
		free(tek_TaskGroups);
	}
	/* print in buffer information about not started processes */
	buffer[pos++] = '(';
	for(i = 0; i < number_of_norma_tasks; i++)
		if(cur_part_info[i] == NULL)
		{
			sprintf(&buffer[pos], " %d", i);
			pos += strlen(&buffer[pos]);
		}
	buffer[pos++] = ')';
	buffer[pos] = '\0';

	/* Take semaphore that we may use the send buffer */	
	sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
	/* be sure that there is no data in buffer */
	assert(task_buffer_send[number_of_norma_tasks] == NULL);
	mess_id_send[number_of_norma_tasks] = ND_NOTIF_OPERATOR_START;
	task_buffer_send[number_of_norma_tasks] = buffer;
#ifdef DEBUG
	fprintf(stderr, "Report to UI: %s\n", buffer);
#endif

#ifndef NDEBUG
	sem_getvalue(&send_thread_sem_has_work[number_of_norma_tasks], &i);
	assert(i == 0);
#endif
	sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
	DestroyTaskGroups();
}
/********* This function check reached breakpoints for the condition ANY/ALL
          and return TRUE if exists really reached breakpoins and FALSE if not *******************/
static BOOL TestBreakpoinsType(int *IDs)
{
	int i;

	if(IDs == NULL)
		return FALSE;

	if(IDs[0] == 0)
		return TRUE; /* Not breakpoint, step */

	for(i = 0; i < number_of_norma_tasks; i++)
		if((task_wait_answer[i] == 0) &&
			(part_wait_cont_signal[i] != 1) /*&& (part_wait_cont_signal[i] != 3)*/)
			break; /* task do anything */

	if(i == number_of_norma_tasks)
		return TRUE; /* In any case all tasks are stoped */

	for(i = 0; IDs[i]; i++)
	{
		Breakpoint *tek_breakpoint = beg_breakpoint;
		while(tek_breakpoint)
		{
			if(tek_breakpoint->id == IDs[i])
				break;
			tek_breakpoint = tek_breakpoint->next;
		}
		assert(tek_breakpoint);
		if(tek_breakpoint->is_any || (tek_breakpoint->filename[0] == 0))
			return TRUE;
	}
	return FALSE;
}

/********* This function check one task groups for breakpoint occurence
                and, if any, call report *******************/
static void CreateTasksGroups(BOOL for_check /* or for report */);
static void OnContinueExecution(int type, const int *task_numbers, int number_of_tasks);
static void StartPWaitTimer(UINT value);
static void CheckTasksGroup()
{
	if(!tasksGroupsCreated) /* this may happens when it was BREAK_EXECUTION between groups check */
		return;

	if(!tasksGroupsActual)
	{
		CreateTasksGroups(TRUE);
		return;
	}

	if(beg_TaskGroups == NULL)
		return; /* No task group were created - all processes are in run state */

	if(current_looked_TaskGroup)
	{
		if((current_looked_TaskGroup = current_looked_TaskGroup->next) == NULL)
		{ /* all groups were looked */
			current_looked_TaskGroup = beg_TaskGroups;
			while(current_looked_TaskGroup)
			{
				if(TestBreakpoinsType(current_looked_TaskGroup->IDs))
					break; /*  some breakpoints have worked */
				current_looked_TaskGroup = current_looked_TaskGroup->next;
			}
			if(current_looked_TaskGroup)
			{
				int i;
				for(i = 0; i < number_of_norma_tasks; i++)
					if((task_wait_answer[i] == 0) && (part_wait_cont_signal[i] == 0))
						break;
				if(i < number_of_norma_tasks)
				{
					/* wait before report the breakpoint for UI */
					task_groups_shall_be_reported = TRUE;
					StartPWaitTimer(PWAIT_TIMER_VALUE);
				}
				else
					ReportTasksGroups();
			}
			else
			{ /* no breakpoint have been reached - continue program execution */
				/* say ourself tasks to continue */
				if(number_stops == 0)
					OnContinueExecution(0, NULL, 0);
				DestroyTaskGroups();
			}
			return;
		}
	}
	else
		current_looked_TaskGroup = beg_TaskGroups;

	/* start the thread that will check the breakpoints */
	sem_wait_not_intr(&sem_IsStopHere_busy);
	sem_post(&sem_IsStopHere_has_work);
}			  
/********* This function prepare task groups for
                checking and reporting *******************/
static void CreateTasksGroups(BOOL for_check /* or for report */)
{
	int i;

	DestroyTaskGroups();
	for(i = 0; i < number_of_norma_tasks; i++)
	{
		if(task_wait_answer[i])
		{
			if(task_wait_answer[i] != 3)
			{
				TaskGroups *tek_TaskGroups = beg_TaskGroups;
				while(tek_TaskGroups)
				{
					int index = tek_TaskGroups->task_numbers[0];
					if((cur_part_info[index]->current_operator_line_begin ==
						cur_part_info[i]->current_operator_line_begin) &&
						(strcmp(cur_part_info[index]->filename, cur_part_info[i]->filename) == 0))
					{ /* add to existing group */
						tek_TaskGroups->task_numbers[tek_TaskGroups->amount++] = i;
						break;
					}
					tek_TaskGroups = tek_TaskGroups->next;
				}
				if(tek_TaskGroups == NULL) /* new group */
				{
					if(((tek_TaskGroups = malloc(sizeof(TaskGroups))) == NULL) ||
						((tek_TaskGroups->task_numbers = malloc(number_of_norma_tasks * sizeof(int))) == NULL))
					{
						out_memory();
						return;
					}
					tek_TaskGroups->amount = 1;
					tek_TaskGroups->task_numbers[0] = i;
					tek_TaskGroups->IDs = NULL;
					tek_TaskGroups->next = beg_TaskGroups;
					beg_TaskGroups = tek_TaskGroups;
				}
			}
		}
//		else /* task_wait_answer[i] == 0 */
//			if((part_wait_cont_signal[i] != 1) && (part_wait_cont_signal[i] != 2))
//				part_wait_cont_signal[i] = 4; /* we havn't waited it */
	}

	if(for_check && (beg_TaskGroups == NULL))
		return;

	tasksGroupsCreated = TRUE;
	tasksGroupsActual = TRUE;
	current_looked_TaskGroup = NULL;
#ifdef DEBUG
	fputs("TaskGroups created\n", stderr);
#endif
	if(for_check)
		CheckTasksGroup();
	else
		ReportTasksGroups();
}
/************ This function is callback for wait timer *******************/
#ifdef WIN32
static VOID CALLBACK PWaitTimerElapsed(
  HWND hwnd,         // handle to window
  UINT uMsg,         // WM_TIMER message
  UINT idEvent,  // timer identifier
  DWORD dwTime       // current system time
)
#else /* UNIX */
void PWaitTimerElapsed(int idEvent)
#endif
{
#ifdef DEBUG
	fprintf(stderr, "PWaitTimerElapsed: pwait_timer: %d, idEvent: %d\n", pwait_timer, idEvent);
#endif
	assert(idEvent == pwait_timer);
	StopPWaitTimer();
#ifdef WIN32
	if(WaitForSingleObject(main_thread_sem_busy, 0) != WAIT_OBJECT_0)
	{/* We can't take the semaphore for main thread. This means that any another thread already
	    has taken it - we should release execution thread and try again little bit later */
		StartPWaitTimer(1);
		return;
	}
#else /* UNIX */
	if(sem_trywait(&main_thread_sem_busy) != 0)
	{ /* Semaphore is busy. We can't wait it on UNIX, because we is in the same context as main thread */
		StartPWaitTimer(0);
		return;
	}
#endif
	active_thread_recv = -1; /* means that it isn't receive, but the timer elapsed */
	sem_post(&main_thread_sem_has_work);
#ifdef WIN32
	PostThreadMessage(main_thread_ID, event_message, (WPARAM) NULL, (LPARAM) NULL);
#endif
}

/****************** This function starts wait timer *******************/
static void StartPWaitTimer(UINT value)
{
#ifndef WIN32 /* UNIX */
	static BOOL was_inited = FALSE;
	struct itimerval timer_val;
#endif

#ifdef DEBUG
	fprintf(stderr, "StartPWaitTimer\n");
#endif

#ifdef WIN32
	if((pwait_timer = SetTimer(NULL, 0, value, PWaitTimerElapsed)) == 0)
		PrintErrorMessage("Can't start timer");
#else /* UNIX */
	if(!was_inited)
	{
		signal(SIGALRM, PWaitTimerElapsed);
		was_inited = TRUE;
	}
	memset(&timer_val, 0, sizeof(struct itimerval));
	if(value == 0)
		timer_val.it_value.tv_usec = 1;
	else
	{
		timer_val.it_value.tv_sec = value / 1000;
		timer_val.it_value.tv_usec = (value % 1000) * 1000;
	}
	if(setitimer(ITIMER_REAL, &timer_val, NULL) != 0)
	{
		pwait_timer = 0;
		PrintErrorMessage("Can't start timer");
	}
	else
		pwait_timer = SIGALRM;
#endif

#ifdef DEBUG
	fprintf(stderr, "pwait_timer: %d, timeout = %d\n", pwait_timer, value);
#endif
}
/****************** This function stops wait timer *******************/
void StopPWaitTimer(void)
{
#ifdef DEBUG
	fprintf(stderr, "StopPWaitTimer: pwait_timer: %d\n", pwait_timer);
#endif

	if(pwait_timer)
	{
#ifndef WIN32 /* UNIX */
		struct itimerval timer_val;
		memset(&timer_val, 0, sizeof(struct itimerval));
		setitimer(ITIMER_REAL, &timer_val, NULL);
#else
		KillTimer(NULL, pwait_timer);
		pwait_timer = 0;
#endif
#ifdef DEBUG
	fprintf(stderr, "StopPWaitTimer finished\n");
#endif
	}
}
/****************** This function check is all tasks finished any
                    operator and wait the answer from monitor   *******************/
static void CheckTasksState(void)
{
	/* stop wait timer */
	StopPWaitTimer();

	if(task_groups_shall_be_reported)
	{
		DestroyTaskGroups();
		StartPWaitTimer(0);
		return;
	}

	if(tasksGroupsCreated)
	{ /* this event come when we already started to check 
		breakpoints - it is necessary to start from beginning */
		tasksGroupsActual = FALSE;
		return;
	}

	StartPWaitTimer(0);
}
/************ End functions that checks the states of tasks *******************/

/****************** Debugged program start Main Part *******************/
static void OnMainPartStart(void)
{
	int i, pos;
	char name[MAX_PART_NAME], filename[MAX_FILE_NAME];
	int start_line;
	long CRC;
	
#ifdef DEBUG
	fputs("OnMainPartStart\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);

	/* search the end of file name */
	i = pos = 0;
	while(task_buffer_recv[active_thread_recv][pos])
	{
		if(task_buffer_recv[active_thread_recv][pos] == '\"')
			if(++i == 2)
				break;
		pos++;
	}
	if(i != 2)
	{
		PrintErrorMessage("error processing OnMainPartStart:");
		return;
	}
	task_buffer_recv[active_thread_recv][pos] = 0;
		
	if((sscanf(task_buffer_recv[active_thread_recv],
		"%s \"%s", 
		name, filename) != 2) ||
		(sscanf(&task_buffer_recv[active_thread_recv][pos + 1],
		" %d %ld", 
		&start_line, &CRC) != 2))
	{
		PrintErrorMessage("error processing OnMainPartStart:");
		return;
	}

#ifdef DEBUG
	fprintf(stderr, "filename: %s\nMessage: %s\n", filename, task_buffer_recv[active_thread_recv]);
#endif
	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;

	CheckFileCRC(filename, CRC);
	main_part_thread_ID = active_thread_recv;
	StartNewPart(name, filename, active_thread_recv, FALSE);
	cur_part_info[active_thread_recv]->current_operator_line_begin =
		cur_part_info[active_thread_recv]->current_operator_line_end = start_line;
	
	task_wait_answer[active_thread_recv] = 1;
	CheckTasksState();
}
/****************** Debugged program end Main Part *******************/
static void OnMainPartEnd(void)
{
	int end_line;
#ifdef DEBUG
	fputs("OnMainPartEnd\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv] != NULL);
	
	if(sscanf(task_buffer_recv[active_thread_recv],	"%d", &end_line) != 1)
		PrintErrorMessage("error processing OnMainPartEnd");

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;

	RotateOperartorInfo(cur_part_info[active_thread_recv], 0, end_line, end_line);

	/* set that the task wait the answer from UI */
	task_wait_answer[active_thread_recv] = 5;
	CheckTasksState();
}

/*--------------------------------------------------------*/
/* This functions sends "Continue" to the active task */
static void ContinueActiveTask(void)
{
	sem_wait_not_intr(&send_thread_sem[active_thread_recv]);
	assert(task_buffer_send[active_thread_recv] == NULL);
	mess_id_send[active_thread_recv] = ND_CONTINUE_EXECUTION;
	sem_post(&send_thread_sem_has_work[active_thread_recv]);
}
/****************** Quick test is breakpoint may occure *******************/
/* test the active task state and return TRUE if any break may occure and FALSE if may noy */
static BOOL QuickBreakCheck(void)
{
	Breakpoint *tek_breakpoint;
//	return TRUE;
	if(cur_part_info[active_thread_recv]->is_step_into || cur_part_info[active_thread_recv]->is_step_over)
		return TRUE;
	tek_breakpoint = beg_breakpoint;
	while(tek_breakpoint)
	{
		if(tek_breakpoint->filename[0]) /* breakpoin related any place in file */
		{
			if((tek_breakpoint->line >= cur_part_info[active_thread_recv]->current_operator_line_begin) &&
				(tek_breakpoint->line <= cur_part_info[active_thread_recv]->current_operator_line_end) &&
				(!((tek_breakpoint->line >= cur_part_info[active_thread_recv]->previous_operator_line_begin) &&
				(tek_breakpoint->line <= cur_part_info[active_thread_recv]->previous_operator_line_end))) &&
				(strcasecmp(tek_breakpoint->filename, cur_part_info[active_thread_recv]->filename) == 0))
			/* the place matched */
				return TRUE;
		}
		else /* breakpoin not related any place in file */
			return TRUE;
		tek_breakpoint = tek_breakpoint->next;
	}
	return FALSE;
}

/****************** Debugged program start DISTRIBUTION Part *******************/
static void OnDistributionPartStart(void)
{
	static int number_started_parts = 0;
	static int expected_number_of_started_parts = -1;
	char name[MAX_PART_NAME], filename[MAX_FILE_NAME],
		iname[MAX_VARIABLE_NAME], jname[MAX_VARIABLE_NAME];
	int start_line;
	long CRC;
	int i, j, myi, myj, ipoints, jpoints;

	int pos;;

#ifdef DEBUG
	fputs("OnDistributionPartStart\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);
	
	/* search the end of file name */
	i = pos = 0;
	while(task_buffer_recv[active_thread_recv][pos])
	{
		if(task_buffer_recv[active_thread_recv][pos] == '\"')
			if(++i == 2)
				break;
		pos++;
	}
	if(i != 2)
	{
		PrintErrorMessage("error processing OnDistributionPartStart:");
		return;
	}
	task_buffer_recv[active_thread_recv][pos] = 0;
		
	if((sscanf(task_buffer_recv[active_thread_recv],
		"%s \"%s", 
		name, filename) != 2) ||
		(sscanf(&task_buffer_recv[active_thread_recv][pos + 1],
		" %d %ld %d %d %d %d %d %d %s %s", 
		&start_line, &CRC, &i, &j, 
		&myi, &myj, &ipoints, &jpoints, iname, jname) != 10))
	{
		PrintErrorMessage("error processing OnDistributionPartStart:");
		return;
	}

#ifdef DEBUG
	fprintf(stderr, "iname = %s, jname = %s\n", iname, jname);
#endif

	StartNewPart(name, filename, active_thread_recv, FALSE);
	cur_part_info[active_thread_recv]->i = i;
	cur_part_info[active_thread_recv]->j = j;
	cur_part_info[active_thread_recv]->myi = myi;
	cur_part_info[active_thread_recv]->myj = myj;
	cur_part_info[active_thread_recv]->ipoints = ipoints;
	cur_part_info[active_thread_recv]->jpoints = jpoints;
	assert(strlen(iname) < MAX_VARIABLE_NAME);
	assert(strlen(jname) < MAX_VARIABLE_NAME);
	strcpy(cur_part_info[active_thread_recv]->iname, iname);
	strcpy(cur_part_info[active_thread_recv]->jname, jname);
	cur_part_info[active_thread_recv]->current_operator_line_begin =
		cur_part_info[active_thread_recv]->current_operator_line_end = start_line;
	CheckFileCRC(filename, CRC);

	if(number_started_parts == 0)
	{ /* the first process from group */
		assert(expected_number_of_started_parts == -1);
		expected_number_of_started_parts = i * j;
		assert(expected_number_of_started_parts < number_of_norma_tasks);
	}

	if(++number_started_parts == expected_number_of_started_parts)
	{ /* all parts have started */
		number_started_parts = 0;
		expected_number_of_started_parts = -1;
	}
	cur_part_info[active_thread_recv]->is_step_over = FALSE;
	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;
	if(QuickBreakCheck())
	{
		/* set that the task wait the answer from UI or from Monitor */
		task_wait_answer[active_thread_recv] = 1;
		CheckTasksState();
	}
	else
		ContinueActiveTask();
}

/****************** Debugged program ends DISTRIBUTION Part *******************/
static void OnDistributionPartEnd(void)
{
	static int number_ended_parts = 0;
	static int expected_number_of_ended_parts = -1;

	int end_line;

#ifdef DEBUG
	fputs("OnDistributionPartEnd\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv] != NULL);

	if(sscanf(task_buffer_recv[active_thread_recv],	"%d", &end_line) != 1)
		PrintErrorMessage("error processing OnDistributionPartEnd");

	/* set that the current operator was finished */
	RotateOperartorInfo(cur_part_info[active_thread_recv], 0, end_line, end_line);

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;
	/* set that the task wait the answer from UI */
	if(cur_part_info[active_thread_recv]->next)
		task_wait_answer[active_thread_recv] = 2;/* It was Distributed Part */
	else
		task_wait_answer[active_thread_recv] = 4;/* It was Distribution Part */
	CheckTasksState();
}

/****************** Debugged program start Scalar Part *******************/
static void OnScalarPartStart(void)
{
	int i, pos;
	char name[MAX_PART_NAME], filename[MAX_FILE_NAME];
	int start_line, is_External;
	long CRC;

#ifdef DEBUG
	fputs("OnScalarPartStart\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);
	
	/* search the end of file name */
	i = pos = 0;
	while(task_buffer_recv[active_thread_recv][pos])
	{
		if(task_buffer_recv[active_thread_recv][pos] == '\"')
			if(++i == 2)
				break;
		pos++;
	}
	if(i != 2)
	{
		PrintErrorMessage("error processing OnScalarPartStart:");
#ifndef NDEBUG
		PrintErrorMessage(task_buffer_recv[active_thread_recv]);
#endif
		return;
	}
	task_buffer_recv[active_thread_recv][pos] = 0;
		
	if((sscanf(task_buffer_recv[active_thread_recv],
		"%s \"%s", 
		name, filename) != 2) ||
		(sscanf(&task_buffer_recv[active_thread_recv][pos + 1],
		" %d %ld %d", 
		&start_line, &CRC, &is_External) != 3))
	{
		PrintErrorMessage("error processing OnScalarPartStart:");
#ifndef NDEBUG
		PrintErrorMessage(task_buffer_recv[active_thread_recv]);
		PrintErrorMessage(&task_buffer_recv[active_thread_recv][pos + 1]);
#endif
		return;
	}

#ifdef DEBUG
	fprintf(stderr, "OnScalarPartStart: name = %s, filename = %s\n", name, filename);
#endif

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;

	StartNewPart(name, filename, active_thread_recv, is_External);
	cur_part_info[active_thread_recv]->current_operator_line_begin =
		cur_part_info[active_thread_recv]->current_operator_line_end = start_line;
	CheckFileCRC(filename, CRC);

	if(QuickBreakCheck())
	{
		/* set that the task wait the answer from UI or from Monitor */
		task_wait_answer[active_thread_recv] = 1;
		CheckTasksState();
	}
	else
		ContinueActiveTask();

#ifdef DEBUG
	fprintf(stderr, "OnScalarPartStart finished\n");
#endif
}

/****************** Debugged program ends Scalar Part *******************/
static void OnScalarPartEnd(void)
{
	int end_line;

#ifdef DEBUG
	fputs("OnScalarPartEnd\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv] != NULL);

	if(sscanf(task_buffer_recv[active_thread_recv],	"%d", &end_line) != 1)
		PrintErrorMessage("error processing OnScalarPartEnd");

	/* set that the current operator was finished */
	RotateOperartorInfo(cur_part_info[active_thread_recv], 0, end_line, end_line);

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;

	if(QuickBreakCheck())
	{
		/* set that the task wait the answer from UI or from Monitor */
		task_wait_answer[active_thread_recv] = 2;
		CheckTasksState();
	}
	else
	{
		if(EndPart(active_thread_recv)) // tmp_breakpoint set
		{
			/* set that the task wait the answer from UI or from Monitor */
			task_wait_answer[active_thread_recv] = 1;
			CheckTasksState();
		}
		else
			ContinueActiveTask();
	}
}

/***************** UI send continue  ************************/
/* type: what to do with tmp-breakpoint in upper Parts
	0 - Nothing
	1 - Clear
	2 - Set
   tasks_numbers - for what tasks ND_CONTINUE_EXECUTION must be sent
                 = NULL - for all
   number_of_tasks - size of tasks_numbers array
                 = 0 - for all
*/
static void OnContinueExecution(int type, const int *task_numbers, int number_of_tasks)
{
	int i, j, prioritet = 0;
	BOOL was_on_continue = FALSE;
	BOOL is_external_exist = FALSE;

#ifdef DEBUG
	fputs("OnContinueExecution\n", stderr);
#endif

	for(j = 0; j < number_of_norma_tasks; j++)
		if(task_wait_answer[j])
		{
			if(task_wait_answer[j] == 3) /* OnProgramStart */
			{
				prioritet = 1;
				break;
			}
			if(cur_part_info[j]->is_External)
			{
				is_external_exist = TRUE;
				continue;
			}
			if(cur_part_info[j]->number_from_root > prioritet)
				prioritet = cur_part_info[j]->number_from_root;
		}
	if((prioritet == 0) && !is_external_exist)
		return;

	/* Send "continue" to all tasks that wait it */
	for(j = 0; j < number_of_norma_tasks; j++)
		if(task_wait_answer[j] &&
			((task_wait_answer[j] == 3) || 
			(cur_part_info[j]->number_from_root == prioritet) ||
			cur_part_info[j]->is_External))
		{
			if(number_of_tasks)
			{
				for(i = 0; i < number_of_tasks; i++)
					if(j == task_numbers[i]) /* this task should get this command */
						break;
				if(i == number_of_tasks)
					continue; /* this task should not get this command */
			}
			if((task_wait_answer[j] == 4) && !was_on_continue)
			{
				/* notify main part that it may go ahead */
				OnContinueSignal(1);
				was_on_continue = TRUE;
			}
			
			if(task_wait_answer[j] == 5)
				/* notify all slave processes that they may go to receive MPIFLG */
				OnContinueSignal(0);

			if((task_wait_answer[j] == 2) || (task_wait_answer[j] == 4) || (task_wait_answer[j] == 5))
			{
				/* it was waiting on the end of a part */
				if(EndPart(j) && (type == 0)) // was tmp_breakpoint
				{
					/* set that the task wait the answer from UI or from Monitor */
					task_wait_answer[j] = 1;
					CheckTasksState();
				}
			}

			sem_wait_not_intr(&send_thread_sem[j]);
			assert(task_buffer_send[j] == NULL);
			mess_id_send[j] = ND_CONTINUE_EXECUTION;
			sem_post(&send_thread_sem_has_work[j]);
			task_wait_answer[j] = 0;

			if((type > 0) && cur_part_info[j])
			{
				PartInfo *part_info = cur_part_info[j]->next;
				if(type == 2) /* set tmp_breakpoint */
				{
					if(part_info)
						part_info->is_tmp_breakpoint = TRUE;
					else if(!is_fortran_program && (j != main_part_thread_ID))
						cur_part_info[main_part_thread_ID]->is_tmp_breakpoint = TRUE;
				}
				else /* clear tmp_breakpoint */
					while(part_info)
					{
						part_info->is_tmp_breakpoint = FALSE;
						part_info = part_info->next;
					}
			}
		}
}

/*--------------------------------------------------------*/
/* This functions moves the values from source to dest, if dest has no values */
static void MoveNewValues(ValuesProcArray *dest, ValuesProcArray *source)
{
	int i;
	assert(dest->proc_amount == source->proc_amount);
	for(i = 0; i < dest->proc_amount; i++)
	{
		if((dest->array[i].amount == 0) && source->array[i].amount)
		{
			dest->array[i].amount = source->array[i].amount;
			dest->array[i].values = source->array[i].values;
			source->array[i].amount = 0;
			source->array[i].values = NULL;
		}
	}
}
/*--------------------------------------------------------*/
/* This function checks is the value in placeless breakpoint changed
   return value:
   0 - old and new values not comparable
   1 - values not equal
   2 - values equal
   3 - all old values equal the corresponded new ones, but new has some extra values
*/
static int BreakpointValueChanged(ValuesProcArray *vnew, ValuesProcArray *vold)
{
	int i, j;
	BOOL need_copy = FALSE;

	if(vnew->proc_amount != vold->proc_amount)
		return 0; /* not comparable => not changed */
	for(i = 0; i < vnew->proc_amount; i++)
	{
		if(vnew->array[i].amount)
		{
			if(vold->array[i].amount)
			{
				/* Both have the values - they should be equal */
				if(vnew->array[i].amount != vold->array[i].amount)
					 return 0; /* not comparable => not changed */
				for(j = 0; j < vnew->array[i].amount; j++)
					switch(ValueEqual(&vnew->array[i].values[j], &vold->array[i].values[j]))
					{
					case 0: /* equal */
						break;
					case 1:
						return 1; /* not equal */
					case 2:
						return 0; /* not comparable */
					}
			}
			else
				need_copy = TRUE;
		}
	}
	return need_copy ? 3 : 2;
}

static BOOL RemoveBreakpoint(int ID);
	
/****************** Is necessary to stop before the operator *******************/
/* return value:
	NULL - don't stop
	value[0] == 0 - stop not related any breakpoint
	values > 0 - breakpoint's IDs
	value[n] == 0 - end of list

	value should be free by calling function
*/
static int* IsStopHere(int current_operator_line_begin,
					   int current_operator_line_end,
					   int previous_operator_line_begin,
					   int previous_operator_line_end,
					   const PartInfo *part_info)
{
	Breakpoint *tek_breakpoint;
	int *IDs, matched_count = 0;
	ValuesProcArray *value;
	int number_of_values;
	char *error_message;
	BOOL was_error = FALSE;

	if((IDs = malloc(sizeof(int) * (breakpoints_count + 1))) == NULL)
	{
		out_memory();
		return NULL;
	}

	if((part_info->is_step_into || part_info->is_step_over) &&
		(current_operator_line_begin != previous_operator_line_begin))
	{
		IDs[0] = 0;
		return IDs;
	}

	tek_breakpoint = beg_breakpoint;
	while(tek_breakpoint)
	{
#ifdef DEBUG
		fprintf(stderr, "IsStopHere: check N %d\n",
			tek_breakpoint->id);
#endif
		if(tek_breakpoint->filename[0]) /* breakpoin related any place in file */
		{
			if((tek_breakpoint->line >= current_operator_line_begin) &&
				(tek_breakpoint->line <= current_operator_line_end) &&
				(!((tek_breakpoint->line >= previous_operator_line_begin) &&
				(tek_breakpoint->line <= previous_operator_line_end))) &&
				(strcasecmp(tek_breakpoint->filename, part_info->filename) == 0))
			{/* the place matched */
				if(tek_breakpoint->condition)
				{
					if((value = CalculateFormula(tek_breakpoint->condition, &error_message)) == NULL)
						number_of_values = 0;
					else
					{
						int i;

						number_of_values = 1;	
						for(i = 0; i < value->proc_amount; i++)
							if(value->array[i].amount != 1)
							{
								number_of_values = value->array[i].amount;
								break;
							}
					}

					if(number_of_values != 1)
					{ /* error */
						/* send UI notice that error when calculate */
						if(thread_should_stops[number_of_norma_tasks])
							return NULL;
						sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
						assert(task_buffer_send[number_of_norma_tasks] == NULL);
						if((task_buffer_send[number_of_norma_tasks] = malloc(
							value ? 200 : strlen(error_message) + 50)) == NULL)
						{
							out_memory();
							free(IDs);
							return NULL;
						}

						sprintf(task_buffer_send[number_of_norma_tasks],
							"%d -1 ", tek_breakpoint->id);
						if(value == NULL)
							sprintf(&task_buffer_send[number_of_norma_tasks][strlen(task_buffer_send[number_of_norma_tasks])],
								"%s", tek_breakpoint->id, error_message);
						else
						{
							sprintf(&task_buffer_send[number_of_norma_tasks][strlen(task_buffer_send[number_of_norma_tasks])],
								"Variable %s has more than 1 value and can't be used in conditions instead of argument of 'MIN' or 'MAX' function",
								tek_breakpoint->condition->item.variable->name);
							FreeValuesProcArray(value);
						}

						mess_id_send[number_of_norma_tasks] = ND_BREAKPOINT_DYNAMIC_CONDITION_ERROR;
						sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
						was_error = TRUE;
						RemoveBreakpoint(tek_breakpoint->id);
					}
					else
					{
						if(IsAnyValueTrue(value)) /* condition is true too */
							IDs[matched_count++] = tek_breakpoint->id;
						FreeValuesProcArray(value);
					}
				}
				else /* no conditions */
					IDs[matched_count++] = tek_breakpoint->id;
			}/* the place matched */
		}
		else /* breakpoin not related any place in file */
		{
			if((value = CalculateFormula(tek_breakpoint->condition, &error_message)) == NULL)
			{ /* error */
				/* Don't send UI notice that error when calculate
				   because it may be the error in current context
				   and no problem into the another one */
			}
			else
			{
				if(tek_breakpoint->line == 0) /* break when expr changes */
				{
					if(tek_breakpoint->old_value)
					{
						switch(BreakpointValueChanged(value, tek_breakpoint->old_value))
						{ 
						case 0: /* old and new values not comparable */
						case 2: /* values equal */
							FreeValuesProcArray(value);
							break;

						case 1: /* values not equal */
							IDs[matched_count++] = tek_breakpoint->id;
							FreeValuesProcArray(tek_breakpoint->old_value);
							tek_breakpoint->old_value = value;
							break;
						case 3: /* all old values equal the corresponded new ones, but new has some extra values */
							MoveNewValues(tek_breakpoint->old_value, value);
							FreeValuesProcArray(value);
							break;
						}
					}
					else
						tek_breakpoint->old_value = value;
				}
				else /* break when expr is true */
				{
					if(IsAnyValueTrue(value)) /* condition is true too */
						IDs[matched_count++] = tek_breakpoint->id;
					FreeValuesProcArray(value);
				}
			}
		} /* breakpoin (not) related any place in file */
		tek_breakpoint = tek_breakpoint->next;
	}

#ifdef DEBUG
	fprintf(stderr, "IsStopHere: is_step_into: %d, is_step_over: %d match: %d\n",
		part_info->is_step_into, part_info->is_step_over, matched_count);
#endif

	if((matched_count > 0) || was_error)
		IDs[matched_count++] = 0;

	if(matched_count == 0)
	{
		free(IDs);
		return NULL;
	}
	return IDs;
}
/****************** Setting the breakpoint *******************/
static THREADRETURN thread_CalculateBrInitCond(void *arg)
{
	Breakpoint *breakpoint = (Breakpoint *)arg;
	char *error_message;

	sem_wait_not_intr(&calc_ready_sem);
#ifdef DEBUG_CALC_SEM
	fprintf(stderr, "sem_wait_not_intr(&calc_ready_sem); on line %d\n", __LINE__);
#endif
	breakpoint->old_value = CalculateFormula(breakpoint->condition, &error_message);
	/* set that we are ready to calculate more */
	sem_post(&calc_ready_sem);
#ifdef DEBUG_CALC_SEM
	fprintf(stderr, "sem_post(&calc_ready_sem); on line %d\n", __LINE__);
#endif
}

static void OnSetBreakpoint(void)
{
	char filename[MAX_FILE_NAME];
	Formula *condition;
	int error_pos;
	char *error_message;
	int i, start, pos, line, ID, is_any = 0;
	Breakpoint *new_breakpoint;

	assert(active_thread_recv == number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);

	if((sscanf(task_buffer_recv[active_thread_recv], "%d %d %d", &ID, &line, &is_any) < 2) || (ID < 1))
		PrintErrorMessage("error processing OnSetBreakpoint");
	/* search the end of file name */
	i = pos = 0;
	while(task_buffer_recv[active_thread_recv][pos])
	{
		if(task_buffer_recv[active_thread_recv][pos] == '\"')
		{
			if(++i == 1)
				start = pos + 1;
			else
				break;
		}
		pos++;
	}
	if(i != 2)
	{
		PrintErrorMessage("error processing OnSetBreakpoint");
		return;
	}
	if(start != pos)
		memcpy(filename, &task_buffer_recv[active_thread_recv][start], pos - start);
	filename[pos - start] = '\0';

	/* look for condition */
	pos++;
	while(task_buffer_recv[active_thread_recv][pos] == ' ')
		pos++;
	if(task_buffer_recv[active_thread_recv][pos])
	{
		condition = BuildFormula(&task_buffer_recv[active_thread_recv][pos],
			strlen(&task_buffer_recv[active_thread_recv][pos]),
			&error_pos, &error_message);
		if(condition == NULL)
		{
			assert(error_message);
			sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
			assert(task_buffer_send[number_of_norma_tasks] == NULL);
			if((task_buffer_send[number_of_norma_tasks] = malloc(
				strlen(error_message) + 50)) == NULL)
			{
				out_memory();
				return;
			}

			sprintf(task_buffer_send[number_of_norma_tasks], "%d %d %s", 
				ID, error_pos, error_message);
			mess_id_send[number_of_norma_tasks] = ND_BREAKPOINT_CONDITION_ERROR;
			sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
			return;
		}
	}
	else
		condition = NULL;


#ifdef DEBUG
	fprintf(stderr, "OnSetBreakpoint: ID: %d, Filename: %s, line: %d\n",
		ID, filename, line);
#endif
	new_breakpoint = beg_breakpoint;
	while(new_breakpoint)
	{
		if(new_breakpoint->id == ID)
		{ /* such ID exists! */
			sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
			assert(task_buffer_send[number_of_norma_tasks] == NULL);
			mess_id_send[number_of_norma_tasks] = ND_FAIL;
			sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);

			PrintErrorMessage("Breakpoint ID exists!");
			return;
		}
		new_breakpoint = new_breakpoint->next;
	}

	/* add new breakpoint */
	if((new_breakpoint = malloc(sizeof(Breakpoint))) == NULL)
	{
		out_memory();
		return;
	}
	assert(strlen(filename) < MAX_FILE_NAME);
	new_breakpoint->id = ID;
	strcpy(new_breakpoint->filename, filename);
	new_breakpoint->line = line;
	new_breakpoint->is_any = is_any;
	new_breakpoint->condition = condition;
	new_breakpoint->next = beg_breakpoint;
	new_breakpoint->old_value = NULL;
	if((filename[0] == 0) && (line == 0)) /* break when expr changed */
	{ /* Condition only - it is necessary to caluculate it */
		pthread_t calc_thread;
		assert(new_breakpoint->condition);
		/* start the calculating thread */
		if(pthread_create(&calc_thread, &attr, 
			thread_CalculateBrInitCond, (void *)new_breakpoint))
		{
			PrintErrorMessage("error_init thread: thread_CalculateBrInitCond");
			return;
		}
	}

	beg_breakpoint = new_breakpoint;
	breakpoints_count++;

	sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
	assert(task_buffer_send[number_of_norma_tasks] == NULL);
	mess_id_send[number_of_norma_tasks] = ND_OK;
	sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
}

static BOOL RemoveBreakpoint(int ID)
{
	Breakpoint *tek_breakpoint = beg_breakpoint, *pred_breakpoint = NULL;
	
	while(tek_breakpoint)
	{
		if(tek_breakpoint->id == ID)
		{
			if(pred_breakpoint)
				pred_breakpoint->next = tek_breakpoint->next;
			else
				beg_breakpoint = tek_breakpoint->next;
			if(tek_breakpoint->condition)
				FreeFormula(tek_breakpoint->condition);
			FreeValuesProcArray(tek_breakpoint->old_value);
			free(tek_breakpoint);
			breakpoints_count--;
			return TRUE;
		}
		pred_breakpoint = tek_breakpoint;
		tek_breakpoint = tek_breakpoint->next;
	}
	return FALSE;
}

static void OnRemoveBreakpoint(void)
{
	int ID;

	if(sscanf(task_buffer_recv[active_thread_recv],	"%d", &ID) != 1)
		PrintErrorMessage("error processing OnRemoveBreakpoint");
	if(!RemoveBreakpoint(ID))
		PrintErrorMessage("error processing OnRemoveBreakpoint: no such breakpoint");
}
static void RemoveAllBreakpoints(void)
{
	Breakpoint *tek_breakpoint = beg_breakpoint, *pred_breakpoint;
	while(tek_breakpoint)
	{
		pred_breakpoint = tek_breakpoint;
		tek_breakpoint = tek_breakpoint->next;
		if(pred_breakpoint->condition)
			FreeFormula(pred_breakpoint->condition);
		free(pred_breakpoint);
	}
	breakpoints_count = 0;
}
/****************** Debugged program before exec any Norma operator *******************/
/* thread that will calculate the formula for the breakpoint condition checking */
static THREADRETURN thread_IsStopHere(void *arg)
{
	int task_number;

	while(1)
	{
		sem_post(&sem_IsStopHere_busy); /* set that we are free */
		sem_wait_not_intr(&sem_IsStopHere_has_work);
		if(current_looked_TaskGroup == NULL)
		{ /* Exit */
			sem_post(&sem_IsStopHere_busy);
			return;
		}
		sem_wait_not_intr(&calc_ready_sem); /* wait if any process calculate formula */
	#ifdef DEBUG_CALC_SEM
		fprintf(stderr, "sem_wait_not_intr(&calc_ready_sem); on line %d\n", __LINE__);
	#endif

		task_number = current_looked_TaskGroup->task_numbers[0];
		current_looked_TaskGroup->IDs = IsStopHere(
			cur_part_info[task_number]->current_operator_line_begin,
			cur_part_info[task_number]->current_operator_line_end,
			cur_part_info[task_number]->previous_operator_line_begin,
			cur_part_info[task_number]->previous_operator_line_end,
			cur_part_info[task_number]);
		/* set that we are ready to calculate more */
		sem_post(&calc_ready_sem);
	#ifdef DEBUG_CALC_SEM
		fprintf(stderr, "sem_post(&calc_ready_sem); on line %d\n", __LINE__);
	#endif
		/* say for main thread that we should check the another group */
		sem_wait_not_intr(&main_thread_sem_busy);
		active_thread_recv = -2; /* means that it isn't receive, but we should check the next TasksGroup */
		sem_post(&main_thread_sem_has_work);
#ifdef WIN32
		PostThreadMessage(main_thread_ID, event_message, (WPARAM) NULL, (LPARAM) NULL);
#endif
	}
}

static void OnOperatorStart(void)
{
	int previous_operator,
		previous_operator_line_begin, 
		previous_operator_line_end, 
		current_operator,
		current_operator_line_begin,
		current_operator_line_end;

#ifdef DEBUG
	fputs("OnOperatorStart\n", stderr);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);
	
	if(sscanf(task_buffer_recv[active_thread_recv], "%d %d %d %d %d %d",
		&previous_operator,
		&previous_operator_line_begin, 
		&previous_operator_line_end, 
		&current_operator,
		&current_operator_line_begin,
		&current_operator_line_end) != 6)
		PrintErrorMessage("error processing OnOperatorStart");

#ifdef DEBUG
	fprintf(stderr, "OnOperatorStart :%d thread: %d\n", 
		current_operator, active_thread_recv);
#endif

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;
	cur_part_info[active_thread_recv]->previous_operator = previous_operator;
	cur_part_info[active_thread_recv]->previous_operator_line_begin = previous_operator_line_begin;
	cur_part_info[active_thread_recv]->previous_operator_line_end = previous_operator_line_end;
	cur_part_info[active_thread_recv]->current_operator = current_operator;
	cur_part_info[active_thread_recv]->current_operator_line_begin = current_operator_line_begin;
	cur_part_info[active_thread_recv]->current_operator_line_end = current_operator_line_end;
	if(QuickBreakCheck())
	{
		/* set that the task wait the answer from UI or from Monitor */
		task_wait_answer[active_thread_recv] = 1;
		CheckTasksState();
	}
	else
		ContinueActiveTask();
}	

/****************** Debugged program before exec any Fortran operator *******************/
static void OnExOperatorStart(void)
{
	int is_potentially_lock,
		previous_operator_line_begin, 
		previous_operator_line_end, 
		current_operator_line_begin,
		current_operator_line_end;

#ifdef DEBUG
/*	fputs("OnExOperatorStart\n", stderr);*/
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);
	assert(cur_part_info[active_thread_recv] && cur_part_info[active_thread_recv]->is_External);
	
	if(sscanf(task_buffer_recv[active_thread_recv], "%d %d %d %d %d",
		&is_potentially_lock,
		&previous_operator_line_begin, 
		&previous_operator_line_end, 
		&current_operator_line_begin,
		&current_operator_line_end) != 5)
	{
		PrintErrorMessage("error processing OnExOperatorStart");
		return;
	}

#ifdef DEBUG
/*	fprintf(stderr, "OnExOperatorStart line: %d thread: %d\n", 
		current_operator_line_begin, active_thread_recv);*/
#endif

	free(task_buffer_recv[active_thread_recv]);
	task_buffer_recv[active_thread_recv] = NULL;

	cur_part_info[active_thread_recv]->previous_operator_line_begin = previous_operator_line_begin;
	cur_part_info[active_thread_recv]->previous_operator_line_end = previous_operator_line_end;
	cur_part_info[active_thread_recv]->current_operator_line_begin = current_operator_line_begin;
	cur_part_info[active_thread_recv]->current_operator_line_end = current_operator_line_end;

/*	if(is_potentially_lock)
		part_wait_cont_signal[active_thread_recv] = 3;
	else
		part_wait_cont_signal[active_thread_recv] = 0;*/

	if(QuickBreakCheck())
	{
		/* set that the task wait the answer from UI or from Monitor */
		task_wait_answer[active_thread_recv] = 1;
		CheckTasksState();
	}
	else
		ContinueActiveTask();
}	
/****************** Debugged program started *******************/
static void OnProgramStart(void)
{
	static int number_started_programs = 0;
	char filename[MAX_FILE_NAME], programname[MAX_PART_NAME];
	long CRC;
	int i, j, start_line;
#ifdef DEBUG
	fprintf(stderr, "OnProgramStart %d\n", active_thread_recv);
#endif

	assert(active_thread_recv < number_of_norma_tasks);
	assert(task_buffer_recv[active_thread_recv]);
	number_started_programs++;
	if((number_started_programs == number_of_norma_tasks) || is_fortran_program)
	{
		for(i = j = 0; task_buffer_recv[active_thread_recv][i] && (task_buffer_recv[active_thread_recv][i] != '\"'); i++)
			if((is_fortran_program) && (task_buffer_recv[active_thread_recv][i] != ' '))
				programname[j++] = task_buffer_recv[active_thread_recv][i];
		if(is_fortran_program)
			programname[j++] = 0;
		if(task_buffer_recv[active_thread_recv][i] != '\"')
		{
			PrintErrorMessage("error processing OnProgramStart");
			return;
		}
		for(j = ++i; task_buffer_recv[active_thread_recv][j] && task_buffer_recv[active_thread_recv][j] != '\"'; j++);
		if(task_buffer_recv[active_thread_recv][j] != '\"')
		{
			PrintErrorMessage("error processing OnProgramStart");
			return;
		}
		assert(j - i < MAX_FILE_NAME);
		memcpy(filename, &task_buffer_recv[active_thread_recv][i], j - i);
		filename[j - i] = 0;
		if(sscanf(&task_buffer_recv[active_thread_recv][j + 1], " %d %ld", &start_line, &CRC) != 2)
		{
			PrintErrorMessage("error processing OnProgramStart");
			return;
		}

		if(number_started_programs == number_of_norma_tasks)
		{ /* all programs have started */
			/* send UI information */
			/* Take semaphore that we may use the send buffer */	
			sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
			/* be sure that there is no data in buffer */
			assert(task_buffer_send[number_of_norma_tasks] == NULL);
			mess_id_send[number_of_norma_tasks] = ND_NOTIF_PROGRAM_START;
			task_buffer_send[number_of_norma_tasks] = task_buffer_recv[active_thread_recv];
			sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
#ifdef DEBUG
		fprintf(stderr, "Before call CheckFileCRC() %d\n", active_thread_recv);
#endif
			CheckFileCRC(filename, CRC);
		}
		else
			free(task_buffer_recv[active_thread_recv]);
	}
	else
		free(task_buffer_recv[active_thread_recv]);

	task_buffer_recv[active_thread_recv] = NULL;

	if(is_fortran_program)
	{
		StartNewPart(programname, filename, active_thread_recv, TRUE);
		cur_part_info[active_thread_recv]->current_operator_line_begin =
			cur_part_info[active_thread_recv]->current_operator_line_end = start_line;
		
		/* set that the task wait the answer from UI */
		task_wait_answer[active_thread_recv] = 1;
		CheckTasksState();
	}
	else
		/* set that the task wait the answer from UI */
		task_wait_answer[active_thread_recv] = 3;
}
/***************** Debugged program finished ************************/
/* returt TRUE when the last task will finish */
static BOOL OnProgramEnd(void)
{
	static int number_stops = 0;

	if(++number_stops == number_of_norma_tasks)
	{ /* all programs have finished */
		/* send UI information */
		/* Take semaphore that we may use the buffer */	
		sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
		assert(task_buffer_send[number_of_norma_tasks] == NULL);
		mess_id_send[number_of_norma_tasks] = ND_NOTIF_PROGRAM_END;
		sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
	}
	thread_should_stops[active_thread_recv] = TRUE;

	FreeVarInfo(global_var_info[active_thread_recv]);
	global_var_info[active_thread_recv] = NULL;

	if(number_stops == number_of_norma_tasks)
		return TRUE;
	else
		return FALSE;
}
/***************** UI send break  ************************/
void OnBreakExecution(TCPIPMessageType msg)
{
	int j;

#ifdef DEBUG
	fprintf(stderr, "OnBreakExecution\n");
#endif

	StopPWaitTimer();
	DestroyTaskGroups();

	/* Send "break" to all tasks that wait the answer */
	for(j = 0; j < number_of_norma_tasks; j++)
	{
		if(!thread_should_stops[j])
		{
			sem_wait_not_intr(&send_thread_sem[j]);
			assert(task_buffer_send[j] == NULL);
			mess_id_send[j] = msg;
			sem_post(&send_thread_sem_has_work[j]);
		}
	}

	StartPWaitTimer(TIMEOUT_TIMER_VALUE);

#ifdef DEBUG
	fprintf(stderr, "OnBreakExecution END\n");
#endif
}	

/***************** Debugged program breaked ************************/
/* returt TRUE when the last task will breaked */
BOOL OnNotifyBreakExecution(void)
{
	int running_task = number_of_norma_tasks;
	int i;

	thread_should_stops[active_thread_recv] = TRUE;

	for(i = 0; i < number_of_norma_tasks; i++)
		if(thread_should_stops[i])
			running_task--;

#ifdef DEBUG
	fprintf(stderr, "OnNotifyBreakExecution: running_task = %d\n", running_task);
#endif

	if(running_task == 0)
		return TRUE;
	else
		return FALSE;
}
/***************** Register variable at the start of the part ************************/
static void	OnRegisterVar(BOOL is_global)
{
	VarInfo *var;

	if((var = malloc(sizeof(VarInfo))) == NULL)
	{
		out_memory();
		return;
	}

	if(sscanf(task_buffer_recv[active_thread_recv], "%s %d %d",
		var->name, &var->type, &var->number_of_dimensions) != 3)
		PrintErrorMessage("error processing OnRegisterVar");
	if(var->type >= CHARACTER_ID_BASE)
	{
		var->len = var->type - CHARACTER_ID_BASE;
		var->type = CHARACTER_T;
	}

	if(var->number_of_dimensions)
	{
		if((var->dimensions = malloc(sizeof(DimInfo) * var->number_of_dimensions)) == NULL)
		{
			out_memory();
			return;
		}
	}
	if(cur_part_info[active_thread_recv] && !is_global)
	{
		var->next = cur_part_info[active_thread_recv]->vars;
		cur_part_info[active_thread_recv]->vars = var;
	}
	else
	{
		var->next = global_var_info[active_thread_recv];
		global_var_info[active_thread_recv] = var;
	}
	if((is_global ||
		(cur_part_info[active_thread_recv] == NULL) ||
		cur_part_info[active_thread_recv]->is_External) &&
		var->number_of_dimensions)
	{ /* external part and var is an array */
		int i, j, k;
		for(i = 0, j = 0; task_buffer_recv[active_thread_recv][i] && (j < 3); i++)
			if(task_buffer_recv[active_thread_recv][i] == ' ')
				j++;
		for(j = 0; j < var->number_of_dimensions; j++)
		{
			if(sscanf(&task_buffer_recv[active_thread_recv][i], "%d %d",
				&var->dimensions[j].start, &var->dimensions[j].end) != 2)
				PrintErrorMessage("error processing OnRegisterVar");
			var->dimensions[j].declared_start = var->dimensions[j].start;
			var->dimensions[j].declared_end = var->dimensions[j].end;
			for(k = 0; task_buffer_recv[active_thread_recv][i] && (k < 2); i++)
				if(task_buffer_recv[active_thread_recv][i] == ' ')
					k++;
		}
	}

	free(task_buffer_recv[active_thread_recv]);
}

static void	OnRegisterDim(void)
{
	DimInfo dim;
	int i;

	assert(cur_part_info[active_thread_recv]);
	if(sscanf(task_buffer_recv[active_thread_recv], "%d %s %d %d %d %d",
		&i, dim.name, &dim.declared_start, &dim.declared_end,
		&dim.start, &dim.end) != 6)
		PrintErrorMessage("error processing OnRegisterDim");
	assert(i >= 0);
	assert(i < cur_part_info[active_thread_recv]->vars->number_of_dimensions);
	memcpy(&cur_part_info[active_thread_recv]->vars->dimensions[i], &dim, sizeof(DimInfo));
	free(task_buffer_recv[active_thread_recv]);
}

/***************** Calculation of expression ************************/

/* Whait is supported in expressions?
Variables : a, B, Hren;
Operators: +-/*, <>=;
Values: Integer and float: 34, 2.56
Index construct: [20], [1..2, k=5..6];
Function: MIN, MAX;
Operation orders: ();
*/

/***** Calculate variable ******/

static BOOL GetAndConvertValue(int task_num, int amount, 
	Value *values, NormaVarType type,
	char **error_message, BOOL retrieve_index_info)
{
	static char msg[MAX_VARIABLE_NAME + 50];
	int l, pos, pos1;
	static char internal_error[] = "Internal Error: Value expected";

	if(thread_should_stops[task_num])
		return FALSE;
	
	assert(task_wait_answer[task_num] && (task_wait_answer[task_num] != 3));

	mess_id_send[task_num] = ND_GET_VALUE;
#ifdef DEBUG
	fprintf(stderr, "GetAndConvertValue before send (%d) amount = %d\n",
		task_num, amount);
#endif
	sem_post(&send_thread_sem_has_work[task_num]);
	/* wait the answer */
	sem_wait_not_intr(&calc_wait_sem);
#ifdef DEBUG
	fprintf(stderr, "GetAndConvertValue after get value\n");
#endif
#ifdef DEBUG
	fprintf(stderr, "GetAndConvertValue: value_received_str: %s\n", value_received_str);
#endif
	if(value_received_error)
	{
		strncpy(msg, value_received_str, sizeof(msg) - 1);
		free(value_received_str);
		*error_message = msg;
#ifdef DEBUG
		fprintf(stderr, "GetAndConvertValue: before return FALSE\n");
#endif
		return FALSE;
	}
	pos = 0;
	for(l = 0; l < amount; l++)
	{
		while(value_received_str[pos] == ' ')
			pos++;
		if(value_received_str[pos] == '\0')
		{
#ifdef DEBUG
			fprintf(stderr, "GetAndConvertValue: l = %d, amount = %d, pos = %d, value_received_str: %s\n",
				l, amount, pos, value_received_str);
#endif
			free(value_received_str);
			*error_message = internal_error;
			return FALSE;
		}
		pos1 = pos;
		if(type == CHARACTER_T)
		{
			assert(value_received_str[pos1] == '\"'); /* char value */
			for(pos1++; value_received_str[pos1]; pos1++)
				if((value_received_str[pos1] == '\\') && value_received_str[pos1 + 1])
					pos1++;
				else if(value_received_str[pos1] == '\"')
					break;
			assert(value_received_str[pos1] == '\"'); /* end of char value */
			pos1++;
		}
		else
		{
			while((value_received_str[pos1] != ' ') &&
				(value_received_str[pos1] != '\0'))
				pos1++;
		}
		value_received_str[pos1] = 0;
		values[l].type = type;
		switch(type)
		{
		case INTEGER_T:
			values[l].value.int_value = atoi(&value_received_str[pos]);
			break;
		case REAL_T:
			sscanf(&value_received_str[pos], "%f", &values[l].value.real_value);
			break;
		case DOUBLE_T:
			sscanf(&value_received_str[pos], "%lf", &values[l].value.double_value);
			break;
		case CHARACTER_T:
			if((values[l].value.char_value = malloc(pos1 - pos + 1)) == NULL)
			{
				free(value_received_str);
				*error_message = OUT_MEMORY;
				return FALSE;
			}
			strcpy(values[l].value.char_value, &value_received_str[pos]);
			break;
		default:
			assert(FALSE);

		}
		pos = pos1 + 1;
		if(retrieve_index_info)
		{
			if(value_received_str[pos] != '(')
			{
				free(value_received_str);
				*error_message = internal_error;
				return FALSE;
			}
			pos1 = ++pos;
			while((value_received_str[pos1] != ')') &&
				(value_received_str[pos1] != '\0'))
				pos1++;
			if(pos1 != pos)
			{
				if((values[l].indexes = malloc(pos1 - pos +1)) == NULL)
				{
					free(value_received_str);
					*error_message = OUT_MEMORY;
					return FALSE;
				}
				memcpy(values[l].indexes, &value_received_str[pos], pos1 - pos);
				values[l].indexes[pos1 - pos] = '\0';
			}
			pos = pos1 + 1;
		}
	}
	free(value_received_str);
	return TRUE;
}

static char NO_CONTEXT[] = "No context";
/* Calculate Fortran-variable */
static ValuesProcArray *CalculateVariableInEachProc(const Variable *variable, char **error_message)
{
	VarInfo *vars = NULL;
	ValuesProcArray *values;
	int i, j, amount;
	Index *curr_index;
	static char msg[100];

	if((values = MakeValuesProcArray(number_of_norma_tasks, 0)) == NULL)
	{
		*error_message = OUT_MEMORY;
		return NULL;
	}

	for(j = 0; j < number_of_norma_tasks; j++)
	{
		if((!task_wait_answer[j]) || (cur_part_info[j] == NULL))
			continue;
		
		vars = cur_part_info[j]->vars;
		while(vars)
		{
			if(strcasecmp(vars->name, variable->name) == 0)
				break;
			vars = vars->next;
		}
		
		if(vars == NULL)
		{
			vars = global_var_info[j];
			while(vars)
			{
				if(strcasecmp(vars->name, variable->name) == 0)
					break;
				vars = vars->next;
			}
			if(vars == NULL)
				continue;
		}

		amount = 1; /* Calculate the amount of values */
		curr_index = variable->index;
		for(i = 0; i < vars->number_of_dimensions; i++)
		{
			if(curr_index == NULL) /* this i index hasn't been specified */
			{
				vars->dimensions[i].requested_start = vars->dimensions[i].declared_start;
				vars->dimensions[i].requested_end = vars->dimensions[i].declared_end;
			}
			else
			{
				if((curr_index->from < vars->dimensions[i].declared_start) ||
					(curr_index->to > vars->dimensions[i].declared_end))
				{
					sprintf(msg, "Index %d..%d out of boundary", curr_index->from, curr_index->to);
					*error_message = msg;
					FreeValuesProcArray(values);
					return NULL;
				}
				vars->dimensions[i].requested_start = curr_index->from;
				vars->dimensions[i].requested_end = curr_index->to;
				curr_index = curr_index->next;
			}
			/* Calculate the amount of values */
			amount = amount * (vars->dimensions[i].requested_end - vars->dimensions[i].requested_start + 1);
		}
		if((values->array[j].values = MakeValues(amount)) == NULL)
		{
			*error_message = OUT_MEMORY;
			FreeValuesProcArray(values);
			return NULL;
		}
		values->array[j].amount = amount;

		sem_wait_not_intr(&send_thread_sem[j]);
		assert(task_buffer_send[j] == NULL);
		if((task_buffer_send[j] = malloc(MAX_VARIABLE_NAME + vars->number_of_dimensions * 20)) == NULL)
		{
			FreeValuesProcArray(values);
			*error_message = OUT_MEMORY;
			return NULL;
		}
		strcpy(task_buffer_send[j], vars->name);
		if(vars->number_of_dimensions)
			strcat(task_buffer_send[j],	" 0");
		for(i = 0; i < vars->number_of_dimensions; i++)
			sprintf(&task_buffer_send[j][strlen(task_buffer_send[j])],
				" %d %d", vars->dimensions[i].requested_start, vars->dimensions[i].requested_end);
		if(GetAndConvertValue(j, 
			values->array[j].amount, 
			values->array[j].values, 
			vars->type,	error_message,
			FALSE) == FALSE)
		{
			FreeValuesProcArray(values);
			return NULL;
		}
	}

	return values;
}
/* Calculate variable */
ValuesProcArray *CalculateVariable(const Variable *variable, char **error_message, BOOL retrieve_index_info)
{
	VarInfo *vars = NULL;
	DimInfo *i_index = NULL, *j_index = NULL;
	int i, j, k, l, prioritet = 0, first_task, amount, value_pos;
	static char msg[MAX_VARIABLE_NAME + 50];
	Index *curr_index;
	ValuesProcArray *values;
	int start_proc_i, end_proc_i, start_i_point, end_i_point;
	int start_proc_j, end_proc_j, start_j_point, end_j_point;
	BOOL individual = FALSE;


#ifdef DEBUG
	fprintf(stderr, "CalculateVariable\n");
#endif

	for(j = 0; j < number_of_norma_tasks; j++)
		if(task_wait_answer[j])
		{
			if(task_wait_answer[j] == 3) /* OnProgramStart */
				break;
			if(cur_part_info[j]->is_External)
				individual = TRUE;
			else if(cur_part_info[j]->number_from_root > prioritet)
			{
				prioritet = cur_part_info[j]->number_from_root;
				first_task = j;
			}
		}
	if(prioritet == 0)
	{
		if(individual) /* Calculate Fortran-variable */
			return CalculateVariableInEachProc(variable, error_message);
		*error_message = NO_CONTEXT;
		return NULL;
	}
	/* Calculate NORMA-variable */
	assert(first_task < number_of_norma_tasks);
	if(cur_part_info[first_task] == NULL)
	{
		*error_message = NO_CONTEXT;
		return NULL;
	}
	vars = cur_part_info[first_task]->vars;

	while(vars)
	{
		if(strcasecmp(vars->name, variable->name) == 0)
			break;
		vars = vars->next;
	}
	if(vars == NULL)
	{
		if(individual) /* Calculate Fortran-variable */
			return CalculateVariableInEachProc(variable, error_message);
		sprintf(msg, "Variable %s doesn't present in context", variable->name);
		*error_message = msg;
		return NULL;
	}

	amount = 1; /* Calculate the amount of values */
	for(i = 0; i < vars->number_of_dimensions; i++)
	{
		curr_index = variable->index;
		while(curr_index)
		{
			if(strcasecmp(vars->dimensions[i].name, curr_index->name) == 0)
				break;
			curr_index = curr_index->next;
		}
		if(curr_index == NULL) /* this i index hasn't been specified */
		{
			vars->dimensions[i].requested_start = vars->dimensions[i].declared_start;
			vars->dimensions[i].requested_end = vars->dimensions[i].declared_end;
		}
		else
		{
			if((curr_index->from < vars->dimensions[i].declared_start) ||
				(curr_index->to > vars->dimensions[i].declared_end))
			{
				sprintf(msg, "Index %s out of boundary", vars->dimensions[i].name);
				*error_message = msg;
				return NULL;
			}
			curr_index->found = TRUE;
			vars->dimensions[i].requested_start = curr_index->from;
			vars->dimensions[i].requested_end = curr_index->to;
		}
		if(cur_part_info[first_task]->i != 0)
		{ /* Distr* part - look the distr indexes */
			if(strcasecmp(vars->dimensions[i].name, cur_part_info[first_task]->iname) == 0)
				i_index = &vars->dimensions[i];
			else if(strcasecmp(vars->dimensions[i].name, cur_part_info[first_task]->jname) == 0)
				j_index = &vars->dimensions[i];
		}
		/* Calculate the amount of values */
		amount = amount * (vars->dimensions[i].requested_end - vars->dimensions[i].requested_start + 1);
	}
	/* check that all indexes that where specified by user were found */
	curr_index = variable->index;
	while(curr_index)
	{
		if(!curr_index->found)
		{
			sprintf(msg, "Index %s doesn't exist", curr_index->name);
			*error_message = msg;
			return NULL;
		}
		curr_index = curr_index->next;
	}
			
	if((values = MakeValuesProcArray(1, amount)) == NULL)
	{
		*error_message = OUT_MEMORY;
		return NULL;
	}

	if((i_index == NULL) && (j_index == NULL))
	{/* not distributed var */
		/* Request Norma task for a value */
		if(thread_should_stops[first_task])
		{
			FreeValuesProcArray(values);
			*error_message = NO_CONTEXT;
			return NULL;
		}

		sem_wait_not_intr(&send_thread_sem[first_task]);
		assert(task_buffer_send[first_task] == NULL);
		if((task_buffer_send[first_task] = malloc(MAX_VARIABLE_NAME +
			vars->number_of_dimensions * 20)) == NULL)
		{
			FreeValuesProcArray(values);
			*error_message = OUT_MEMORY;
			return NULL;
		}
		strcpy(task_buffer_send[first_task], vars->name);
		if(vars->number_of_dimensions)
			sprintf(&task_buffer_send[first_task][strlen(task_buffer_send[first_task])],
				" %d", retrieve_index_info);
		for(i = 0; i < vars->number_of_dimensions; i++)
			sprintf(&task_buffer_send[first_task][strlen(task_buffer_send[first_task])],
				" %d %d", vars->dimensions[i].requested_start, vars->dimensions[i].requested_end);
		if(GetAndConvertValue(first_task, 
			values->array[0].amount, 
			values->array[0].values, 
			vars->type,	error_message,
			retrieve_index_info && (vars->number_of_dimensions > 0)) == FALSE)
		{
			FreeValuesProcArray(values);
			return NULL;
		}
	}
	else
	{
		if(i_index != NULL)
		{
			start_proc_i = ((i_index->requested_start - 1) / cur_part_info[first_task]->ipoints) + 1;
			end_proc_i = ((i_index->requested_end - 1) / cur_part_info[first_task]->ipoints) + 1;
			start_i_point = ((i_index->requested_start - 1) % cur_part_info[first_task]->ipoints) + 1;
			end_i_point = ((i_index->requested_end - 1) % cur_part_info[first_task]->ipoints) + 1;
		}
		else
			start_proc_i = end_proc_i = 1;
		if(j_index != NULL)
		{
			start_proc_j = ((j_index->requested_start - 1) / cur_part_info[first_task]->jpoints) + 1;
			end_proc_j = ((j_index->requested_end - 1) / cur_part_info[first_task]->jpoints) + 1;
			start_j_point = ((j_index->requested_start - 1) % cur_part_info[first_task]->jpoints) + 1;
			end_j_point = ((j_index->requested_end - 1) % cur_part_info[first_task]->jpoints) + 1;
		}
		else
			start_proc_j = end_proc_j = 1;
		value_pos = 0;
		for(i = start_proc_i; i <= end_proc_i; i++)
		{
			for(j = start_proc_j; j <= end_proc_j; j++)
			{
				for(k = 0; k < number_of_norma_tasks; k++)
					if(task_wait_answer[k] &&
						(cur_part_info[k]->number_from_root == prioritet))
					{
						if((cur_part_info[k]->myi == i) &&
							(cur_part_info[k]->myj == j))
							break;
					}
				if(k == number_of_norma_tasks)
				{
					*error_message = CANT_CALC_REQUESTED_TASK_NOT_AVAILABLE;
					FreeValuesProcArray(values);
					return NULL;
				}

				if(thread_should_stops[k])
				{
					*error_message = NO_CONTEXT;
					FreeValuesProcArray(values);
					return NULL;
				}
				sem_wait_not_intr(&send_thread_sem[k]);

				assert(task_buffer_send[k] == NULL);
				if((task_buffer_send[k] = malloc(MAX_VARIABLE_NAME +
					vars->number_of_dimensions * 20)) == NULL)
				{
					FreeValuesProcArray(values);
					*error_message = OUT_MEMORY;
					return NULL;
				}
				strcpy(task_buffer_send[k], vars->name);
				if(vars->number_of_dimensions)
					sprintf(&task_buffer_send[k][strlen(task_buffer_send[k])],
						" %d", retrieve_index_info);
				amount = 1;
				for(l = 0; l < vars->number_of_dimensions; l++)
				{
					int sval, eval;

					if(i_index == &vars->dimensions[l])
					{
						if(i == start_proc_i)
							sval = start_i_point;
						else
							sval = 1;
						if(i == end_proc_i)
							eval = end_i_point;
						else
							eval = cur_part_info[first_task]->ipoints;
					}
					else if(j_index == &vars->dimensions[l])
					{
						if(j == start_proc_j)
							sval = start_j_point;
						else
							sval = 1;
						if(j == end_proc_j)
							eval = end_j_point;
						else
							eval = cur_part_info[first_task]->jpoints;
					}
					else
					{
						sval = vars->dimensions[l].requested_start;
						eval = vars->dimensions[l].requested_end;
					}
					assert(sval >= vars->dimensions[l].start);
					assert(eval <= vars->dimensions[l].end);
					sprintf(&task_buffer_send[k][strlen(task_buffer_send[k])],
						" %d %d", sval, eval);
					amount = amount * (eval - sval + 1);
				}
				if(GetAndConvertValue(k, amount, 
					&(values->array[0].values[value_pos]), vars->type,	error_message, 
					retrieve_index_info && (vars->number_of_dimensions > 0)) == FALSE)
				{
					FreeValuesProcArray(values);
					return NULL;
				}
				value_pos += amount;
			} /* for j */
		} /* for i */
		assert(values->array[0].amount == value_pos);
	}

	return values;
}

/* thread that will calculate the formula */
static THREADRETURN thread_CalculateFormula(void *arg)
{
	ValuesProcArray *value;
	char *error_message = NULL;
	FormulaList *tmp_calc_formula_list;

	while(1)
	{
		sem_wait_not_intr(&sem_calc_formula_list);
		if(beg_calc_formula_list == NULL)
		{/* wait the work inited by external thread */
#ifdef DEBUG
			fprintf(stderr, "thread_CalculateFormula waits the task\n");
#endif
			sem_post(&sem_calc_formula_busy); /* set that we are free */
			sem_post(&sem_calc_formula_list);
			sem_wait_not_intr(&sem_calc_formula_has_work);
			if(beg_calc_formula_list == NULL)
			{ /* Exit */
				sem_post(&sem_calc_formula_busy);
				return;
			}
#ifdef DEBUG
			fprintf(stderr, "thread_CalculateFormula starting the task\n");
#endif
		}
		else
		{
#ifdef DEBUG
			fprintf(stderr, "thread_CalculateFormula get the next formula from queue\n");
#endif
			sem_post(&sem_calc_formula_list);
		}
		assert(beg_calc_formula_list);
		sem_wait_not_intr(&calc_ready_sem);
#ifdef DEBUG_CALC_SEM
		fprintf(stderr, "sem_wait_not_intr(&calc_ready_sem); on line %d\n", __LINE__);
#endif
		value = CalculateFormula(beg_calc_formula_list->formula, &error_message);
		/* set that we are ready to calculate more */
		sem_post(&calc_ready_sem);
	#ifdef DEBUG_CALC_SEM
		fprintf(stderr, "sem_post(&calc_ready_sem); on line %d\n", __LINE__);
	#endif

		if(thread_should_stops[number_of_norma_tasks]) // UI has finished
		{
			FreeValuesProcArray(value);
			sem_wait_not_intr(&sem_calc_formula_list);
			FreeFormulaList(beg_calc_formula_list);
			beg_calc_formula_list = NULL;
			sem_post(&sem_calc_formula_list);
			continue;
		}
		/* send UI calculated value */
		sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
		assert(task_buffer_send[number_of_norma_tasks] == NULL);
		if(value == NULL) /* error when calculate */
		{
			assert(error_message);
			if((task_buffer_send[number_of_norma_tasks] = malloc(
				MAX_NUMBER_LEN + strlen(error_message) + 1)) == NULL)
			{
				out_memory();
				sem_wait_not_intr(&sem_calc_formula_list);
				FreeFormulaList(beg_calc_formula_list);
				beg_calc_formula_list = NULL;
				sem_post(&sem_calc_formula_list);
				continue;
			}

			sprintf(task_buffer_send[number_of_norma_tasks], "%d ", beg_calc_formula_list->ID);
			strcat(task_buffer_send[number_of_norma_tasks], error_message);
			mess_id_send[number_of_norma_tasks] = ND_TAKE_VALUE_ERROR;
		}
		else
		{
			task_buffer_send[number_of_norma_tasks] = PrintValuesProcArray(value, beg_calc_formula_list->ID);
			FreeValuesProcArray(value);
			mess_id_send[number_of_norma_tasks] = ND_TAKE_VALUE;
		}
		sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);

#ifdef DEBUG
		fprintf(stderr, "thread_CalculateFormula has finished one formula processing\n");
#endif

		sem_wait_not_intr(&sem_calc_formula_list);
		tmp_calc_formula_list = beg_calc_formula_list;
		beg_calc_formula_list = beg_calc_formula_list->next;
		sem_post(&sem_calc_formula_list);
		FreeFormulaListItem(tmp_calc_formula_list);
	}
}

/* request from the UI for getting current status - what happenning? */
static void OnGetStatus(void)
{
	char *mes;
	int pos, i;

#ifdef DEBUG
	fputs("OnGetStatus\n", stderr);
#endif
	assert(active_thread_recv == number_of_norma_tasks);

	if((mes = malloc(number_of_norma_tasks * (100 + MAX_FILE_NAME + MAX_PART_NAME))) == NULL)
	{
		out_memory();
		return;
	}

	for(i = 0, pos = 0; i < number_of_norma_tasks; i++)
	{
		if(cur_part_info[i])
		{
			sprintf(&mes[pos], "Task %d now in the Part %s\n", i, cur_part_info[i]->name);
			pos += strlen(&mes[pos]);
			if(cur_part_info[i]->i)
			{
				sprintf(&mes[pos], "This is DISTRIBUTION Part. %s = %d, %s = %d\n",
					cur_part_info[i]->iname, cur_part_info[i]->myi,
					cur_part_info[i]->jname, cur_part_info[i]->myj);
				pos += strlen(&mes[pos]);
			}
			if(cur_part_info[i]->previous_operator)
				sprintf(&mes[pos], "The last operator that was complete is operator N %d from line %d file %s\n",
					cur_part_info[i]->previous_operator, 
					cur_part_info[i]->previous_operator_line_begin,
					cur_part_info[i]->filename);
			else
				sprintf(&mes[pos], "No one operators from this Part were completed\n");
			pos += strlen(&mes[pos]);
			if(cur_part_info[i]->current_operator)
			{
				sprintf(&mes[pos], "Now operator N %d from line %d file %s %s\n",
					cur_part_info[i]->current_operator, 
					cur_part_info[i]->current_operator_line_begin,
					cur_part_info[i]->filename,
					(task_wait_answer[i] ? "is ready to be executed" : "is being execute"));
				pos += strlen(&mes[pos]);
			}
			else if(cur_part_info[i]->previous_operator)
			{
				sprintf(&mes[pos], "The Part is ready to be finished\n");
				pos += strlen(&mes[pos]);
			}
		}
		else
		{
			sprintf(&mes[pos], "Task %d has not started any Part\n", i);
			pos += strlen(&mes[pos]);
		}
		mes[pos++] = '\n';
		mes[pos] = '\0';
	}

	sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
	task_buffer_send[number_of_norma_tasks] = mes;
	mess_id_send[number_of_norma_tasks] = ND_TAKE_STATUS;
	sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
}

/* request from the UI for getting (calculating) value */
static void OnGetValue(void)
{
	Formula *formula;
	int error_pos;
	char *error_message;
	FormulaList *tek_calc_formula_list, *new_calc_formula_list;
	int busy;
	int ID;
	int pos = 0;

#ifdef DEBUG
	fputs("OnGetValue\n", stderr);
#endif

	assert(active_thread_recv == number_of_norma_tasks);
	assert(task_buffer_recv[number_of_norma_tasks]);

	if(sscanf(task_buffer_recv[number_of_norma_tasks], "%d", &ID) == 1)
	{
		while(isdigit(task_buffer_recv[number_of_norma_tasks][pos]))
			pos++;
		formula = BuildFormula(&task_buffer_recv[number_of_norma_tasks][pos],
			strlen(&task_buffer_recv[number_of_norma_tasks][pos]),
			&error_pos, 
			&error_message);
	}
	else
	{
		formula = NULL;
		error_message = "Internal Error";
	}

	free(task_buffer_recv[number_of_norma_tasks]);
	if(formula == NULL) /* error when build formula */
	{ 
		if(thread_should_stops[number_of_norma_tasks])
			return;
		sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
		assert(task_buffer_send[number_of_norma_tasks] == NULL);
		if((task_buffer_send[number_of_norma_tasks] = malloc(MAX_NUMBER_LEN +
			(error_message ? strlen(error_message) : 0) + 50)) == NULL)
		{
			out_memory();
			return;
		}

		if(error_message)
			sprintf(task_buffer_send[number_of_norma_tasks], "%d %d %s", ID, error_pos, error_message);
		else
			sprintf(task_buffer_send[number_of_norma_tasks], "%d %s", ID, "Error: empty formula");

		mess_id_send[number_of_norma_tasks] = ND_TAKE_VALUE_ERROR;
		sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
		return;
	}

	/* start the calculating thread */
	if((new_calc_formula_list = MakeFormulaListItem(formula, ID)) == NULL)
	{
		out_memory();
		return;
	}
#ifdef DEBUG
	fprintf(stderr, "OnGetValue before insert new formula in queue\n");
#endif
	sem_wait_not_intr(&sem_calc_formula_list);
#ifdef DEBUG
	fprintf(stderr, "OnGetValue inserting new formula in queue....\n");
#endif
	tek_calc_formula_list = beg_calc_formula_list;
	while(tek_calc_formula_list && tek_calc_formula_list->next)
		tek_calc_formula_list = tek_calc_formula_list->next;
	if(tek_calc_formula_list)
		tek_calc_formula_list->next = new_calc_formula_list;
	else
		beg_calc_formula_list = new_calc_formula_list;
#ifdef DEBUG
	fprintf(stderr, "OnGetValue finish inserting new formula in queue\n");
#endif
	sem_getvalue(&sem_calc_formula_busy, &busy);
	sem_post(&sem_calc_formula_list);
#ifdef DEBUG
	fprintf(stderr, "OnGetValue sem_calc_formula_busy: %d\n", busy);
#endif
	if(busy == 1)
	{ /* start calc thread */
		sem_wait_not_intr(&sem_calc_formula_busy);
		sem_post(&sem_calc_formula_has_work);
	}
	/* otherwise Calc thread will handle the new element from calc_formula_list when finish process the previous one */
}

static void OnTakeValue(void)
{
#ifdef DEBUG
	fputs("OnTakeValue\n", stderr);
#endif
	value_received_str = task_buffer_recv[active_thread_recv];
	value_received_error = FALSE;
	sem_post(&calc_wait_sem);
}

static void OnErrorTakeValue(void)
{
#ifdef DEBUG
	fputs("OnErrorTakeValue\n", stderr);
#endif
	value_received_str = task_buffer_recv[active_thread_recv];
	value_received_error = TRUE;
	sem_post(&calc_wait_sem);
}

/***************** Get version info ************************/
static void OnGetVersionID(void)
{
	if(active_thread_recv == number_of_norma_tasks)
	{	/* this is the request from UI - ask the 0 task about MonitorConnect ver */
		sem_wait_not_intr(&send_thread_sem[0]);
		assert(task_buffer_send[0] == NULL);
		mess_id_send[0] = ND_GET_VERSION_ID;
		sem_post(&send_thread_sem_has_work[0]);
	}
	else
	{   /* this is the answer from MonitorConnect */
		assert(active_thread_recv == 0);
		sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
		assert(task_buffer_send[number_of_norma_tasks] == NULL);
		if((task_buffer_send[number_of_norma_tasks] = 
			malloc(strlen(task_buffer_recv[0]) + 100)) == NULL)
		{
			out_memory();
			return;
		}

		sprintf(task_buffer_send[number_of_norma_tasks], 
			"Monitor: version %s built at %s.\n%s",
			VERSIONID, __DATE__, task_buffer_recv[0]);

		mess_id_send[number_of_norma_tasks] = ND_GET_VERSION_ID;
		sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
	}
}
/***************** Get Stack info ************************/
static void OnStackInfo(void)
{
	int proc_num;
	int len;

	assert(active_thread_recv == number_of_norma_tasks);
	assert(task_buffer_recv[number_of_norma_tasks]);

	if((sscanf(task_buffer_recv[number_of_norma_tasks], "%d", &proc_num) != 1) ||
		(proc_num < 0) || (proc_num >= number_of_norma_tasks))
	{
		PrintErrorMessage("Error processing OnStackInfo");
		return;
	}
	free(task_buffer_recv[number_of_norma_tasks]);

	sem_wait_not_intr(&send_thread_sem[number_of_norma_tasks]);
	mess_id_send[number_of_norma_tasks] = ND_TAKE_STACK_INFO;
	if(cur_part_info[proc_num])
	{
		char *mes;
		PartInfo *pi = cur_part_info[proc_num];

		len = (pi->number_from_root + 2) * (MAX_PART_NAME + MAX_FILE_NAME + 20);
		if((mes = malloc(len)) == NULL)
		{
			out_memory();
			return;
		}
		
		len = 0;
		while(pi)
		{
			sprintf(&mes[len], " %s \"%s\" %d", pi->name, pi->filename, pi->current_operator_line_begin);
			len += strlen(&mes[len]);
			pi = pi->next;
		}
		if(!is_fortran_program && (proc_num != main_part_thread_ID))
		/* add the main part */
			sprintf(&mes[len], " %s \"%s\" %d", cur_part_info[main_part_thread_ID]->name,
				cur_part_info[main_part_thread_ID]->filename, cur_part_info[main_part_thread_ID]->current_operator_line_begin);
		task_buffer_send[number_of_norma_tasks] = mes;
	}
	else
	{
		assert(task_buffer_send[number_of_norma_tasks] == NULL);
	}
	sem_post(&send_thread_sem_has_work[number_of_norma_tasks]);
}
/***************** convert tasks: char to int representation ************************/
int String2IntArr(const char *str, int **int_arr)
{
	int amount = 0;
	int j = 0;
	if((*int_arr = malloc(strlen(str) * sizeof(int))) == NULL)
	{
		out_memory();
		return 0;
	}
	while(str[j])
	{
		while(str[j] == ' ') j++;
		if(sscanf(&str[j], "%d", *int_arr + amount) != 1)
			break;
		amount++;
		while(isdigit(str[j])) j++;
	}
	if(amount == 0)
		free(*int_arr);
	return amount;
}

/***************** User Actions ************************/
static void OnActionCommand()
{
	int number_of_tasks = 0;
	int *task_numbers;
	int i, j;

#ifdef DEBUG
	switch(mess_id_recv[number_of_norma_tasks])
	{
	case ND_RUN:
		fprintf(stderr, "ND_RUN %s\n", task_buffer_recv[number_of_norma_tasks] ? task_buffer_recv[number_of_norma_tasks] : "");
		break;
	case ND_PAUSE:
		fprintf(stderr, "ND_PAUSE %s\n", task_buffer_recv[number_of_norma_tasks] ? task_buffer_recv[number_of_norma_tasks] : "");
		break;
	case ND_STEP_OVER:
		fprintf(stderr, "ND_STEP_OVER %s\n", task_buffer_recv[number_of_norma_tasks] ? task_buffer_recv[number_of_norma_tasks] : "");
		break;
	case ND_STEP_INTO:
		fprintf(stderr, "ND_STEP_INTO %s\n", task_buffer_recv[number_of_norma_tasks] ? task_buffer_recv[number_of_norma_tasks] : "");
		break;
	case ND_UNTIL_RETURN:
		fprintf(stderr, "ND_UNTIL_RETURN %s\n", task_buffer_recv[number_of_norma_tasks] ? task_buffer_recv[number_of_norma_tasks] : "");
		break;
	}
#endif

	assert(active_thread_recv == number_of_norma_tasks);
	if(task_buffer_recv[number_of_norma_tasks])
	{
		number_of_tasks = String2IntArr(task_buffer_recv[number_of_norma_tasks], &task_numbers);
		free(task_buffer_recv[number_of_norma_tasks]);
	}
	for(i = 0; i < number_of_norma_tasks; i++)
		if(cur_part_info[i])
		{
			if(number_of_tasks)
			{
				for(j = 0; j < number_of_tasks; j++)
					if(task_numbers[j] == i) /* this task should get this command */
						break;
				if(j == number_of_tasks)
					continue;  /* this task should not get this command */
			}
			switch(mess_id_recv[number_of_norma_tasks])
			{
			case ND_RUN:
				cur_part_info[i]->is_step_over = FALSE;
				cur_part_info[i]->is_step_into = FALSE;
				break;
			case ND_PAUSE:
				assert(number_of_tasks == 0);
				cur_part_info[i]->is_step_into = TRUE;
				break;
			case ND_STEP_OVER:
				cur_part_info[i]->is_step_over = TRUE;
				cur_part_info[i]->is_step_into = FALSE;
				break;
			case ND_STEP_INTO:
				cur_part_info[i]->is_step_over = FALSE;
				cur_part_info[i]->is_step_into = TRUE;
				break;
			case ND_UNTIL_RETURN:
				cur_part_info[i]->is_step_over = FALSE;
				cur_part_info[i]->is_step_into = FALSE;
				break;
			}
		}

	switch(mess_id_recv[number_of_norma_tasks])
	{
	case ND_RUN:
	case ND_STEP_OVER:
		is_step_into = FALSE;
		OnContinueExecution(1, task_numbers, number_of_tasks); /* clear tmp_breakpoint */
		break;
	case ND_STEP_INTO:
		is_step_into = TRUE;
		OnContinueExecution(1, task_numbers, number_of_tasks); /* clear tmp_breakpoint */
		break;
	case ND_UNTIL_RETURN:
		is_step_into = FALSE;
		OnContinueExecution(2, task_numbers, number_of_tasks); /* set is_tmp_breakpoint in upper Part */
		break;
	case ND_PAUSE:
		task_groups_shall_be_reported = TRUE;
		StopPWaitTimer();
		StartPWaitTimer(PWAIT_TIMER_VALUE);
		break;
	}

	if(number_of_tasks)
		free(task_numbers);
}

/***************** MAIN LOOP  ************************/
static int main_loop(void)
{
	int i;
	int common_tasks_counter = 0; /* for calculating and synchronization */
	

	pwait_timer = 0;
	tasksGroupsCreated = FALSE;
	task_groups_shall_be_reported = FALSE;
	number_stops = 0; /* 0 - nobody stopped, 1 - tasks, 2 - tasks and UI */
	while(number_stops < 2)
	{
		/* set that we are not busy */
#ifndef NDEBUG
		sem_getvalue(&main_thread_sem_busy, &i);
		assert(i == 0);
#endif
		if(sem_post(&main_thread_sem_busy))
		{
			PrintErrorMessage("error_post sem");
			return 1;
		}

#ifdef WIN32
		{
			MSG msg;
#if 0
			while(1)
			{
				if(WaitForSingleObject(main_thread_sem_has_work, 0) == WAIT_OBJECT_0)
					break; /* We've taken semaphore that we have work */
				if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) == TRUE)
				{ /* loop message processing to be able process timer messages */
					TranslateMessage(&msg); 
					DispatchMessage(&msg);
				}
//				else if(WaitForSingleObject(main_thread_sem_has_work, 20) == WAIT_OBJECT_0)
//					break; /* We've taken semaphore that we have work */
			}
#endif
			while(GetMessage(&msg, NULL, 0, 0))
			{
				if(msg.message == event_message)
					break;
				/* loop message processing to be able process timer messages */
				TranslateMessage(&msg); 
				DispatchMessage(&msg);
			}
			sem_wait_not_intr(&main_thread_sem_has_work);	/* Take semaphore that we have work */
		}
#else /* UNIX */
		sem_wait_not_intr(&main_thread_sem_has_work);	/* Take semaphore that we have work */
#endif

#ifndef NDEBUG
		/* assert that semaphore is set busy */
		sem_getvalue(&main_thread_sem_busy, &i);
		assert(i == 0);
#endif

		if(is_combined_debugger)
			CD_main_switch(&number_stops);
		else
		{
			switch(active_thread_recv)
			{
			case -1:/* this is wait timer message */
				if(number_stops == 1)
				{ /* wait tasks to be finished fail - exit anyway */
#ifndef NDEBUG
					fprintf(stderr, "Stop by timer\n");
#endif
					DestroyTaskGroups();
					number_stops = 2;
					continue;
				}

				if(task_groups_shall_be_reported)
				{
					if(tasksGroupsCreated)
						ReportTasksGroups();
					else
						CreateTasksGroups(FALSE); /* create and report tasks */
				}
				else
					CreateTasksGroups(TRUE); /* check tasks */
				continue;
			case -2:
				CheckTasksGroup();
				continue;
			default:
				assert(active_thread_recv >= 0);
				assert(active_thread_recv <= number_of_norma_tasks);
				break;
			}
			/* Main switch - what kind of message received */
			switch(mess_id_recv[active_thread_recv])
			{
			/* Debugged program started */
			case ND_NOTIF_PROGRAM_START:
				OnProgramStart();
				break;

			/* Debugged program finished */
			case ND_NOTIF_PROGRAM_END:
				if(number_stops == 0)
				{
					if(OnProgramEnd())
						number_stops = 1;
				}
				break;

			/* UI send continue */
			case ND_CONTINUE_EXECUTION:
#ifdef DEBUG
				fputs("Message: ND_CONTINUE_EXECUTION\n", stderr);
#endif
				assert(active_thread_recv == number_of_norma_tasks);
				if(number_stops == 1)
				{ /* this is the STOP message */
					number_stops = 2;
					thread_should_stops[active_thread_recv] = TRUE;
				}
				/* Else - send it to all tasks that waits it */
				else
					OnContinueExecution(0, NULL, 0);
				break;

			/* UI send break */
			case ND_BREAK_EXECUTION:
#ifdef DEBUG
				fputs("Message: ND_BREAK_EXECUTION\n", stderr);
#endif
				if(active_thread_recv == number_of_norma_tasks)
				{/* this is UI command to break execution */
					OnBreakExecution(ND_BREAK_EXECUTION);
					number_stops = 1;
				}
				/* this is notification that task breaking execution */
				assert(number_stops == 1);
				if(OnNotifyBreakExecution())
				{
					StopPWaitTimer();
					number_stops = 2;
				}
				break;

			case ND_NOTIF_MAIN_PART_START:
				if(number_stops == 0)
					OnMainPartStart();
				break;
			
			case ND_NOTIF_MAIN_PART_END:
				if(number_stops == 0)
					OnMainPartEnd();
				break;

			case ND_NOTIF_DISTRIBUTION_PART_START:
				if(number_stops == 0)
					OnDistributionPartStart();
				break;

			case ND_NOTIF_DISTRIBUTION_PART_END:
				if(number_stops == 0)
					OnDistributionPartEnd();
				break;

			case ND_NOTIF_PART_START:
				if(number_stops == 0)
					OnScalarPartStart();
				break;

			case ND_NOTIF_PART_END:
				if(number_stops == 0)
					OnScalarPartEnd();
				break;

			case ND_NOTIF_OPERATOR_START:
				if(number_stops == 0)
					OnOperatorStart();
				break;

			case ND_NOTIF_EX_OPERATOR_START:
				if(number_stops == 0)
					OnExOperatorStart();
				break;

			case ND_WAIT_SIGNAL:
				if(number_stops == 0)
					OnWaitSignal();
				break;
			
			case ND_CONTINUE_SIGNAL0:
				if(number_stops == 0)
					OnContinueSignal(0);
				break;

			case ND_CONTINUE_SIGNAL1:
				if(number_stops == 0)
					OnContinueSignal(1);
				break;

			case ND_SET_BREAKPOINT:
#ifdef DEBUG
				fputs("Message: ND_SET_BREAKPOINT\n", stderr);
#endif
				OnSetBreakpoint();
				free(task_buffer_recv[active_thread_recv]);
				task_buffer_recv[active_thread_recv] = NULL;
				break;

			case ND_REMOVE_BREAKPOINT:
#ifdef DEBUG
				fputs("Message: ND_REMOVE_BREAKPOINT\n", stderr);
#endif
				OnRemoveBreakpoint();
				free(task_buffer_recv[active_thread_recv]);
				task_buffer_recv[active_thread_recv] = NULL;
				break;

			case ND_RUN:
			case ND_PAUSE:
			case ND_STEP_OVER:
			case ND_STEP_INTO:
			case ND_UNTIL_RETURN:
				OnActionCommand();
				break;

			case ND_REGISTER_VAR:
				OnRegisterVar(FALSE);
				break;

			case ND_REGISTER_GLOBAL_VAR:
				OnRegisterVar(TRUE);
				break;

			case ND_REGISTER_DIM:
				OnRegisterDim();
				break;

			case ND_GET_VALUE:
				assert(active_thread_recv == number_of_norma_tasks);
				OnGetValue();
				break;

			case ND_TAKE_VALUE:
				assert(active_thread_recv != number_of_norma_tasks);
				OnTakeValue();
				break;

			case ND_TAKE_VALUE_ERROR:
				assert(active_thread_recv != number_of_norma_tasks);
				OnErrorTakeValue();
				break;

			case ND_GET_STATUS:
				assert(active_thread_recv == number_of_norma_tasks);
				OnGetStatus();
				break;

			case ND_GET_VERSION_ID:
#ifdef DEBUG
				fputs("Message: ND_GET_VERSION_ID\n", stderr);
#endif
				OnGetVersionID();
				break;

			case ND_GET_STACK_INFO:
				assert(active_thread_recv == number_of_norma_tasks);
				OnStackInfo();
				break;

			default:
				PrintErrorMessage("Unknown message!");
			}
		}
		/* Release the recv thread semaphore */
#ifndef NDEBUG
		sem_getvalue(&recv_thread_sem[active_thread_recv], &i);
		assert(i == 0);
#endif
		sem_post(&recv_thread_sem[active_thread_recv]);
	}
#ifdef DEBUG
	fprintf(stderr, "number_stops = %d\n", number_stops);
#endif
	sem_post(&main_thread_sem_busy); /* because thread IsStopHere may be in the state wait this semaphore */
	/* Finish the send threads */
	for(i = 0; i <= number_of_norma_tasks; i++)
	{
#ifdef DEBUG
		fprintf(stderr, "Before wait send_thread_sem[%d]\n", i);
#endif
		/* Take semaphore that we may use the buffer */	
		sem_wait_not_intr(&send_thread_sem[i]);
		sem_post(&send_thread_sem_has_work[i]);
		sem_wait_not_intr(&send_thread_sem[i]);
	}
	
#ifdef DEBUG
	fprintf(stderr, "Before RemoveAllBreakpoints\n");
#endif
	RemoveAllBreakpoints();

	return 0;
}

/***************** SEND thread loop  ************************/

static THREADRETURN start_task_thread_send(void *arg)
{
	int myID;

	myID = *((int*)arg);
	assert(myID <= number_of_norma_tasks);

	while(1)
	{
		sem_wait_not_intr(&send_thread_sem_has_work[myID]);
		if(thread_should_stops[myID])
		{
			sem_post(&send_thread_sem[myID]);
			break; /* thread shoud ends */
		}
		/* send the message */
		if(!TCPIPSendMessage(tcp_ip_handle[myID], 
			mess_id_send[myID], task_buffer_send[myID]))
		{
			PrintErrorMessage("Error send message");
			thread_should_stops[myID] = TRUE;
			return;
		}

#ifdef DEBUG
		if(myID == number_of_norma_tasks)
			fprintf(stderr, "Send message to UI: %d %s\n", 
				mess_id_send[myID], task_buffer_send[myID] ? task_buffer_send[myID] : "<null>");
#endif
		if(task_buffer_send[myID])
		{
			free(task_buffer_send[myID]);
			task_buffer_send[myID] = NULL;
		}
		sem_post(&send_thread_sem[myID]);
	}
	return;
}
	
/***************** RECV thread loop  ************************/

static THREADRETURN start_task_thread_recv(void *arg)
{
	int myID;

	myID = *((int*)arg);
	assert(myID <= number_of_norma_tasks);
#ifdef DEBUG
	printf("Receive thread N %d started\n", myID);
#endif

	if(sem_init(&recv_thread_sem[myID], 0, 1))
	{
		PrintErrorMessage("error_init sem recv");
		return;
	}
	
	if(myID == number_of_norma_tasks)
		sem_post(&calc_ready_sem); /* the last thread was started */

	while(1)
	{
		sem_wait_not_intr(&recv_thread_sem[myID]);	/* Take semaphore that we use the buffer */
		if(thread_should_stops[myID])
			break; /* thread shoud ends */

		if(!TCPIPGetMessage(tcp_ip_handle[myID], 
			&mess_id_recv[myID], &task_buffer_recv[myID]))
		{
			fprintf(stderr, "MyID = %d\n", myID);
			PrintErrorMessage("Error get message");
			thread_should_stops[myID] = TRUE;
			if(myID != number_of_norma_tasks)
				TCPIPSendMessage(tcp_ip_handle[number_of_norma_tasks], SYS_CRASH, NULL);
			return;
		}
		
#ifdef DEBUG
//		if(myID == number_of_norma_tasks)
//			fprintf(stderr, "Receive message from UI: %d\n", mess_id_recv[myID]);
#endif
		sem_wait_not_intr(&main_thread_sem_busy);
		active_thread_recv = myID;
#ifdef WIN32
		PostThreadMessage(main_thread_ID, event_message, (WPARAM) NULL, (LPARAM) NULL);
#endif
		sem_post(&main_thread_sem_has_work);
#ifdef DEBUG
//		if(myID == number_of_norma_tasks)
//			fprintf(stderr, "Message from UI %d started to be processed\n", mess_id_recv[myID]);
#endif
	}
	return;
}






