/* Header of Monitor-communication library */
/* All these functions called from Fortran */
#ifndef _MONITOR_CONNECT_H
#define _MONITOR_CONNECT_H

#include "CommonMonitorDefs.h"

#ifdef WIN32
#define FNAME(x) x
#define FORTRAN_INTEGER int
#else
	#ifdef __INTEL_COMPILER
	#define FNAME(x) x
	#else /* gcc */
	#define FNAME(x) x##_
	#endif
#define FORTRAN_INTEGER int
#endif

/* Start of all program
   This function should initialize the Monitor */ 
int FNAME(ndstprg)(const char *main_part_name, const char *filename,
			 FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC, FORTRAN_INTEGER *number_of_tasks, FORTRAN_INTEGER *mynumber);
/* Fortran program */
int FNAME(ndstpex)(const char *program_name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC);
int NDStartProgram(const char *main_part_name, const char *filename,
			 FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC, FORTRAN_INTEGER number_of_tasks, FORTRAN_INTEGER mynumber);
void FNAME(ndenprg)(void);
void NDEndProgram(const char *, FORTRAN_INTEGER);
/* register of "sizeof" of standard types:  INTEGER_T, REAL_T and DOUBLE_T */
void FNAME(ndregsz)(FORTRAN_INTEGER *int_size, FORTRAN_INTEGER *real_size, FORTRAN_INTEGER *double_size);
void NDRegisterSize(FORTRAN_INTEGER int_size, FORTRAN_INTEGER real_size, FORTRAN_INTEGER double_size);

/* Start of Main Part */ 
void FNAME(ndstmp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC);
void NDStartMainPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC);
void FNAME(ndenmp)(const char *filename, FORTRAN_INTEGER *end_line);
void NDEndMainPart(const char *filename, FORTRAN_INTEGER end_line);
/* Scalar part */
int FNAME(ndstp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC);
int NDStartPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC);
void FNAME(ndenp)(const char *filename, FORTRAN_INTEGER *end_line);
void NDEndPart(const char *filename, FORTRAN_INTEGER end_line);
/* External Fortran function */
int FNAME(ndstexf)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC);
void FNAME(ndenexf)(const char *filename, FORTRAN_INTEGER *end_line);
/* Distribution part */
void FNAME(ndstdp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC,
	FORTRAN_INTEGER *i, FORTRAN_INTEGER *j, FORTRAN_INTEGER *myi, FORTRAN_INTEGER *myj, FORTRAN_INTEGER *ipoints, FORTRAN_INTEGER *jpoints,
	const char *iname, const char *jname);
void NDStartDistributionPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC,
	FORTRAN_INTEGER i, FORTRAN_INTEGER j, FORTRAN_INTEGER myi, FORTRAN_INTEGER myj, FORTRAN_INTEGER ipoints, FORTRAN_INTEGER jpoints,
	const char *iname, const char *jname);
void FNAME(ndendp)(const char *filename, FORTRAN_INTEGER *end_line);
void NDEndDistributionPart(const char *filename, FORTRAN_INTEGER end_line);

/* Registration of variable "name".
   type - INTEGER_T, REAL_T or DOUBLE_T
   number_of_dimensions - 0 if scalar.
   and then for each dimension:
   char *index_name;
   FORTRAN_INTEGER declared_start_point;
   FORTRAN_INTEGER declared_end_point;
   FORTRAN_INTEGER norma_start_point;
   FORTRAN_INTEGER norma_end_point;
   Norma                 Fortran              data
   B(i=1..40)            B(10)            "B\000", B, 1, 1, "i\000", 1, 40, 1, 10, 
   C(i=1..40; k=5..20)   C(0:11, 5:20)    "C\000", C, 1, 2, "i\000", 1, 40, 0, 11, "k\000", 5, 20, 5, 20
   N                     N                "N\000", N, 2, 0
*/

void FNAME(ndregvar)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...);
void NDRegisterVariable(const char *name, void *pointer, NormaVarType type, 
			  FORTRAN_INTEGER number_of_dimensions, ...);

/* Variable in External Fortran function */
/* Registration of variable "name".
   type - INTEGER_T, REAL_T, DOUBLE_T, or CHARACTER - (100 + len)
   number_of_dimensions - 0 if scalar.
   and then for each dimension:
   char *index_name;
   FORTRAN_INTEGER start_point;
   FORTRAN_INTEGER end_point;
*/
void FNAME(ndregvex)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...);

/* Global Variable in External Fortran function
   The differnce with ndregvex() is only logical
 */
void FNAME(ndregvg)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...);

/* called before each operator.
   operator == 0 => it is the first operator */
int FNAME(ndopstar)(FORTRAN_INTEGER *previous_operator,
			  FORTRAN_INTEGER *previous_operator_line_begin,
			  FORTRAN_INTEGER *previous_operator_line_end,
			  FORTRAN_INTEGER *current_operator,
			  FORTRAN_INTEGER *current_operator_line_begin,
			  FORTRAN_INTEGER *current_operator_line_end);
int NDOperatorStart(FORTRAN_INTEGER previous_operator,
			  FORTRAN_INTEGER previous_operator_line_begin,
			  FORTRAN_INTEGER previous_operator_line_end,
			  FORTRAN_INTEGER current_operator,
			  FORTRAN_INTEGER current_operator_line_begin,
			  FORTRAN_INTEGER current_operator_line_end);
/* Operator in External Fortran function */
int FNAME(ndopstex)(FORTRAN_INTEGER *is_potentially_lock,
			  FORTRAN_INTEGER *previous_operator_line_begin,
			  FORTRAN_INTEGER *previous_operator_line_end,
			  FORTRAN_INTEGER *current_operator_line_begin,
			  FORTRAN_INTEGER *current_operator_line_end);

/* called in start-part of Distribution part before send parameters and receive the rezult */
void FNAME(ndwait)(void);
void NDWait(void);
/* called in Distribution part to notify that main part should continue */
void FNAME(ndcont)(FORTRAN_INTEGER *for_main);
void NDContinue(FORTRAN_INTEGER for_main);

#endif
