#include "TCPIPlib.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#include <errno.h>

typedef int socklen_t;

#else

#include <netinet/tcp.h>

#endif

#include "Thread_Semaphore.h"

#define PACKAGE_HEADER_LEN 30

static SOCKET server_socket;
static BOOL is_server_socket_inited = FALSE;
static BOOL cancel_accept_socket;
static int number_of_ports_number_try = NUMBER_OF_PORTS_NUMBERS_TRY;

static sem_t sem_access_conlist;

typedef struct ServerConnections
{
	TCPIPConnectionHandle handle;
	int type;
	BOOL connected;
	char address[MAX_INTERNET_ADDRESS_SIZE + 1];
	struct ServerConnections *next;
} ServerConnections;
static ServerConnections *beg_ServerConnections = NULL;

char **error_message_pointer = NULL;

static void PrintErrorMessage(const char *msg)
{
	static char out_memory[] = "Out memory";

	if(error_message_pointer == NULL)
	{
#ifdef WIN32
		MessageBox(NULL, msg, "Error", MB_OK);
#else
		fputs(msg, stderr);
		fputc('\n',stderr);
#endif
	}
	else
	{
		if((*error_message_pointer != NULL) &&
			(*error_message_pointer != out_memory))
			free(*error_message_pointer);
		if((*error_message_pointer = (char *)malloc(strlen(msg) + 1)) == NULL)
			*error_message_pointer = out_memory;
		strcpy(*error_message_pointer, msg);
	}
}

#ifdef WIN32
/*   */
static void RegisterInSocketLib(void)
{
	static BOOL is_registred = FALSE;
	WSADATA wsdata;
	WORD wVersionRequested;

	if(!is_registred)
	{
		wVersionRequested = MAKEWORD(2,2);
		if(WSAStartup(wVersionRequested, &wsdata) != 0)
		{
			PrintErrorMessage("WSAStartup error");
			return;
		}
		is_registred = TRUE;
	}
}
#endif

BOOL TCPIPServerCreateSocket(IPInfo *ipinfo)
{
	int i, flag = 1;
	struct sockaddr_in serv_addr;
	struct hostent *hp;
	char buf [MAX_INTERNET_ADDRESS_SIZE + 1];
	struct in_addr addr;

	assert(!is_server_socket_inited);
#ifdef WIN32
	RegisterInSocketLib();
#endif
	/* ,   TCP*/
	if((server_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		PrintErrorMessage("socket creation error");
		return FALSE;
	}

	/* set TCP_NODELAY for immediately send data */
	if(setsockopt(server_socket,            /* socket affected */
		IPPROTO_TCP,     /* set option at TCP level */
		TCP_NODELAY,     /* name of option */
		(char *) &flag,  /* the cast is historical
							cruft */
		sizeof(int)) < 0)    /* length of option value */
		PrintErrorMessage("Error set TCP_NODELAY");

	/*   */
	memset(&serv_addr, 0, sizeof(serv_addr) );
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;

	/*    */
	for(i = 0; TRUE ; )
	{
		/*      */
		serv_addr.sin_port = htons((u_short) ipinfo->port_number);
		/*    */
		if(bind(server_socket, (struct sockaddr *)&serv_addr,
								sizeof (serv_addr)) == -1)
		{
			/*   -   */
#ifdef WIN32
			if(WSAGetLastError() != WSAEADDRINUSE)
#else
			if(errno != EADDRINUSE)
#endif
			{
				PrintErrorMessage("Port binding error");
				return FALSE;
			}
			if(++i < number_of_ports_number_try)
				ipinfo->port_number++;
			else
				break;
		}
		else
			break;
	}
	if(i == number_of_ports_number_try)
	{
		PrintErrorMessage("Port binding error: no free ports");
		return FALSE;
	}
	
	if(listen(server_socket, /*SOMAXCONN*/50) == -1)
	{
		PrintErrorMessage("Port lesten error");
		return FALSE;
	}
	TCPIPSetConnectionNonBlockedMode(server_socket);
/*  */
	if((gethostname(buf, sizeof(buf)) != 0) ||
			((hp = gethostbyname(buf)) == NULL))
	{
		PrintErrorMessage("gethostbyname error");
		return FALSE;
	}
	strcpy(ipinfo->address, hp->h_name);
	assert(hp->h_addr_list[0]);
	addr.s_addr = *((unsigned long *) (hp->h_addr_list[0]));
	strcpy(ipinfo->dot_address, inet_ntoa(addr));
	ipinfo->alias_addresses[0] = '\0';
	for(i = 0; (hp->h_addr_list[i + 1] != NULL) && (i < MAX_ALIASES_AMOUNT); i++)
	{
		addr.s_addr = *((unsigned long *) (hp->h_addr_list[i + 1]));
		sprintf(&ipinfo->alias_addresses[strlen(ipinfo->alias_addresses)],
			"%s%s <%s>", i > 0 ? ", " : "",
			&hp->h_addr_list[i + 1][hp->h_length],
			inet_ntoa(addr));
	}
	
	if(error_message_pointer == NULL)
	{
		printf("Server %s <%s> ready on port %d\n", ipinfo->address, ipinfo->dot_address, ipinfo->port_number);
		if(i > 0)
			printf("Alias names: %s\n", ipinfo->alias_addresses);
	}

	is_server_socket_inited	= TRUE;
	/* create semaphore for access to connection list */
	if(sem_init(&sem_access_conlist, 0, 1))
	{
		PrintErrorMessage("error_init sem");
		return 1;
	}

	return TRUE;
}

BOOL TCPIPGetConnectionFromThread(IPInfo *ipinfo, TCPIPConnectionHandle *tcp_ip_handle, int number_of_retry)
{
	ServerConnections *tek_ServerConnections;
	int i = 0;
	char buffer[PACKAGE_HEADER_LEN];
	
	if(!ipinfo->is_server || !is_server_socket_inited)
	{
		assert(FALSE);
		return FALSE;
	}

	while(1)
	{
		sem_wait_not_intr(&sem_access_conlist); /* wain the access sem */
		tek_ServerConnections = beg_ServerConnections;
		while(tek_ServerConnections)
		{
			if((tek_ServerConnections->type == ipinfo->type) &&
				(tek_ServerConnections->connected == FALSE))
			{ /* the type of called connection is exists in the queue */
				*tcp_ip_handle = tek_ServerConnections->handle;
				tek_ServerConnections->connected = TRUE;
				sprintf(buffer, "%d", SYS_OK);
				send(*tcp_ip_handle, buffer, PACKAGE_HEADER_LEN, 0);
				strcpy(ipinfo->address, tek_ServerConnections->address);
				sem_post(&sem_access_conlist);
				return TRUE;
			}
			tek_ServerConnections = tek_ServerConnections->next;
		}
		sem_post(&sem_access_conlist);
		if(++i > number_of_retry)
			break;
		Sleep(3000);
	}
	return FALSE;
}

BOOL TCPIPConnect(IPInfo *ipinfo, 
				  TCPIPConnectionHandle *tcp_ip_handle,
				  void * dummy/* dummy for what??? */)
{
	int addrlen;
	struct sockaddr_in clnt_addr;
	SOCKET ns;
	struct sockaddr_in serv_addr;
	struct hostent *hp;
	char buffer[PACKAGE_HEADER_LEN];

	if(ipinfo->is_server) /* we want to be a server */
	{
		if(!is_server_socket_inited && (TCPIPServerCreateSocket(ipinfo) == FALSE))
			return FALSE;

		/* first try to get this connection from list */
		if(TCPIPGetConnectionFromThread(ipinfo, tcp_ip_handle, 0))
			return TRUE;
		
		while(1)
		{
			ServerConnections *tek_ServerConnections;

			addrlen = sizeof(clnt_addr);
			memset(&clnt_addr, 0, addrlen);
/* .   ns   
  */
			cancel_accept_socket = FALSE;
accept_again:
			if((ns = accept(server_socket, (struct sockaddr *)&clnt_addr,
				(socklen_t *)&addrlen)) == -1)
			{
#ifdef WIN32
				if(WSAGetLastError() == WSAEWOULDBLOCK)
#else
				if((errno == EAGAIN) || (errno == EWOULDBLOCK))
#endif
				{
					Sleep(100);
					if(cancel_accept_socket)
						return FALSE;
					goto accept_again;
				}
				PrintErrorMessage("accept error");
				return FALSE;
			}
			TCPIPSetConnectionBlockedMode(ns);
	/*   */
#ifndef NDEBUG
			if(error_message_pointer == NULL)
				printf("Client address %s\n", inet_ntoa(clnt_addr.sin_addr));
#endif
	/*   client*/
			if(recv(ns, buffer, PACKAGE_HEADER_LEN, 0) != PACKAGE_HEADER_LEN)
			{
				char buf[100];
				sprintf(buf, "Error connect with client %s", inet_ntoa(clnt_addr.sin_addr));
				PrintErrorMessage(buf);
				closesocket(ns);
				return FALSE;
			}

			if((tek_ServerConnections = (ServerConnections *)malloc(sizeof(ServerConnections))) == NULL)
			{
				PrintErrorMessage("Out memory");
				return FALSE;
			}
			tek_ServerConnections->handle = ns;
			tek_ServerConnections->type = atoi(buffer);
			strncpy(tek_ServerConnections->address,
				inet_ntoa(clnt_addr.sin_addr), MAX_INTERNET_ADDRESS_SIZE);
			sem_wait_not_intr(&sem_access_conlist); /* wain the access sem */
			tek_ServerConnections->next = beg_ServerConnections;
			beg_ServerConnections = tek_ServerConnections;
			if(tek_ServerConnections->type == ipinfo->type)
			{
				*tcp_ip_handle = ns;
				sprintf(buffer, "%d", SYS_OK);
				send(*tcp_ip_handle, buffer, PACKAGE_HEADER_LEN, 0);
				tek_ServerConnections->connected = TRUE;
				strcpy(ipinfo->address, tek_ServerConnections->address);
				sem_post(&sem_access_conlist);
				return TRUE;
			}
			tek_ServerConnections->connected = FALSE;
			sem_post(&sem_access_conlist);
		}
	}
	/* we want to be client */
#ifdef WIN32
	RegisterInSocketLib();
#endif

/*    */
	if((hp = gethostbyname(ipinfo->address)) == NULL)
	{
#ifdef WIN32/*gethostbyname   IP */
		unsigned long addr;
		addr = inet_addr(ipinfo->address);
		if((hp = gethostbyaddr((char *)&addr, 4, AF_INET)) == NULL)
#endif
		{
			char buf[MAX_INTERNET_ADDRESS_SIZE + 100];
			sprintf(buf, "Error resolve address: %s\n", ipinfo->address);
			PrintErrorMessage(buf);
			return FALSE;
		}
	}

	while(1)
	{
		int i;

		for(i = 0; i < 1; i++)
		{
			int flag = 1;
			memset(&serv_addr, 0, sizeof(serv_addr));
			memcpy(&serv_addr.sin_addr, hp->h_addr, hp->h_length);
			serv_addr.sin_family = hp->h_addrtype;
			serv_addr.sin_port = htons((u_short) ipinfo->port_number);
		/* */
			if((ns = socket(AF_INET, SOCK_STREAM, 0)) == -1)
			{
				PrintErrorMessage("socket creation error");
				return FALSE;
			}

			/* set TCP_NODELAY for immediately send data */
			if(setsockopt(ns,            /* socket affected */
				IPPROTO_TCP,     /* set option at TCP level */
				TCP_NODELAY,     /* name of option */
				(char *) &flag,  /* the cast is historical
									cruft */
				sizeof(int)) < 0)    /* length of option value */
				PrintErrorMessage("Error set TCP_NODELAY");

		/*    */
#ifndef NDEBUG
			if(error_message_pointer == NULL)
				printf("Server address: %s\n",inet_ntoa(serv_addr.sin_addr));
#endif
		/*  */
			if(connect(ns, (struct sockaddr *)&serv_addr, sizeof (serv_addr)) == -1)
			{
				if(error_message_pointer == NULL)
				{
#ifdef WIN32
					printf("Try: %d, Error: %X\n", i, WSAGetLastError());
#else
					char str[50];
					sprintf(str, "Try: %d, Errno (%d): ", i, errno);
					perror(str);
#endif
				}
				Sleep(3000);
				closesocket(ns);
			}
			else
				break;
		}
		if(i == 10)
		{
			PrintErrorMessage("socket connect error");
			return FALSE;
		}
	/*   about my type */
		sprintf(buffer, "%d", ipinfo->type);
		if(send(ns, buffer, PACKAGE_HEADER_LEN, 0) == SOCKET_ERROR)
		{
			PrintErrorMessage("send error");
			return FALSE;
		}
	/*   */
		if(recv(ns, buffer, PACKAGE_HEADER_LEN, 0) != PACKAGE_HEADER_LEN)
		{
			PrintErrorMessage("recv error");
			return FALSE;
		}
		if(atoi(buffer) == SYS_OK)
		{
			*tcp_ip_handle = ns;
			break;
		}
		
		/* here should be the attempt to connect to another port */
		assert(FALSE);
	}
	return TRUE;
}

BOOL TCPIPDisconnectEx(TCPIPConnectionHandle tcp_ip_handle, BOOL close_server)
{
#ifdef WIN32
	shutdown(tcp_ip_handle, SD_BOTH);
#else
	shutdown(tcp_ip_handle, SHUT_RDWR);
#endif
	Sleep(100); /* This fix for crash, that may occure when one thread is in recv function
                       and the another thread calls TCPIPDisconnect. If closesocket before
                       recv be finished, crash occure */

	if(is_server_socket_inited)
	{
		ServerConnections *pred_ServerConnections, *tek_ServerConnections;

		sem_wait_not_intr(&sem_access_conlist); /* wain the access sem */
		tek_ServerConnections = beg_ServerConnections;
		while(tek_ServerConnections)
		{
			if(tek_ServerConnections->handle == tcp_ip_handle)
			{ /* this was the server connection */
				assert(tek_ServerConnections->connected == TRUE);
				assert(is_server_socket_inited);
				if(tek_ServerConnections == beg_ServerConnections)
					beg_ServerConnections = tek_ServerConnections->next;
				else
					pred_ServerConnections->next = tek_ServerConnections->next;
				free(tek_ServerConnections);
				sem_post(&sem_access_conlist);
				if(closesocket(tcp_ip_handle) != 0)
					return FALSE;
				if(close_server && (beg_ServerConnections == NULL))
				{ /* The last server connection closed - close server socket */
					cancel_accept_socket = TRUE;
					if(closesocket(server_socket) != 0)
						return FALSE;
					sem_destroy(&sem_access_conlist);
					is_server_socket_inited = FALSE;
				}
				return TRUE;
			}
			pred_ServerConnections = tek_ServerConnections;
			tek_ServerConnections = tek_ServerConnections->next;
		}
	}
	/* client connection */
	if(closesocket(tcp_ip_handle) != 0)
		return FALSE;
	return TRUE;
}

BOOL TCPIPSendMessage(TCPIPConnectionHandle tcp_ip_handle,
					  TCPIPMessageType mess_id, const char *data)
{
	char *buffer;
	int data_len = data ? strlen(data) + 1 : 0;

	if((buffer = (char *)malloc(PACKAGE_HEADER_LEN + data_len)) == NULL)
	{
		PrintErrorMessage("Out memory");
		return FALSE;
	}
	memset(buffer, 0, PACKAGE_HEADER_LEN);
	sprintf(buffer, "%d %d", mess_id, data_len);
	if(data_len)
		memcpy(&buffer[PACKAGE_HEADER_LEN], data, data_len);

	if(send(tcp_ip_handle, buffer, PACKAGE_HEADER_LEN + data_len, 0) == SOCKET_ERROR)
	{
		PrintErrorMessage("send error");
		free(buffer);
		return FALSE;
	}
	free(buffer);

	return TRUE;
}

BOOL TCPIPGetMessage(TCPIPConnectionHandle tcp_ip_handle,
					  TCPIPMessageType *mess_id, char **data)
{
	char package_header[PACKAGE_HEADER_LEN];
	char *buffer;
	int retcode;
	int byte_pos = 0;
	int data_len;

recv_again1:
	assert(byte_pos < PACKAGE_HEADER_LEN);
	if((retcode = recv(tcp_ip_handle, &package_header[byte_pos],
		PACKAGE_HEADER_LEN - byte_pos, 0)) != PACKAGE_HEADER_LEN - byte_pos)
	{
		if((retcode == -1) && (errno == EINTR))
			goto recv_again1;
#ifdef WIN32
		if(WSAGetLastError() == WSAEWOULDBLOCK)
#else
		if(errno == EAGAIN)
#endif
		{ /* it was non-blocked socket */
			*mess_id = SYS_NO_MESSAGES;
			if(data != NULL)
				*data = NULL;
			return TRUE;
		}
		if(retcode > 0)
		{
			byte_pos += retcode;
			goto recv_again1;
		}
		if(error_message_pointer == NULL)
#ifdef WIN32
			fprintf(stderr, "retcode = %d error = %X\n", retcode, WSAGetLastError());
#else
			fprintf(stderr, "retcode = %d errno = %d\n", retcode, errno);
#endif
		PrintErrorMessage("recv error");
		return FALSE;
	}
	if(sscanf(package_header, "%d %d", mess_id, &data_len) != 2)
	{
		PrintErrorMessage("recv error");
		return FALSE;
	}

	if(data_len)
	{
		int received_bytes = 0;
		if((buffer = (char *)malloc(data_len)) == NULL)
		{
			PrintErrorMessage("Out memory");
			return FALSE;
		}
	recv_again2:
		if((retcode = recv(tcp_ip_handle, &buffer[received_bytes], data_len - received_bytes, 0)) != data_len - received_bytes)
		{
			if((retcode == -1) && (errno == EINTR))
				goto recv_again2;

			if(retcode > 0)
			{
#ifdef DEBUG
				if(error_message_pointer == NULL)
					fprintf(stderr, "TCPIPGetMessage: get %d bytes instead of expected %d. Trying again.\n", retcode, data_len - received_bytes);
#endif
				received_bytes += retcode;
				goto recv_again2;
			}
			if(error_message_pointer == NULL)
				fprintf(stderr, "retcode = %d errno = %d\n", retcode, errno);
			PrintErrorMessage("recv error");
			return FALSE;
		}
		if(data != NULL)
			*data = buffer;
		else
			free(buffer);
	}
	else
		if(data != NULL)
			*data = NULL;
	return TRUE;
}

void TCPIPCancelServer(void)
{
	if(!is_server_socket_inited)
		return;
	cancel_accept_socket = TRUE;
	closesocket(server_socket);
	is_server_socket_inited = FALSE;
}

void TCPIPSetConnectionBlockedMode(TCPIPConnectionHandle tcp_ip_handle)
{ /* Set Socket blocking mode */
#ifdef WIN32
	u_long mode = 0;
	if(ioctlsocket(tcp_ip_handle, FIONBIO, &mode) != 0)
		PrintErrorMessage("Set Socket blocking mode error");
#else
	int flags = fcntl(tcp_ip_handle, F_GETFL);
	if(fcntl(tcp_ip_handle, F_SETFL, flags & ~((long)O_NONBLOCK)) != 0)
		PrintErrorMessage("Set Socket blocking mode error");
#endif
}

void TCPIPSetConnectionNonBlockedMode(TCPIPConnectionHandle tcp_ip_handle)
{ /* Set Socket non-blocking mode */
#ifdef WIN32
	u_long mode = 1;
	if(ioctlsocket(tcp_ip_handle, FIONBIO, &mode) != 0)
		PrintErrorMessage("Set Socket non-blocking mode error");
#else
	int flags = fcntl(tcp_ip_handle, F_GETFL);
	if(fcntl(tcp_ip_handle, F_SETFL, flags | O_NONBLOCK) != 0)
		PrintErrorMessage("Set Socket non-blocking mode error");
#endif
}

void TCPIPSetMaxPortsNum(int num) { number_of_ports_number_try = num;}
