/* Header of TCP/IP communicating library */
#ifndef _TCPIPlib_H
#define _TCPIPlib_H

#include <assert.h>

#ifdef WIN32
#include <Winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#endif

#ifndef WIN32

#define closesocket close
typedef int SOCKET;
#define SOCKET_ERROR -1
#define Sleep(s) usleep(s*1000)

#endif


#ifndef TRUE
typedef int BOOL;
#define TRUE 1
#define FALSE 0
#endif

#define MAX_INTERNET_ADDRESS_SIZE 256
#define MAX_DOT_ADDRESS_SIZE 15
#define MAX_ALIASES_AMOUNT 5
#define DEFAULT_PORT_NUMBER 1500
#define NUMBER_OF_PORTS_NUMBERS_TRY 10

typedef struct IPInfo
{
	char address[MAX_INTERNET_ADDRESS_SIZE + 1];
	char dot_address[MAX_DOT_ADDRESS_SIZE + 1];
	char alias_addresses[(MAX_INTERNET_ADDRESS_SIZE + MAX_DOT_ADDRESS_SIZE + 5) * MAX_ALIASES_AMOUNT];
	int port_number;
	BOOL is_server; /* or client */
	int type; /* 0 - UI, 1 - debugged program, 2 - Monitor */
} IPInfo;

typedef SOCKET TCPIPConnectionHandle;
typedef int TCPIPMessageType;

void TCPIPSetMaxPortsNum(int num);
BOOL TCPIPServerCreateSocket(IPInfo *ipinfo);

BOOL TCPIPConnect(IPInfo *ipinfo, 
				  TCPIPConnectionHandle *tcp_ip_handle,
				  void * /* dummy for what??? */);
BOOL TCPIPGetConnectionFromThread(IPInfo *ipinfo, TCPIPConnectionHandle *tcp_ip_handle, int number_of_retry);

BOOL TCPIPDisconnectEx(TCPIPConnectionHandle tcp_ip_handle, BOOL close_server);
#define TCPIPDisconnect(tcp_ip_handle) TCPIPDisconnectEx(tcp_ip_handle, TRUE)

BOOL TCPIPSendMessage(TCPIPConnectionHandle tcp_ip_handle,
					  TCPIPMessageType mess_id, const char *data);

BOOL TCPIPGetMessage(TCPIPConnectionHandle tcp_ip_handle,
					  TCPIPMessageType *mess_id, char **data);

void TCPIPCancelServer(void);

void TCPIPSetConnectionBlockedMode(TCPIPConnectionHandle tcp_ip_handle);
void TCPIPSetConnectionNonBlockedMode(TCPIPConnectionHandle tcp_ip_handle);

extern char **error_message_pointer;

/* System Messages */
#define SYS_FAIL 0
#define SYS_OK 1
#define SYS_NO_MESSAGES 2
#define SYS_CRASH 3
#define SYS_GATE_ID 4
/* %d or no data when no gate */

#endif
