#include <assert.h>
#include <errno.h>

#ifdef WIN32
#include <process.h>
#include <direct.h>
#else
#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#endif

#include "Thread_Semaphore.h"

/* Thread and Semaphores defines for Windows */
#ifdef WIN32

int pthread_create(pthread_t *thr_id, pthread_attr_t *attr, 
				   void (*start_func)(void*), 
				   void *args)
{
	HANDLE handle = CreateThread(NULL, 0, 
		(LPTHREAD_START_ROUTINE) start_func,
		args, 0, thr_id);
	if(handle)
	{
		CloseHandle(handle);
		return 0; /* OK */
	}
	return 1; /* Fail */
}

int sem_init(sem_t *handle, long pshare, long init_count)
{
	*handle = CreateSemaphore(NULL, init_count, 255, NULL);
	if(*handle)
		return 0; /* OK */
	return 1; /* Fail */
}

int sem_post(sem_t *handle)
{
	if(ReleaseSemaphore(*(handle), 1, NULL) == 0)
		return -1;
	return 0;
}

int sem_getvalue(sem_t *handle, int *value)
{
	assert(sizeof(int) == sizeof(long));
	switch(WaitForSingleObject(*(handle), 0))
	{
	case WAIT_OBJECT_0:
        if(!ReleaseSemaphore(*(handle), 1, (long *)value))
			return 1; /* Fail */
		(*value)++;
		break;
	case WAIT_TIMEOUT:
		*value = 0;
		break;
	default:
		assert(FALSE);
		return 1; /* Fail */
	}
	return 0; /* OK */
}

#else /* UNIX */

int sem_wait_not_intr(sem_t *sem)
{
	int rez;
	do
	{
		errno = 0;
		rez = sem_wait(sem);
	} while(errno == EINTR);
	return rez;
}

#endif
