/* $Id: mod2c.c,v 1.2 1999/09/24 10:43:53 alena Exp $ */
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include "status.h"

#define N_EOS 59
#define N_SJUMP 1

#define SIZE_HEADER 36
#define SIZE_ENT 11
#define SIZE_EXT 9
#define SIZE_FUNC 11
#define SIZE_RLD 4
#define SIZE_NAME 8
#define SIZE_TBL 8

#define TYP_FUNC 0
#define TYP_SYST 1
#define TYP_SS   2
#define TYP_XS   3
#define TYP_BOX  4
#define TYP_EMP  5

char rcx_id_mod2c[]="$Id: mod2c.c,v 1.2 1999/09/24 10:43:53 alena Exp $";

typedef struct tfunc {
	int addr;
	char name[9];
	} tfunc;

#define MAXLINE 512 
#define MAX_FUN 10*1024 
#define MAX_MOD 128 
/* x-abo */
#define MAX_RLD 20*1024
#define MAX_EXT 1024
#define MAX_LOC MAX_FUN/10
#define CODE_LENGTH 332*1024

typedef struct {
        char *addr;
	int mod;
	unsigned char status;
	unsigned char trace;
} func; /* for size */

typedef unsigned int uint4;
typedef unsigned int uint2;
 
    /********* table of function **********/ 
#define FL_NONE     0
#define FL_LOCAL    1
#define FL_CALL     2
#define FL_LOC_SPEC 4
#define FL_ENTRY    8
#define FL_EXTERN   16
#define FL_SYSTEM   32
#define FL_SWAP     64
#define FL_EMPTY    128
#define FL_ALL   0xff

#define NO_SPEC   -1

struct { 
   char * name; 
   char * alt_name;
   uint4 addr; 
   uint4 spec;
   uint4 enddate;
   unsigned char flag;
   uint2 module;
   uint2 type; /* debug */
       } table[MAX_FUN]; 

struct {
  uint4 addr;
  char name[10];
  uint2 module;
  int type;
} table_ext[MAX_EXT];

struct {
  uint2 module;
  uint2 idx;
  uint4 addr;
} table_rld[MAX_RLD];

struct {
  uint4 addr;
} table_max_loc[MAX_LOC];

struct {
  char name[9]; /* module name */
  uint4 len;
  int fn;
  int enn;
  int exn;
  uint2 first;
  uint4 addr;
} moduls[MAX_MOD];


void in_byte(int f);
unsigned char in_code(int f);
void in_word(int f, int*);
void in_long(int f, long*);
void in_string(int f, int len, char * buf);
void read_header(int f);
void read_ent(int ff,int num);
void read_ext(int ff, int num);
tfunc* read_func(int ff,int num);
void read_text(int ff,int len, int num, tfunc*);
void read_rld(int ff,int num);
void read_end(int ff);
int func_cmp(const void * f1, const void * f2);


void register_code(unsigned char byte);
uint2 register_func(char * fp,uint4 addr,int type, int mod);
uint2 register_mod(char * fp,uint4 len,int fn,int enn, int exn);
void register_rld(uint4 addr, uint2 tab);
static char * well_ident(char * id);
int type_recoding(int type);
void add_rld();
void check_ext();
static char *  deexp(char * f1, char * e2);

static void print_extern_functions();
static void print_name_table();
static void print_func_table();
static void print_struct_func();
static void print_global_var();
static void print_code_table();
static void make_new_file();
static int  print_statistics();


/*---------------------------*/

int module_length=0;
int function_number=0;
int entry_number=0;
int extern_number=0;
int relocation_number=0;

uint2 max_fun=0;     /* max index function in table */ 
uint2 max_mod=0;     /* max index moduls in moduls  */ 
uint2 max_ext=0;     /* max index function in  table_ext */ 
uint2 max_rld=0;     /* max index in  rld_ext */ 
uint2 max_loc=0;     /* max index in  table_max_loc */ 
uint4 max_code=0;     /* max index  of code */

#define MAX_STRING 10

FILE * newf;
tfunc * func_table=NULL;
unsigned char * code;
uint2 seek_task=-1;

uint4 lenght = CODE_LENGTH;
uint4 lenght_fun= MAX_FUN;
uint4 lenght_mod= MAX_MOD;
uint4 lenght_ext= MAX_EXT;
uint4 lenght_loc= MAX_LOC;
uint4 lenght_rld= MAX_RLD;

#define byte2(b2,b1) (((b2)<<8) + (b1) )
#define byte4(b4,b3,b2,b1) (((b4)<<24) + ((b3)<<16) + ((b2)<<8) + (b1))

static unsigned char string_buffer[MAX_STRING+1];

int eof(int des)
{ int i;
  char b[2];
  i = read(des,b,1);
  if (i==0 || i==-1) return 1;
  lseek(des,-1,SEEK_CUR);
  return 0;
}

void work(int mod_file)
{
	while (!eof(mod_file)) {
	read_header( mod_file );
	read_ent( mod_file, entry_number );
	read_ext( mod_file, extern_number );
	func_table = read_func( mod_file, function_number );
	read_text( mod_file, module_length, function_number, func_table );
	read_rld( mod_file, relocation_number );
	read_end( mod_file );
	};
}
char file_name[255]="",
     mod_name[255]; 

     int stat_flag=0;

void help()
{
       printf("mac2c [ -Options] [ Files ]\n"
	/*"-m[t|s|m|l] model memory\n" */
        "-o Name output file\n"
	"-s Internal statistics\n"); 
}
int main(int argc, char *argv[])
{
int mod_file;
char *p;
int  file_ind;

printf("REFAL-M Linker (C) Copyleft Alena Laskavaya %s\n",__DATE__);
if (argc==1) 
       { 
       help();
       exit(0); 
       }; 
 
  argv++; 
  while ( argc>1 && (*argv)[0]=='-') 
    { 
      switch ((*argv)[1]) 
      { 
      case 'm':  
               switch( (*argv)[2]) 
               { 
               case 't': lenght=CODE_LENGTH/4;break; 
               case 's': lenght=CODE_LENGTH/2;break; 
               case 'm': lenght=CODE_LENGTH*4;break; 
               case 'l': lenght=CODE_LENGTH*10;break; 
               default : fprintf(stderr,"Illegal model %c ( t,s,m,l)",*argv[2]); 
               }; 
               break; 
      case 'o' : 
               if ((*argv)[2]!=0) p=(*argv)+2; 
               else {argv++; p=(* argv); argc--;};
               strcpy(file_name,p);
	       break; 
      case 's' :
	       stat_flag=1;
	       break; 
      } 
    argv++; 
    argc--;
    };
 
  if (argc<=1)
  {
    help();
    exit(1);
  }
  if ((code=(char *)malloc(lenght))==NULL) 
     fprintf(stderr,"Out of memory.\nChoose other model.\n"); 
  if (file_name[0]==0)
    strcpy(file_name,deexp(*argv,"c"));
  for (file_ind=1; file_ind<argc; file_ind++) 
  {
     assert(argv[file_ind-1]!=NULL);
     strcpy(mod_name,deexp(argv[file_ind-1],"mod"));
     fprintf(stdout,"Processing %s\n",mod_name);
     mod_file = open( mod_name, O_RDONLY );
      if ( mod_file == -1 ) {
	      printf("Can't open file %s\n", mod_name);
	      exit(1);
	      }
  
      work(mod_file);
      close( mod_file );
   };
   
   check_ext();
   add_rld();
  
   if ((newf=fopen(file_name,"w"))==NULL)  
                { 
                fprintf (stderr,"Cannot open  output file %s\n",
                         file_name); 
                newf=stdout; 
                }; 
 
    fprintf(stdout,"Output file: ");
  if (newf==stdout) 
    fprintf(stdout,"dev stdout\n");
  else
    fprintf(stdout,"%s\n",file_name);

   make_new_file();
   fclose(newf);
   if (stat_flag==1)
     print_statistics();
return 0;
} /* main */

void read_header( int ff )
{
    in_string( ff, 5 ,string_buffer); /* "\tSignature" */
    if (strncmp(string_buffer,"RULEX",5)!=0) {
      fprintf(stderr,"File %s is not refal mod file\n",mod_name);
      exit(1);
    }
    in_word( ff, &module_length ); /* "\tModule length" */
    in_word( ff, &function_number ); /* "\tFunction number" */
    in_word( ff, &entry_number ); /* "\tEntry number" */
    in_word( ff, &extern_number ); /* "\tExtern number" */
    in_word( ff, &relocation_number ); /* "\tRelocation number" */
    in_long( ff, NULL ); /* "\tFile length" */
    in_string( ff, 8 ,string_buffer); /* "\tModule name" */
    in_byte( ff); /* "\tVersion"  */
    in_byte( ff); /*  "\tRelease"  */
 
    in_byte( ff); /*  "\tHour"  */
    in_byte( ff); /*  "\tMinutes"  */
    in_byte( ff); /*  "\tDay"  */
    in_byte( ff); /*  "\tMonth"  */
    in_word( ff, NULL ); /*  "\tYear" */

    in_byte( ff); /*  "\tCRC"  */

    register_mod(string_buffer,module_length,function_number,entry_number,extern_number);
} /* read_header */

void read_ent( int ff, int num)
{
	int i;
	uint2 addr;
	int type;
	for( i=0;i<num;i++)
		{
		in_string(ff,8,string_buffer); /* NULL */
		in_word(ff,&addr); 
		type=in_code(ff); 
		register_func(string_buffer,addr,type,FL_ENTRY);
		}
} /* read_ent */
void read_ext( int ff, int num )
{
	int i;
	int type;
	for( i=0;i<num;i++)
		{
		in_string(ff,8,string_buffer); /* NULL */
		type=in_code(ff); 
		register_func(string_buffer,-1,type,FL_EXTERN);
		}
} /* read_ext */
tfunc* read_func( int ff,int num )
{
	int i,type;
	int addr;
	tfunc* pf;
	pf=(tfunc*)malloc( num*sizeof(tfunc));
	assert( pf!=NULL);
	
	for( i=0;i<num;i++)
		{
		type = in_code(ff); /* NULL */
		in_word(ff,&addr); /* NULL */
		pf[i].addr=addr;
		in_string(ff,8,string_buffer); /* NULL */
		strcpy(pf[i].name,string_buffer);

	        register_func(pf[i].name,pf[i].addr,type,FL_LOCAL);
		
		}
	/*qsort( pf, num, sizeof(tfunc), func_cmp ); */
	return pf;
} /* read_func */

void read_text( int ff,int len, int num, tfunc* pf )
{
	int i=0;
	int k=0,z=0;
	int fnum=0;
	unsigned char byte;
	
	for(i=0,k=0,fnum=z;i<len;i++) {
		byte = in_code(ff);
		register_code(byte); 
		}
} /* read_text */
void read_rld( int ff, int num )
{
	int i;
	uint4 addr;
	uint2 tab;
	for( i=0;i<num;i++)
		{
		in_word(ff,&addr); /* NULL */
		in_word(ff,&tab); /* NULL */
		register_rld(addr,tab);
		}
} /* read_rld */
void read_end(int ff )
{
        max_mod++;
	in_byte(ff); /*  "\t"  */
} /* read_end */

/*---------------------------------*/

void in_byte( int ff )
{
	unsigned char byte;
	int status;

	status=read(ff,&byte,1);
	assert( status==1);
	return;
} /* in_byte */

unsigned char in_code( int ff )
{
	unsigned char byte;
	int status;

	status=read(ff,&byte,1);
	assert( status==1);
	return byte;
} /* in_code */

void in_word( int ff, int *pi )
{
	unsigned int word;	
	int status;
	unsigned char b1,b2;

	status=read(ff,&b1,1);
	assert(status==1);
	status=read(ff,&b2,1);
	word=byte2(b2,b1);
	assert(status==1);
	if( pi!=NULL ) *pi=word;
	return;
} /* in_word */
void in_long( int ff,  long *pl ) 
{
	unsigned long lword;	
	int status;
	unsigned char b1,b2,b3,b4;


	status=read(ff,&b1,1);
	assert(status==1);
	status=read(ff,&b2,1);
	assert(status==1);
	status=read(ff,&b3,1);
	assert(status==1);
	status=read(ff,&b4,1);
	assert(status==1);
	
	lword=byte4(b4,b3,b2,b1);
	if( pl!=NULL ) *pl=lword;
	return;
} /* in_long */

void in_string( int ff, int len,char * string_buffer)
{
	int status;
	status=read(ff,string_buffer,len);
	string_buffer[len]='\0';
	assert(status==len);
	return;
} /* in_string */

/*---------------------------*/
int scan(char *st1, char *st2) 
 {
  int i=0, l1=strlen(st1), l2=strlen(st2);
  while(st1[i]==st2[i] && i<=l1 && i<=l2) i++;
  return i;  
 }
 
uint2 seek_in_table_ext( char* f) 
{ 
uint2 p;
int l=strlen(f);
  for(p=max_ext;p>0;p--) 
    if (scan(table_ext[p-1].name,f)==l+1)
    {
      return p-1; 
    }
return ((uint2)(-1)); 
}
void check_ext()
{
  uint2 t,p;
for (t=0;t<max_ext;t++)
{
    p=table_ext[t].addr;
    if (p==-1)
    if (table_ext[t].type==TYP_FUNC)
	  {
          fprintf(stderr,"Undefined function %s in %s\n",
	  table_ext[t].name, moduls[table_ext[t].module].name);
	  }
    else if (table_ext[t].type==TYP_SYST)
    {
          table_ext[t].addr=register_func(table_ext[t].name,-1,table_ext[t].type,FL_SYSTEM); 
    }


}
}
void add_rld()
{
  uint2 p,t,z;
  register int i;
  int f=0;
  for (i=0;i<max_rld;i++) {
    t=table_rld[i].idx;
    p=table_ext[t].addr;
    z=table_rld[i].addr;
    if (p==-1)
	  {
	  p=0;
          f=1;
	  }

    t=p*sizeof(func);
    p=moduls[table_rld[i].module].addr+z;
    code[p]=(unsigned char)t;
    code[p+1]=(unsigned char)(t>>8);
    }
    /*if (f) exit(1);*/

}
void register_rld(uint4 addr, uint2 tab)
{
  assert(addr!=65535 && addr!=-1);
if (tab!=(uint2)65535) /* sys deps */
{
  table_rld[max_rld].addr=addr;
  table_rld[max_rld].module=max_mod;
  table_rld[max_rld].idx=table_max_loc[tab].addr;
  max_rld++;
}
else
{
  uint2 p=moduls[max_mod].addr,t;
  t=code[p+addr]+ (code[p+addr+1]<<8);
  t=(moduls[max_mod].first+t)*sizeof(func);
  code[p+addr]=(unsigned char)t;
  code[p+addr+1]=(unsigned char)(t>>8);
}
if (max_rld>lenght_rld-1)
 {
 fprintf(stderr,"Too big rld (use bigest model)\n");
 exit(1);
 }
}

void register_code(unsigned char byte)
{
code[max_code++]=byte;
if (max_code>lenght-2)
{
 fprintf(stderr,"Too big code (use bigest model)\n");
 exit(1);
};
}

uint2 register_mod(char * fp,uint4 len,int fn,int enn, int exn)
{
 uint2 m=max_mod;
if (max_mod>lenght_mod-1)
 {
 fprintf(stderr,"Too big mod (use bigest model)\n");
 exit(1);
 }
  max_loc=0;
  strcpy(moduls[m].name,fp);
  moduls[m].len=len;
  moduls[m].fn=fn;
  moduls[m].enn=enn;
  moduls[m].exn=exn;
  moduls[m].first=max_fun;
  moduls[m].addr=max_code;
  return m;
}


uint2 register_new_func(char * fp,uint4 addr,int type, int mod)
{
 uint2 m=max_fun;
 int l=strlen(fp);
 table[m].name=(char *)malloc(l+2);
 strcpy(table[m].name,fp);
 table[m].addr=addr;
 table[m].flag=mod; 
 table[m].spec=NO_SPEC; 
 table[m].module=max_mod; 
 table[m].type=type; 
 max_fun++;
 return m;
}
 
void register_ext(char * fp,uint4 addr,int type)
{
       table_ext[max_ext].addr=addr;
       table_ext[max_ext].type=type;
       table_ext[max_ext].module=max_mod;
       strcpy(table_ext[max_ext].name,fp);
       max_ext++;
if (max_ext>lenght_ext-1)
 {
 fprintf(stderr,"Too big ext (use bigest model)\n");
 exit(1);
 }
}
uint2 register_func(char * fp,uint4 addr,int type, int mod)
{
 uint2 m=max_fun,n;
 if (max_fun>lenght_fun-1)
 {
 fprintf(stderr,"Too big func (use bigest model)\n");
 exit(1);
 }
 switch(mod)
 {
   case FL_LOCAL: 
      return register_new_func(fp,addr,type,mod);
     break;
   case FL_ENTRY:
     if ((n=seek_in_table_ext(fp))==-1)
     {
       register_ext(fp,moduls[max_mod].first+addr,type);
     }else
     {
       table_ext[n].addr=moduls[max_mod].first+addr;
     }
     if (scan(fp,"TASK")==5) seek_task=moduls[max_mod].first+addr;
     
     break;
   case FL_EXTERN:
     if ((n=seek_in_table_ext(fp))==-1)
     {
       n=max_ext;
       register_ext(fp,-1,type);
     }
     table_max_loc[max_loc++].addr=n; 
      if (max_loc>lenght_loc-1)
       {
       fprintf(stderr,"Too big loc (use bigest model)\n");
       exit(1);
       }
     return -1;
     break;
   case FL_SYSTEM:
      return register_new_func(fp,addr,type,mod);
     break;
    default:
     assert(0);
     
 }
 return m;
}
/*-----------------------------_*/
static char * uniq(uint2 i)
{
  static char ns[200];
  if (!(table[i].flag&(FL_EXTERN|FL_SYSTEM)))
  {
    sprintf(ns,"%s_%s",table[i].name,moduls[table[i].module].name);
    return ns;
  }
  else return table[i].name;
}

char * low(char * s)
{
  static char r[10];
  int i;
for (i=0;i<strlen(s);i++)
  if (s[i]>= 'A' && s[i]<='Z') r[i]=tolower(s[i]);
   else r[i]=s[i];
r[i]=0;
return &r[0];
};

static char * well_ident(char * id)
{
int i;
static char nid[100];
strcpy(nid,id);
if ( (strcmp(nid,"for")==0)    || /* c-keywoed */
     (strcmp(nid,"do")==0)     ||
     (strcmp(nid,"const")==0)  ||
     (strcmp(nid,"if")==0)     ||
     (strcmp(nid,"double")==0) ||
     (strcmp(nid,"int")==0)    ||
     (strcmp(nid,"func")==0)      /* name of type in c-file */
     ) strcat(nid,"_x"); 
for (i=0;i<strlen(nid);i++)
 {
   if (nid[i]=='-') nid[i]='_';    /* minus */
   if ((unsigned)nid[i]>'z')       /* russian liters */
   { int 
     p=((unsigned)nid[i]-127)%52;
     if (p>25) nid[i]=97-26+p;
     else  nid[i]=65+p;
   }
 };
return nid;
}
  
/* --------------------------- */
void print_extern_functions()
{
uint2 p;
fprintf(newf,"\n");
for(p=0;p<max_fun;p++) 
     /* if (!(table[p].flag & (FL_LOCAL|FL_LOC_SPEC))) */
     if ((table[p].flag & (FL_SYSTEM)))
     {
       fprintf(newf,"extern int %s();\n",well_ident(low(table[p].name)));
     } 
fprintf(newf,"\n");
};

void print_name_table() 
{ 
 int f=0,i,j,l;
 char name[9];
 fprintf(newf,"\n\nchar nt[]= {"); 
 for (i=0;i<max_fun;i++) 
  {
   strcpy(name,table[i].name);
   l=strlen(name);
   for (j=0;j<8;j++)    
   { 
    if (f!=0)   fprintf(newf,",");  
    else f=1;  
    if (j==0 && i!=0 ) fprintf(newf,"\n            ");
    if (j>=l)  fprintf(newf,"0"); 
    else fprintf(newf,"'%c'",name[j]); 
   }
  } 
 fprintf(newf,"};\n"); 
}

void print_func_table() 
{
int f=0,i;
int type;
char name[100];
 fprintf(newf,"func ft[]= {"); 
 for (i=0;i<max_fun;i++) 
  { 
   {
    if (f==0) f=1;
    else  if (i!=0) 
      fprintf(newf,",\n            ");
    strcpy(name,well_ident(uniq(i)));
    if (table[i].flag&FL_LOCAL) {
          int d=table[i].addr;
          type=type_recoding(table[i].type);
            fprintf(newf,"{%s+%u,0,%d,0} /* %s %4d*/",
		moduls[table[i].module].name,d,type,table[i].name,
		i*sizeof(func)); 
	  }
    else  
    if ((table[i].flag&(FL_SYSTEM))) {
      fprintf(newf,"{(char *)%s,0,3,0}",well_ident(low(table[i].name))); 
    } 
    else {
      fprintf(newf,"{(char *)0,0,0,0}"); 
    };
   }
  } 
 fprintf(newf,"};\n"); 
} 

void print_struct_func()
{
 fprintf(newf,"\n"
              "typedef struct {\n"
              "  char *addr;\n"
        "  int mod;\n"
        "  unsigned char status;\n"
        "  unsigned char trace;\n"
        "  } func;\n"); 
}

void print_global_var()
{
 if (seek_task==-1) {
            fprintf(stderr,"TASK not find\n");
            seek_task=0;
 }
 fprintf(newf,"\n"
              "func *func_base=ft;\n"
        "func *func_end=&ft[%u];\n"
        "char *name_base=&nt[0];\n"
        "char *name_end=&nt[%u];\n"
        "func *start_base=&ft[%u];\n",max_fun-1,(max_fun-1)*8,seek_task); 
};
uint2 seek_addr(uint2 addr, uint2 mod)
{
uint2 f=moduls[mod].first;  
uint2 i;
for (i=f;i<f+moduls[mod].fn;i++)
{
 if (table[i].addr==addr && 
     ( table[i].type==TYP_FUNC || 
       table[i].type==TYP_SS)) return i;
}
return -1;
};

 void
print_code_table()
{
 int i,li,j=0,v=0; 
 int z=0,k=0,fnum=0,len=0;
 unsigned char byte;
 
 for (i=0;i<max_mod;i++) 
 { 
  fprintf(newf,"\nchar %s[]={\n",well_ident(moduls[i].name)); 
  j=moduls[i].addr; 
  v=j+moduls[i].len-1;
  z=moduls[i].first;
  len=z+moduls[i].fn;
  /*while( table[z].addr==0 && z<len) */
  fnum=seek_addr(0,i);
  if (fnum==-1)
  {
  fprintf(stderr,"No addr %d %s \n",i,moduls[i].name);
  /*assert(fnum!=-1);*/
  };
  if (fnum!=-1) fprintf(newf,"/* %-8s %4d: */\n        ",table[fnum].name,0);
  for(li=0,k=0;j<v;li++,j++) {
      byte = code[j];
      fprintf(newf,"%u," ,byte);
      /*
      if ( k==1 ) {
	      k1=byte;
	      }
      if ( k==2 ) {
	      k2=byte;
	      }
      */ 
      if ( (fnum=seek_addr(li+1,i))!=-1 ) {
	      fprintf(newf,"\n/* %-8s %4d: */\n        ",
		  table[fnum].name,li+1);
	      k=0;
	      continue;
	      }
      if ( k!=0 && k%18==0 ) {
	      fprintf(newf,"\n%-8s"," ");
	      }
      k++; 
      }
  if (j==v) fprintf(newf,"%u",(uint2)code[j]); 
  fprintf(newf,"};\n"); 
 } 
  
};

void  make_new_file() 
{ 
 /* make newf*/ 
 print_struct_func();
 print_extern_functions();
 print_code_table();
 print_name_table();  
 print_func_table(); 
 print_global_var();
} 
 

int type_recoding(int type)
{
      if (type==TYP_BOX) return STATUS_UNINIT;
      else if (type==TYP_EMP) return STATUS_EMPTY;
      else return STATUS_FUNC ;

}

/*------------------------------------------------------------ 
* NAME: deexp
* ARGS: f1 - fiel name, e2 new suffix
* RETURNS: file s drugim rasshireniem
* PURPOSE:zamenyaet rasshirenie u file f1 na e2
* NOTE:
------------------------------------------------------------*/
static char *  deexp(char * f1, char * e2)
{
int i;
static char w[MAX_STRING];
for (i=strlen(f1)-1;i>0 && f1[i]!='/';i--) 
  if (f1[i]=='.') break;
strcpy(w,f1);
if (f1[i]=='.') 
  if (e2[0]!=0)
     strcpy(w+i+1,e2);
  else
    w[i]=0;
else
{ i=strlen(f1);
  w[i]='.';
  strcpy(w+i+1,e2);
}
return w;
}

/*----------------------------*/
int func_cmp( const void * f1, const void * f2)
{
	return ((tfunc*)f1)->addr - ((tfunc*)f2)->addr;
} /* func_cmp */

int print_statistics()
{
  printf("Functions: %d (%2.1f%%)\n",max_fun,((double)(max_fun)/(double)(MAX_FUN)*100));
  printf("Modules: %d (%2.1f%%)\n",max_mod,(double)(max_mod)*100/MAX_MOD);
  printf("External: %d (%2.1f%%)\n",max_ext,(double)(max_ext)*100/MAX_EXT);
  printf("Max rld table: %d (%2.1f%%)\n",max_rld,((double)(max_rld)/(double)(MAX_RLD)*100));
  printf("Byte code length: %d (%2.1f%%)\n",max_code,(double)(max_code)*100/lenght);

  return 1;
}
